//This is a modified version of TautologyGenerator.java with the following
//differences supposed to improve the "quality" of the generated tautologies:
//(NOTE: some of the following checks may currently be commented out)
//- TND-tiles may only be used right after an mp-tile
//  (This implies that a TND tile must not be used as first tile)
//- The tautology is discarded if a maximum parentheses depth is exceeded
//- Trivial truths "_|_->?" and "A->A" are eliminated from the tautology, if this
//  is not possible, the tautology is discarded
//- The tautology is discarded if it has been generated through a proof that has
//  too many unnecessary arrow introductions at the end (unnecessary meaning that the
//  generated bonus square is not used)
//- The tautology is discarded if its postfix representation contains ->->->

/* Copyright (C) 2002 Matthias S. Benkmann <m.s.b@gmx.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License (ONLY THIS VERSION).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


package domino.core;

import java.util.Vector;
import java.util.List;
import java.util.Iterator;
import java.util.Random;
import domino.core.Debugging;
import java.lang.Double; //to help javadoc
import java.lang.Integer; //to help javadoc
import domino.logic.*;

/**
* <p>This class creates propositional logic tautologies.</p>
*/
public class LevelGenerator
{
  private static final boolean debugRules=false;

  private static final int MAX_PAREN_DEPTH=7; //maximum parentheses depth allowed for tautologies
  private static final int MAX_UNNECESSARY_ARROW_INTROS=2;
  
  protected DominoTree tndTile=createTNDTile();
  protected DominoTree peTile=createPETile();
  protected DominoTree mpTile=createMPTile();
  protected Random myRandom;
  
  protected static DominoTree createTNDTile()
  {
    String[] leaves={"_|_"};
    String[] bonus={"X0->_|_"};
    return FormulaDominoTree.newFromInfix("X0",leaves,bonus);
  };
  
  protected static DominoTree createPETile()
  {
    String[] leaves={"X1"};
    String[] bonus={"X0"};
    return FormulaDominoTree.newFromInfix("X0->X1",leaves,bonus);
  };
  
  protected static DominoTree createMPTile()
  {
    String[] leaves={"X0","X0->X1"};
    return FormulaDominoTree.newFromInfix("X1",leaves,null);
  };
 
 
  /**
  * <p>Creates a new TautologyGenerator with a new random number generator.</p>
  */
  public LevelGenerator()
  {
    //Math.random() is multiplied with the initial seed not because it adds
    //more randomness (it doesn't), but because this makes it more likely
    //that different TautologyGenerators get different seeds, even when started 
    //in the same millisecond.
    myRandom=new Random((long)(Math.random()*Long.MAX_VALUE)*System.currentTimeMillis());
  }; 
  
  /**
  * <p>Creates a new TautologyGenerator using random number generator 
  * <code>rand</code>.</p>
  */
  public LevelGenerator(Random rand)
  {
    myRandom=rand;
  }; 
  
  /**
  * <p>Returns the indices of the elements in <code>prob</code>
  * in a random order based on the contents of
  * <code>prob</code>.</p>
  * <p><code>prob</code> must contain {@link Double}s.</p>
  * <p>The returned {@link Vector} will contain the numbers <code>0</code> to
  * <code>prob.size()-1</code> (i.e. the indices of the elements of <code>prob</code>)
  * in random order.</p>
  * <p>Every element of 
  * <code>prob</code> is the relative weight of the index of the element within
  * <code>prob</code>. The chance that index <code>i</code> is returned as element
  * 0 of the result {@link Vector} is <code>prob.get(i)</code> divided by the sum of
  * all weights. The other elements of the result list are determined the same way
  * with the restriction that the same index may not be contained in the result
  * {@link Vector} twice.</p>
  *
  * @return {@link Vector} of {@link Integer}s.
  * @throws IllegalArgumentException if <code>prob</code> is empty or 
  * <code>null</code> or contains any element <code>&lt;0.0000001</code>.
  */
  public Vector randomPick(List prob)
  {
    final double THRESH=0.0000001;
    if (prob==null || prob.isEmpty()) throw new IllegalArgumentException();
    
    Vector result=new Vector(prob.size());
    
    Vector temp=new Vector(prob);
    double sum=0;
    Iterator i=temp.iterator();
    while (i.hasNext())
    {
      double d=((Double)i.next()).doubleValue();
      if (d<THRESH) throw new IllegalArgumentException();
      sum+=d;
    };
    
    double rnum=myRandom.nextDouble();
    
    for (int j=0; j<temp.size(); ++j)
    {
      double tmpsum=0;
      for (int index=0; index<temp.size(); ++index)
      {
        double d=((Double)temp.get(index)).doubleValue();
        if (d<0) continue;
        tmpsum+=d/sum;
        if (rnum>tmpsum) continue; 
        temp.set(index,new Double(-1));
        sum-=d;
        result.add(new Integer(index));
        break;
      };  
    };
    
    return result;
  };
  
  /**
  * <p>Similar to {@link #randomPick(List)} but returns index pairs.</p>
  * <p><code>prob</code> must contain {@link Double}s.</p>
  * <p>The returned {@link Vector} will contain the pairs <code>(0,1),...,
  * (0,prob.size()-1), (1,0), (1,2),...</code>.</p>
  * <p>Element 0 of the result {@link Vector} is the 1st index of the 1st pair.
  * Element 1 is the 2nd index of the 1st pair. Element 2 is the 1st index of the 2nd
  * pair,... </p>
  * <p>The result {@link Vector} will have <code>2*prob.size()*(prob.size()-1)</code>
  * elements.</p>
  *
  * @return {@link Vector} of {@link Integer}s.
  * @throws IllegalArgumentException if <code>prob</code> contains fewer than 2
  * elements or 
  * <code>null</code> or contains any element <code>&lt;0.0000001</code>.
  */
  public Vector randomPairPick(Vector prob)
  {
    final double THRESH=0.0000001;
    if (prob==null || prob.size()<2) throw new IllegalArgumentException();
        
    int s=prob.size();
    
    Iterator iter=prob.iterator();
    double sum=0;
    while (iter.hasNext())
    {
      double d=((Double)iter.next()).doubleValue();
      if (d<THRESH) throw new IllegalArgumentException();
      sum+=d;
    };
    
    Vector pairprob=new Vector(s*(s-1));
    for (int i=0; i<s; ++i)
    for (int j=0; j<s; ++j)
    {
      if (j==i) continue;
      double p=(((Double)prob.get(i)).doubleValue()/sum)*(((Double)prob.get(j)).doubleValue()/sum);
      pairprob.add(new Double(p));
    };
    
    Vector result=new Vector(2*s*(s-1));

    iter=randomPick(pairprob).iterator();
    while(iter.hasNext())
    {
      int index=((Integer)iter.next()).intValue();
      int i=index/(s-1);
      
      int j=index%(s-1);
      if (j>=i) ++j;
      
      result.add(new Integer(i));
      result.add(new Integer(j));
    };
    
    return result;
  };
  
  /**
  * <p>Creates a random tautology with a random proof.</p>
  * 
  * @param numVars is the maximum number of assumptions to use in the proof.
  * @param prob must contain 4 positive {@link Double}s. These values determine by
  *        their relative values the relative probabilities of the 4 rules used to
  *        create the proof. Element 0 is the weight for TND. Element 1 is the weight
  *        for ->I. Element 2 is the weight for ->E. Element 3 is the weight for
  *        the rule that uses a free assumption to remove a leaf.
  * @param bound if <code>true</code>, the returned tautology will have bound 
  *        leaves with names consisting only of uppercase letters from the
  *        range A-Z.
  * 
  * @return a random proof for a random tautology. The symbol used for false is
  *         "_|_". 
  *
  * @throws IllegalArgumentException if <code>prob</code> is empty or 
  * <code>null</code> or contains any element <code>&lt;0.0000001</code> or
  * if <code>numVars&lt;=0</code>.
  */
  public FormulaDominoTree randomTautologyProof(int numVars, Vector prob, boolean bound)
  {
    if (prob==null || prob.size()!=4 || numVars<=0) throw new IllegalArgumentException();
    
    int[] lastRule=new int[numVars];
    for (int i=0; i<numVars; ++i) lastRule[i]=-1;
    Vector trees=new Vector(numVars);
    for (int i=0; i<numVars; ++i) 
    {
      String[] leaves={"X0"};
      DominoTree var=FormulaDominoTree.newFromInfix("X0",leaves,null);
      trees.add(var);
    };
    
    while (trees.size()>1 || ((DominoTree)trees.get(0)).numberOfOpenLeaves()>0)
    {
      Vector treeprob=new Vector(trees.size());
      for (int i=0; i<trees.size(); ++i) treeprob.add(new Double(1));
      
      Iterator ruleToTry=randomPick(prob).iterator();
      boolean success=false;
      while(!success && ruleToTry.hasNext())
      {
        int rule=((Integer)ruleToTry.next()).intValue();
        
        if (rule<2) //TND or ->I
        {
          DominoTree ruleTile;
          if (rule==0) ruleTile=tndTile.cloneDT();
          else ruleTile=peTile.cloneDT();
          
          Iterator treeToTry=randomPick(treeprob).iterator();
          while (treeToTry.hasNext())
          {
            int treeidx=((Integer)treeToTry.next()).intValue();
            if (rule==0 && lastRule[treeidx]!=2) continue; //do not attach TND to anything that is not an mp tile
            DominoTree tryMe=(DominoTree)trees.get(treeidx);
            try{
              tryMe.attachTo(ruleTile,0);
              trees.set(treeidx,ruleTile);
              lastRule[treeidx]=rule;
              success=true;
              if (debugRules) System.out.println(""+rule+"/"+treeidx);
              break;
            }catch(ConflictingBindException x){};
          };
        } else 
        if (rule==2 && trees.size()>=2) //MP
        {
          DominoTree ruleTile=mpTile.cloneDT();
          Iterator treeToTry=randomPairPick(treeprob).iterator();
          while (treeToTry.hasNext())
          {
            int treeidx1=((Integer)treeToTry.next()).intValue();
            int treeidx2=((Integer)treeToTry.next()).intValue();
            DominoTree tryMe1=(DominoTree)trees.get(treeidx1);
            DominoTree tryMe2=(DominoTree)trees.get(treeidx2);
            try{
              //attach to leaf 1 first because attaching changes leaf indices
              //so if we attached to 0 first we couldn't use index 1 to refer to
              //the 2nd leaf.
              tryMe1.attachTo(ruleTile,1);
              try{
                tryMe2.attachTo(ruleTile,0);
                trees.set(treeidx1,ruleTile);
                lastRule[treeidx1]=rule;
                trees.remove(treeidx2);
                for (int j=treeidx2+1; j<lastRule.length; ++j) 
                  lastRule[j-1]=lastRule[j];
                success=true;
                if (debugRules) System.out.println(""+rule+"/"+treeidx1+"/"+treeidx2);
                break;
              }catch(ConflictingBindException x)
              {
                tryMe1.detachFromParent();
              };  
            }catch(ConflictingBindException x){};
          };  
        } else
        if (rule==3) //attach a bonus square to a leaf
        {
          Iterator treeToTry=randomPick(treeprob).iterator();
    tryTrees:
          while (treeToTry.hasNext())
          {
            int treeidx=((Integer)treeToTry.next()).intValue();
            DominoTree tryMe=(DominoTree)trees.get(treeidx);
            int lnum=tryMe.numberOfOpenLeaves();
            if (lnum==0) continue;
            
            Vector leafprob=new Vector(lnum);
            for (int i=0; i<lnum; ++i) leafprob.add(new Double(1));
            Iterator leafToTry=randomPick(leafprob).iterator();
            while (leafToTry.hasNext())
            {
              int leaf=tryMe.indexOfOpenLeaf(((Integer)leafToTry.next()).intValue());
              int bsnum=tryMe.numberOfBonusSquares(leaf);
              if (bsnum==0) continue;
              
              Vector bsprob=new Vector(bsnum);
              for (int i=0; i<bsnum; ++i) bsprob.add(new Double(1));
              Iterator bsToTry=randomPick(bsprob).iterator();
              while (bsToTry.hasNext())
              {
                int bs=((Integer)bsToTry.next()).intValue();
                try{
                  tryMe.useBonusSquare(leaf,bs);
                  lastRule[treeidx]=rule;
                  success=true;
                  if (debugRules) System.out.println(""+rule+"/"+treeidx+"/"+leaf+"/"+bs);
                  break tryTrees;
                }catch(ConflictingBindException x){};
              };
            };
          };
        };
      };
    
      if (!success) throw new RuntimeException("Can't find applicable rule");
    };
    
    FormulaDominoTree result=(FormulaDominoTree)trees.get(0);
    
    if (bound)
    try{
      String zstr="Z";
      byte[] zdat=zstr.getBytes("ISO8859_1");
      final byte z=zdat[0];
      String astr="A";
      byte[] adat=astr.getBytes("ISO8859_1");
      final byte a=adat[0];
      byte[] dat={a};
      int unboundNum=result.numberOfUnboundVariablesInRoot();
      int unboundNum2=unboundNum+result.numberOfUnboundVariablesInLeaves();
      String[] ids=new String[unboundNum2];
      for (int i=0; i<unboundNum2; ++i)
      {
        ids[i]=new String(dat,"ISO8859_1");
        boolean append=true;
        for (int j=0; j<dat.length; ++j)
        {
          if (dat[j]<z) {++dat[j]; append=false; break;};
          dat[j]=a;
        };
        if (append)
        {
          byte[] newdat=new byte[dat.length+1];
          System.arraycopy(dat,0,newdat,0,dat.length);
          dat=newdat;
          dat[dat.length-1]=a;
        };  
      };  
     
      result.bindUnboundVariablesInRoot(ids);
      unboundNum2=result.numberOfUnboundVariablesInLeaves();
      String[] ids2=new String[unboundNum2];
      System.arraycopy(ids,unboundNum,ids2,0,unboundNum2);
      result.bindUnboundVariablesInLeaves(ids2);
      
    }catch(Exception x)
    {
      throw new RuntimeException("Unexpected exception: "+x.toString());
    };  
    
    
    return result;  
  };
  

  protected static String eliminateTruth(String taut, int i)
  {
    while(true)
    {
      int left=i;
      int plevel=0;
      while (plevel>=0)
      {
        --left;
        if (left<0) return null;
        if (taut.charAt(left)=='(') --plevel; else
        if (taut.charAt(left)==')') ++plevel;
      };
    
      int right=i;
      plevel=0;
      while (plevel<=0)
      {
        ++right;
        if (right>=taut.length()) return null;
        if (taut.charAt(right)=='(') --plevel; else
        if (taut.charAt(right)==')') ++plevel;
      };
    
      while (taut.charAt(left)=='(' && taut.charAt(right)==')') 
      {
        --left;
        ++right;
        if (left<0 || right>=taut.length()) return null;
      };
    
      if (taut.charAt(right)!=')') //can only mean that taut.charAt(right)=='-'
      {
        int rightEnd=right+2;
        plevel=0;
        while (plevel<=0)
        {
          ++rightEnd;
          if (taut.charAt(rightEnd)=='(') --plevel; else
          if (taut.charAt(rightEnd)==')') ++plevel;
        };
        return taut.substring(0,left)+taut.substring(right+2,rightEnd)+
               taut.substring(rightEnd+1);
      };
      //else taut.charAt(left)=='>'
    
      i=left;
    }
  };

  
  /**
  * <p>Outputs a random tautology on stdout.</p>
  */
  public static void main(String[] args)
  {
    if (args.length!=5 && args.length!=1)
    {
      System.err.println("USAGE: numVars TND ->I ->E []");
      System.err.println("  or   numVars");
      System.exit(1);
    };
    Random rand0=new Random();
    long seed=rand0.nextLong();
    
    if (args.length==1)
    {
      String numVars=args[0];
      args=new String[5];
      args[0]=numVars;
      for (int i=1; i<=4; ++i)
      {
        args[i]=""+(Math.round(rand0.nextDouble()*4)+1);
      };
    };

//    seed=7031244344659850909L; System.out.println("Fixed seed!");

    System.err.println("Random seed: "+seed+"  Params: "+args[0]+" "+args[1]+" "+args[2]+" "+args[3]+" "+args[4]);
    rand0=new Random(seed);
    LevelGenerator tgen=new LevelGenerator(rand0);
    Vector prob=new Vector();
    prob.add(new Double(Double.parseDouble(args[1])));
    prob.add(new Double(Double.parseDouble(args[2])));
    prob.add(new Double(Double.parseDouble(args[3])));
    prob.add(new Double(Double.parseDouble(args[4])));
    FormulaDominoTree randomTautProof=tgen.randomTautologyProof(Integer.parseInt(args[0]),prob,true);
    
    FormulaDominoSquare root=(FormulaDominoSquare)randomTautProof.rootSquare();
    String tautStr=root.toInfixRepresentation();
    String tautStrPost=root.toPostfixRepresentation();
    
    FormulaDominoTree rootCopy=(FormulaDominoTree)randomTautProof.cloneDT();
    int numberOfUnnecessaryArrowIntroductions=0;
    while (rootCopy.numberOfSubtreeConnections()==1 &&
           !((FormulaDominoSquare)rootCopy.subtreeConnectionSquare(0)).
           toPrefixRepresentation().equals("(_|_)"))
    {
      int numOfOpenLeaves=rootCopy.numberOfOpenLeaves();
      rootCopy=(FormulaDominoTree)rootCopy.subtree(0);
      rootCopy.detachFromParent();
      if (numOfOpenLeaves==rootCopy.numberOfOpenLeaves())
        ++numberOfUnnecessaryArrowIntroductions;

      if (numberOfUnnecessaryArrowIntroductions>MAX_UNNECESSARY_ARROW_INTROS) 
      {
        System.err.println("Rejecting ->I: "+tautStr);
        System.exit(1); 
      };
    };
    
    
/*    int scanIndex=tautStr.indexOf("_|_->");
    while(scanIndex>=0) 
    { 
      System.err.println("Fixing _|_->: "+tautStr);
      String newTautStr=eliminateTruth(tautStr,scanIndex);
      if (newTautStr==null)
      {
        System.err.println("Rejecting _|_->: "+tautStr);
        System.exit(1); 
      };
      System.err.println("Fixed: "+newTautStr);
      tautStr=newTautStr;
      scanIndex=tautStr.indexOf("_|_->");
    };
    
    scanIndex=tautStr.indexOf("->");
    while (scanIndex>=0)
    {
      int k=scanIndex-1;
      while (k>=0 && Character.isLetter(tautStr.charAt(k))) --k;
      k=k+1;
      if (tautStr.charAt(k)!='-')
      {
        int j=scanIndex+2;
        while (j<tautStr.length() && tautStr.charAt(k)!='-' && 
               tautStr.charAt(k)==tautStr.charAt(j)) {++k; ++j;};
        if (tautStr.charAt(k)=='-' && 
            (j>=tautStr.length() || !Character.isLetter(tautStr.charAt(j))))
            {
               System.err.println("Fixing A->A: "+tautStr);
               String newTautStr=eliminateTruth(tautStr,k);
               if (newTautStr==null)
               {
                 System.err.println("Rejecting A->A: "+tautStr);
                 System.exit(1);
               };  
               System.err.println("Fixed: "+newTautStr);
               tautStr=newTautStr;
               scanIndex=-1;
            };  
      };
      
      scanIndex=tautStr.indexOf("->",scanIndex+1);
    };
    
    if (tautStrPost.indexOf("->->->")>=0) { System.err.println("Rejecting ->->->: "+tautStr);System.exit(1); };
*/  
    if (tautStrPost.endsWith("->->->")) { System.err.println("Rejecting ->->->: "+tautStr);System.exit(1); };
  
    int pdepth=0;
    for (int i=0; i<tautStr.length(); ++i)
    {
      if (tautStr.charAt(i)=='(') 
      {
        ++pdepth; 
        if (pdepth>MAX_PAREN_DEPTH) { System.err.println("Rejecting (()): "+tautStr); System.exit(1);};
      } else
      if (tautStr.charAt(i)==')') --pdepth;
      
      
    };
    
    String[] diag=ProofFormatter.naturalDeductionDiagram(randomTautProof);
    for (int i=0; i<diag.length; ++i) System.out.println(diag[i]);
    System.out.println();

    System.out.println(args[0]+","+args[1]+","+args[2]+","+args[3]+","+args[4]+","+seed+": "+tautStr);
/*    
    FormulaDominoTree sub=(FormulaDominoTree)randomTautProof.subtree(0).subtree(0).subtree(0);
    sub.detachFromParent();
    FormulaDominoTree.UnboundIdAssigner ass=FormulaDominoTree.UnboundIdAssigner.newAssigner();
    diag=ProofFormatter.naturalDeductionDiagram(randomTautProof,ass);
    for (int i=0; i<diag.length; ++i) System.out.println(diag[i]);
    System.out.println();
    diag=ProofFormatter.naturalDeductionDiagram(sub,ass);
    for (int i=0; i<diag.length; ++i) System.out.println(diag[i]);
    System.out.println();
    
    ass.removeMappingsFor(sub);
    
    sub.attachTo(randomTautProof,2);
    diag=ProofFormatter.naturalDeductionDiagram(randomTautProof,ass);
    for (int i=0; i<diag.length; ++i) System.out.println(diag[i]);
    System.out.println();
*/    
  };
};

