/* Copyright (C) 2002 Matthias S. Benkmann <m.s.b@gmx.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License (ONLY THIS VERSION).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package domino.logic;

/**
* <p>This interface is implemented by classes that represent domino tiles connected
* to form a tree.</p>
*/
public interface DominoTree
{
  /**
  * <p>If the <code>DominoTree</code> is a subtree of a larger tree, it is
  * detached from the parent tree. If the tree is not a subtree, nothing happens.</p>
  * <p>All bonus squares in the subtree to be detached, that originate from
  * <code>DominoTree</code>s not part of the detached subtree will be detached
  * from the subtree and cannot be reattached.</p>
  *
  * @return <code>true</code> if subtree was detached; <code>false</code> if tree
  *         had no parent tree to begin with.
  *
  * @see #attachTo(DominoTree,int)
  */
  public boolean detachFromParent();
  
  /**
  * <p>Connects the root of <code>this</code> to <code>leaf</code>
  * of <code>DominoTree</code> <code>to</code>.</p>
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;to.numberOfLeaves() && to.isOpenLeaf(leaf)</code>.</p>
  *
  * @throws IllegalArgumentException if <code>to</code> is a subtree of 
  *         <code>this</code>.
  *         The trees remain unchanged.
  * @throws IllegalNodeTypeException if <code>this</code> is already connected, or is
  *         a bonus square that has been detached or if the leaf to attach to is not
  *         open.
  *         The trees remain unchanged.
  * @throws SquareDoesNotExistException if <code>to==null</code> or 
  *         <code>leaf</code> is not a valid leaf 
  *         number. The trees remain unchanged.
  * @throws ConflictingBindException if the root square does not match the leaf to
  *         connect with. The trees remain unchanged.
  * @throws ClassCastException if <code>this</code> and <code>to</code>
  *         have incompatible types. The trees remain unchanged.
  *
  * @see #attachToSubtreeConnection(DominoTree, int)
  * @see #detachFromParent()
  * @see #numberOfLeaves()
  * @see #isOpenLeaf(int)
  */
  public void attachTo(DominoTree to, int leaf);
  
  
  /**
  * <p>Like {@link #attachTo(DominoTree,int)} with the difference that instead of a leaf
  * index, a subtree connection index is used.</p>
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;to.numberOfSubtreeConnections() && 
  * to.isOpenLeaf(leafIndexForSubtreeConnection(i))</code>.</p>
  *
  * @throws IllegalArgumentException if <code>to</code> is a subtree of 
  *         <code>this</code>.
  *         The trees remain unchanged.
  * @throws IllegalNodeTypeException if <code>this</code> is already connected, or is
  *         a bonus square that has been detached or if the leaf to attach to is not
  *         open.
  *         The trees remain unchanged.
  * @throws SquareDoesNotExistException if <code>to==null</code> or 
  *         <code>i</code> is not a valid subtree connection index. 
  *         The trees remain unchanged.
  * @throws ConflictingBindException if the root square does not match the leaf to
  *         connect with. The trees remain unchanged.
  * @throws ClassCastException if <code>this</code> and <code>to</code>
  *         have incompatible types. The trees remain unchanged.
  *
  * @see #attachTo(DominoTree, int)
  * @see #detachFromParent()
  * @see #numberOfLeaves()
  * @see #isOpenLeaf(int)
  */
  public void attachToSubtreeConnection(DominoTree to, int i);
  
  /**
  * <p>Returns the parent tree of <code>this</code> tree, if there is one, or 
  * <code>null</code> if <code>this</code> tree does not have a parent.</p>
  *
  * @see #parentSubtreeIndex()
  */
  public DominoTree parent();
  
  /**
  * <p>If <code>this</code> tree has a {@link #parent()},
  * <code>parent().subtree(parentSubtreeIndex())==this</code>, otherwise -1 is
  * returned.</p>
  */
  public int parentSubtreeIndex();
  
  
  /**
  * <p>Returns the number of subtrees that may be connected to <code>this</code>.</p>
  * <p>Every <code>DominoTree</code> consists of a root tree with connections
  * for subtrees. This function returns the number of connections regardless of 
  * whether actual subtrees are connected to them or not.</p>
  *
  * @see #subtree(int)
  */
  public int numberOfSubtreeConnections();
  
  /**
  * <p>If a subtree is connected to subtree connection <code>i</code>, then that
  * subtree is returned. Otherwise this function returns <code>null</code>.</p>
  * 
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;to.numberOfSubtreeConnections()</code>.</p>
  *
  * @throws IllegalArgumentException if <code>i</code> has an illegal value.
  *
  * @see #numberOfSubtreeConnections()
  */
  public DominoTree subtree(int i);
  
  /**
  * <p>Returns the number of leaves this tree has that are not bonus squares.</p>
  */
  public int numberOfOpenLeaves();
  
  /**
  * <p>Returns the number of leaves this tree has.</p>
  */
  public int numberOfLeaves();
  
  /**
  * <p>Returns <code>true</code> if <code>leaf</code> is open.</p>
  *
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;numberOfLeaves()</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  */
  public boolean isOpenLeaf(int leaf);
  
  /**
  * <p>Returns the index of the <code>n</code>-th open leaf.</p>
  *
  * <p>Constraints for <code>n</code>: 
  * <code>0&lt;=n&lt;numberOfOpenLeaves()</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>n</code> is not valid.
  */
  public int indexOfOpenLeaf(int n);
  
  
  /**
  * <p>Returns the index of the first leaf found in subtree <code>i</code>.</p>
  *
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;numberOfSubtreeConnections()</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>i</code> is not a valid subtree
  *         connection index.
  */
  public int leafIndexForSubtreeConnection(int i);
  
  /**
  * <p>Returns the number of bonus {@link DominoSquare}s the player may use in a 
  * subtree attached to <code>leaf</code>. This counts bonus squares coming from
  * parent nodes of <code>this</code> as well as new bonus squares not available in
  * parent nodes.</p>
  * <p>A <code>FormulaDominoTree</code> may provide the same bonus square multiple
  * times and the same bonus square may be provided by <code>this</code> and/or
  * several parents at the same time. These duplicates will be included in the count
  * returned by this function!</p>
  * <p>This method may be called even on non-open leaves although bonus squares
  * cannot be used on them. In this case the return value is the number of bonus
  * squares that could be used if the leaf were open.</p>
  *
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;numberOfLeaves()</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  */
  public int numberOfBonusSquares(int leaf);
  
  /**
  * <p>Returns the number of new bonus {@link DominoSquare}s the player may use in a 
  * subtree attached to subtree connection<code>i</code>. This does not count bonus 
  * squares coming from parent nodes of <code>this</code>.
  * <p>A <code>FormulaDominoTree</code> may provide the same bonus square multiple
  * times and the same bonus square may be provided by <code>this</code> and/or
  * several parents at the same time, so the count returned by this function may
  * include squares that are also provided by parent trees.</p>
  *
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;numberOfSubtreeConnections()</code>.</p>
  *
  * @throws IllegalArgumentException if <code>i</code> is not a valid number.
  * 
  * @see #numberOfSubtreeConnections()
  * @see #subtree(int)
  */
  public int numberOfNewBonusSquaresForSubtreeConnection(int i);
  
  /**
  * <p>Attaches to <code>leaf</code> the <code>num</code>-th bonus square that 
  * may be used in subtrees attached to <code>leaf</code>.</p>
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;numberOfLeaves()</code>.</p>
  * <p>Constraints for <code>num</code>: 
  * <code>0&lt;=num&lt;numberOfBonusSquares(leaf)</code>.</p>
  *
  * @return the {@link DominoTree} that was attached.
  *
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  * @throws ConflictingBindException if the bonus square and the leaf are connected
  *                                  in a way that would cause recursion or are
  *                                  otherwise incompatible.
  * @throws IllegalNodeTypeException <code>leaf</code> is not open.
  * @throws IllegalArgumentException if <code>num</code> has an illegal value.
  *
  * @see #rootSquare()
  * @see #leafSquare(int)
  * @see #bonusSquare(int,int)
  */
  public DominoTree useBonusSquare(int leaf, int num);
  
  /**
  * <p>Attaches to subtree connection <code>i</code> the bonus square returned by
  * <code>tree.newBonusSquareForSubtreeConnection(subtree,newBonusForSubtreeIndex)</code>.</p>
  * <p><code>tree</code> must be <code>this</code> or a direct or indirect parent of 
  * <code>this</code>.</p>
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;numberOfSubtreeConnections()</code>.</p>
  * <p>Constraints for <code>subtree</code>: 
  * <code>0&lt;=subtree&lt;tree.numberOfSubtreeConnections()</code>.</p>
  * <p>Constraints for <code>newBonusForSubtreeIndex</code>: 
  * <code>0&lt;=newBonusForSubtreeIndex&lt;tree.numberOfNewBonusSquaresForSubtreeConnection(subtree)</code>.</p>
  *
  * @return the {@link DominoTree} that was attached.
  *
  * @throws SquareDoesNotExistException if <code>i</code> is not a valid subtree
  *                                     connection index or 
  *                                     <code>newBonusForSubtreeIndex</code> 
  *                                     has an illegal value.
  * @throws ConflictingBindException if the bonus square and the leaf are connected
  *                                  in a way that would cause recursion or are
  *                                  otherwise incompatible, or if <code>tree</code>
  *                                  is not a direct or indirect parent of 
  *                                  <code>leaf</code>.
  * @throws IllegalNodeTypeException a subtree is connected to connection 
  *                                  <code>i</code> (i.e. it is not open).
  * @throws IllegalArgumentException if <code>subtree</code> is not a valid index.
  *
  * @see #rootSquare()
  * @see #leafSquare(int)
  * @see #bonusSquare(int,int)
  * @see #newBonusSquareForSubtreeConnection(int,int)
  */
  public DominoTree useBonusSquare(int i, DominoTree tree, int subtree, int newBonusForSubtreeIndex);
  
  /**
  * <p>Returns a copy of the <code>DominoTree</code>. The copy will have the same
  * structure as the original but will not share any physical data, nor will it
  * be connected with the original in any way. This means that the clone can be
  * used independent of the orginal as if the clone had been synthesized directly
  * with no involvement of the original tree.</p>
  * <p>Note that if this function is called on a subtree of a larger tree, the
  * whole tree will be cloned and the return value of this function will be the
  * subtree of the large clone tree that corresponds to the tree being cloned.</p>
  */
  public DominoTree cloneDT();
  
  /**
  * <p>See {@link #cloneDT()}.</p>
  */ 
  public Object clone();
  
  /**
  * <p>Returns the root square.</p>
  * @see #leafSquare(int)
  * @see #bonusSquare(int,int)
  * @see #newBonusSquareForSubtreeConnection(int,int)
  * @see #subtreeConnectionSquare(int)
  */
  public DominoSquare rootSquare();
  
  /**
  * <p>Returns the square for <code>leaf</code>.</p>
  *
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;numberOfLeaves()</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  *
  * @see #rootSquare()
  * @see #bonusSquare(int,int)
  * @see #newBonusSquareForSubtreeConnection(int,int)
  * @see #subtreeConnectionSquare(int)
  */
  public DominoSquare leafSquare(int leaf);
  
  /**
  * <p>Returns <code>num</code>-th bonus square usable for <code>leaf</code>.</p>
  *
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;numberOfLeaves()</code>.</p>
  * <p>Constraints for <code>num</code>: 
  * <code>0&lt;=num&lt;numberOfBonusSquares(leaf)</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  * @throws IllegalArgumentException if <code>num</code> has an illegal value;
  *
  * @see #rootSquare()
  * @see #leafSquare(int) 
  * @see #newBonusSquareForSubtreeConnection(int,int)
  * @see #subtreeConnectionSquare(int)
  */
  public DominoSquare bonusSquare(int leaf, int num);
  
  /**
  * <p>Returns the {@link DominoSquare} for subtree connection <code>i</code>.</p>
  * <p>This method works even for subtree connections with no attached subtree.</p>
  *
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;numberOfSubtreeConnections()</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>i</code> is not a valid
  *         subtree connection index.
  */
  public DominoSquare subtreeConnectionSquare(int i);
  
  /**
  * <p>Returns <code>num</code>-th new bonus square usable for subtree connection
  * <code>i</code>.</p>
  *
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;numberOfSubtreeConnections()</code>.</p>
  * <p>Constraints for <code>num</code>: 
  * <code>0&lt;=num&lt;numberOfNewBonusSquaresForSubtreeConnection(i)</code>.</p>
  *
  * @throws IllegalArgumentException if <code>i</code> has an illegal value;
  * @throws SquareDoesNotExistException if <code>num</code> is not a valid number
  *         for subtree connection <code>i</code>.
  *
  * @see #rootSquare()
  * @see #leafSquare(int) 
  * @see #bonusSquare(int, int)
  * @see #subtreeConnectionSquare(int)
  * @see #numberOfSubtreeConnections()
  * @see #numberOfNewBonusSquaresForSubtreeConnection(int)
  */
  public DominoSquare newBonusSquareForSubtreeConnection(int i, int num);
};


