/* Copyright (C) 2002 Matthias S. Benkmann <m.s.b@gmx.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License (ONLY THIS VERSION).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package domino.logic;

import domino.core.Debugging;
import java.util.Vector;
import java.util.Collection;
import java.util.Map;
import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.lang.ClassCastException; //to help javadoc

/**
* <p>This is a class for {@link DominoTree}s that represent proofs in propositional
* logic.</p>
*/
public class FormulaDominoTree implements DominoTree
{
  /**
  * <p>The methods of this class all assume that the {@link FormulaSyntaxNet}
  * {@link #root_} and those contained in {@link #leaves_} and {@link #bonus_}
  * don't contain forward nodes that have not been created by a call to
  * {@link #attachTo(DominoTree,int)} or {@link #useBonusSquare(int,int)}.</p>
  */
  protected FormulaSyntaxNet root_;
  
  /**
  * <p>The methods of this class assume that leaves_.length==0 characterizes a
  * bonus square. Normal <code>FormulaDominoTree</code>s without leaves are not
  * possible.</p>
  */
  protected FormulaSyntaxNet[] leaves_;
  
  /**
  * <p>NOTE: Right now every FormulaDominoTree may have several bonus squares but does
  * not allow restricting them to certain leaves only. In order to create 
  * FormulaDominoTrees like one for exists-elimination where leaf-dependent bonus 
  * squares are needed, the FormulaDominoTree can be assembled from 3 
  * FormulaDominoTrees.</p>
  */
  protected FormulaSyntaxNet[] bonus_;
  
  /** 
  * <p><code>bonusSquaresInAction_[i]</code> is a Collection of 
  * {@link BonusFormulaDominoTree}s that are 
  * unified with {@link #bonus_ bonus_[i]}.</p>
  */
  protected Collection[] bonusSquaresInAction_; 
  
  private int numOfOpenLeavesCache_;
  private int numOfLeavesCache_;
  private int[] numOfOpenLeavesCacheForConnection_;
  private int[] numOfLeavesCacheForConnection_;

  protected FormulaDominoTree[] leafConnections_;
  
  protected FormulaDominoTree rootConnection_=null;
  /** 
  * <p><code>rootConnection_.leafConnections_[rootConnectionLeafIndex_]==this</code></p>
  * <p>If <code>rootConnection_==null</code> then 
  * <code>rootConnectionLeafIndex_==-1</p>
  */
  protected int rootConnectionLeafIndex_=-1;
  
  
  /**
  * <p>Returns the parent tree of <code>this</code> tree, if there is one, or 
  * <code>null</code> if <code>this</code> tree does not have a parent.</p>
  *
  * @see #parentSubtreeIndex()
  */
  public DominoTree parent()
  {
    return rootConnection_;
  };
  
  /**
  * <p>If <code>this</code> tree has a {@link #parent()},
  * <code>parent().subtree(parentSubtreeIndex())==this</code>, otherwise -1 is
  * returned.</p>
  */
  public int parentSubtreeIndex()
  {
    return rootConnectionLeafIndex_;
  };
  
  /**
  * <p>Returns a new {@link Collection} of the type used for 
  * <code>bonusSquaresInAction_</code>.</p>
  */
  protected static Collection newCollectionUsedForBonusSquaresInAction() //TESTED
  {
    return new HashSet();
  };
  
  /**
  * <p>Returns the number of open leaves connected with index <code>i</code>.</p>
  * <p>If <code>leafConnections_[i]==null</code>, this function returns 1.</p>
  * <p>If <code>leafConnections_[i]!=null</code>, this function returns
  * <code>leafConnections_[i].numberOfOpenLeaves()</code> .</p>
  *
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;leaves_.length</code>.</p>
  *
  */
  protected int numOfOpenLeavesForConnection(int i) //NOT cache-clean (meaning this function must not be called when caches are in an inconsistent state), TESTED
  { 
    if (numOfOpenLeavesCacheForConnection_[i]<0)
    {
      if (leafConnections_[i]!=null) 
      {
        numOfOpenLeavesCacheForConnection_[i]=leafConnections_[i].numberOfOpenLeaves();
      }  
      else //if (leafConnections_[i]==null)
      { 
        numOfOpenLeavesCacheForConnection_[i]=1;
      };  
    };
    
    return numOfOpenLeavesCacheForConnection_[i];
  };
  
  
  /**
  * <p>Returns the number of leaves connected with index <code>i</code>.</p>
  * <p>If <code>leafConnections_[i]==null</code>, this function returns 1.</p>
  * <p>If <code>leafConnections_[i]!=null</code>, this function returns
  * <code>max{leafConnections_[i].numberOfLeaves(),1}</code> .</p>
  *
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;leaves_.length</code>.</p>
  *
  */
  protected int numOfLeavesForConnection(int i) //NOT cache-clean, TESTED
  {
    if (numOfLeavesCacheForConnection_[i]<0)
    {
      if (leafConnections_[i]!=null) 
      {
        int j=leafConnections_[i].numberOfLeaves();
        if (j==0) j=1;
        numOfLeavesCacheForConnection_[i]=j;
      }  
      else //if (leafConnections_[i]==null)
      { 
        numOfLeavesCacheForConnection_[i]=1;
      };  
    };
    
    return numOfLeavesCacheForConnection_[i];
  };
  
  
  
  /**
  * <p>The class used by {@link #mapLeaf(int)}.</p>
  */
  protected static class LeafMapping
  {
    /** <p>The index in <code>leafConnections_</code> of the subtree that
    * contains the leaf.</p> */
    int index;
    /** <p>The leaf index within the subtree.</p>*/
    int subindex;
    public LeafMapping(int ind, int subind) {index=ind; subindex=subind;};
  };
  
  /**
  * <p>Returns the {@link LeafMapping#index index} in <code>leafConnections_</code> 
  * to find the subtree 
  * that knows <code>leaf</code> under index {@link LeafMapping#subindex subindex}.</p>
  * <p>If no subtree is attached at <code>index</code> or if the attached subtree
  * has no leaves (i.e. it is a bonus square), <code>subindex</code> 
  * is set to -1.</p>
  *
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  */
  protected LeafMapping mapLeaf(int leaf) //NOT cache-clean, TESTED
  {
    if (leaf>=0)
    {
      for (int i=0; i<leafConnections_.length; ++i)
      {
        int n=numOfLeavesForConnection(i);
        if (leaf<n)
        {
          if (leafConnections_[i]==null || leafConnections_[i].leaves_.length==0) leaf=-1;
          return new LeafMapping(i,leaf);
        };
        leaf-=n;
      };
    };  
    throw new SquareDoesNotExistException();
  };
  
  
  /**
  * <p>The class used by {@link #locateLeaf(int)}.</p>
  */
  protected static class LeafLocation
  {
    /** <p>The tree that contains the leaf.</p> */
    FormulaDominoTree tree;
    /** <p><code>tree.leafConnections_[index]</code> is the leaf.</p> */
    int index;
    public LeafLocation(FormulaDominoTree t, int ind) {index=ind; tree=t;};
  };
  
  /**
  * <p>Returns the {@link LeafLocation#tree tree} that has
  * <code>leaf</code> under index {@link LeafLocation#index index}.</p>
  * <p>If the returned {@link LeafLocation#index} is -1, this means that 
  * {@link LeafLocation#tree}
  * is already the desired leaf (i.e. the leaf is a bonus square).
  * Otherwise <code>tree.leafConnections_[index]==null</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  */
  protected LeafLocation locateLeaf(int leaf) //NOT cache-clean, TESTED
  {
    FormulaDominoTree tree=this;
    LeafMapping l=tree.mapLeaf(leaf);
    while(l.subindex>=0)
    {
      tree=tree.leafConnections_[l.index];
      leaf=l.subindex;
      l=tree.mapLeaf(leaf);
    };

    if (tree.leafConnections_[l.index]!=null) 
    {
      tree=tree.leafConnections_[l.index];
      l.index=-1;
    };  
    
    return new LeafLocation(tree, l.index);
  };
  
  
  /** 
  * <p>Returns <code>true</code> if <code>leafConnections_[i]==null</code>.</p>
  * 
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;leaves_.length</code>.</p>
  *
  */
  protected boolean isOpenConnection(int i) //cache-clean
  {
    return (leafConnections_[i]==null);
  };
  
  /**
  * <p>Returns <code>true</code> if <code>leaf</code> is open.</p>
  *
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;numberOfLeaves()</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  */
  public boolean isOpenLeaf(int leaf) //NOT cache-clean, TESTED
  {
    LeafLocation l=locateLeaf(leaf);
    
    if (l.index>=0) 
      return true;
    else
      return false;
  };
  
  /**
  * <p>Returns the index of the <code>n</code>-th open leaf.</p>
  *
  * <p>Constraints for <code>n</code>: 
  * <code>0&lt;=n&lt;numberOfOpenLeaves()</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>n</code> is not valid.
  */
  public int indexOfOpenLeaf(int n) //NOT cache-clean, TESTED
  {
    if (n>=0)
    {
      int index=0;
      for (int i=0; i<leafConnections_.length; ++i)
      {
        int num=numOfOpenLeavesForConnection(i);
        if (n<num)
        {
          if (leafConnections_[i]==null)//case leafConnections_[i].leaves_.length==0 does not have to be tested because in this case num would be 0 so that n<num would be false
            return index;
          else return index+leafConnections_[i].indexOfOpenLeaf(n);
        };
        n-=num;
        index+=numOfLeavesForConnection(i);
      };
    };  
    throw new SquareDoesNotExistException();
  };
  
  
  /**
  * <p>Returns the index of the first leaf found in subtree <code>i</code>.</p>
  *
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;numberOfSubtreeConnections()</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>i</code> is not a valid subtree
  *         connection index.
  */
  public int leafIndexForSubtreeConnection(int i) //NOT cache-clean
  {
    if (i<0 || i>=numberOfSubtreeConnections()) throw new SquareDoesNotExistException();
    
    int idx=0;
    for (int j=0; j<i; ++j) idx+=numOfLeavesForConnection(j);
    return idx;
  };
 
 
  /**
  * <p>Removes all forward nodes and forward connections from all
  * {@link FormulaSyntaxNet}s in <code>this</code> and all of its subtrees
  * (but not from parents and siblings unless their
  * {@link FormulaSyntaxNet}s are unified with those of <code>this</code>).</p>
  *
  * @see #recreateAllForwardsDownwards()
  */
  protected void removeAllForwardsDownwards() //cache-clean, TESTED
  {
    FormulaSyntaxNet.removeAllForwards(root_);
    for (int i=0; i<leaves_.length; ++i)
    {
      FormulaSyntaxNet.removeAllForwards(leaves_[i]);
    };
    
    for (int i=0; i<bonus_.length; ++i)
    {
      FormulaSyntaxNet.removeAllForwards(bonus_[i]);
    };
    
    for (int i=0; i<leafConnections_.length; ++i)
    {
      if (leafConnections_[i]!=null) leafConnections_[i].removeAllForwardsDownwards();
    };
  };
  
  
  /**
  * <p>Recursively reunifies the appropriate {@link FormulaSyntaxNet}s 
  * of <code>this</code> and
  * of its subtrees (but not those of parents and siblings).</p>
  * <p>Old forward connections are <b>not</b> removed.</p>
  *
  * @see #removeAllForwardsDownwards()
  */
  protected void recreateAllForwardsDownwards() //cache-clean, TESTED
  {
    for (int i=0; i<leaves_.length; ++i)
    {
      if (leafConnections_[i]!=null) FormulaSyntaxNet.unify(leaves_[i],leafConnections_[i].root_);
    };
    
    for (int i=0; i<bonus_.length; ++i)
    {
      Iterator iter=bonusSquaresInAction_[i].iterator();
      while(iter.hasNext()) 
      {
        FormulaDominoTree bonusInstance=(FormulaDominoTree)iter.next();
        FormulaSyntaxNet.unify(bonus_[i],bonusInstance.root_);
      };  
    };
    
    for (int i=0; i<leafConnections_.length; ++i)
    {
      if (leafConnections_[i]!=null) leafConnections_[i].recreateAllForwardsDownwards();
    };
  };
  
  
  /**
  * <p>Removes all forward connections and then recursively reunifies the whole 
  * tree.</p>
  * <p>This function need not be called on the top-level root. It will automatically 
  * ascend as far as possible before beginning operation.</p>
  */
  protected void reunifyWholeTree() //cache-clean, TESTED
  {
    FormulaDominoTree tree=this;
    while(tree.rootConnection_!=null) tree=tree.rootConnection_;
    tree.removeAllForwardsDownwards();
    tree.recreateAllForwardsDownwards();
  };
  
  
  /**
  * <p>Finds and detaches all bonus squares that originate in a
  * <code>FormulaDominoTree</code> contained in <code>parents</code>.</p>
  * <p>This function does not deunify or reunify the tree or 
  * invalidate any caches.
  * This means that after calling this function the <code>FormulaDominoTree</code>
  * is likely to be in an inconsistent state.</p>
  */
  protected void detachAllBonusSquaresOriginatingIn(Collection parents) //cache-clean, TESTED
  {
    /*
      This function is overridden in BonusFormulaDominoTree to do the actual
      detaching. Non-bonus squares just pass the command on to their children.
    */  
    for (int i=0; i<leafConnections_.length; ++i)
      if (leafConnections_[i]!=null) leafConnections_[i].detachAllBonusSquaresOriginatingIn(parents);
  };
  
  /**
  * <p>If the <code>DominoTree</code> is a subtree of a larger tree, it is
  * detached from the parent tree. If the tree is not a subtree, nothing happens.</p>
  * <p>All bonus squares in the subtree to be detached, that originate from
  * <code>DominoTree</code>s not part of the detached subtree will be detached
  * from the subtree and cannot be reattached.</p>
  *
  * @return <code>true</code> if subtree was detached; <code>false</code> if tree
  *         had no parent tree to begin with.
  *
  * @see #attachTo(DominoTree,int)
  */
  public boolean detachFromParent() //cache-clean, TESTED
  {
    if (rootConnection_==null) return false;
    
    //collect all parent nodes
    Collection parents=new HashSet();
    FormulaDominoTree parent=rootConnection_;
    FormulaDominoTree masterRoot=parent;
    while (parent!=null)
    {
      masterRoot=parent;
      parents.add(parent);
      parent=parent.rootConnection_;
    };
    
    //now find and detach all bonus squares whose originating DominoTree is in
    //the collection of parent nodes
    detachAllBonusSquaresOriginatingIn(parents);
    
    FormulaDominoTree oldRootConnection=rootConnection_;
    rootConnection_=null;
    oldRootConnection.leafConnections_[rootConnectionLeafIndex_]=null;
    rootConnectionLeafIndex_=-1;
    
    //it would be great if we could just use FormulaSyntaxNet.extractByOwner() to
    //detach the tree. However (see doc for extractByOwner()) there are some
    //unifications that might remain. So we have no choice but to reunify the whole
    //tree.
    masterRoot.reunifyWholeTree(); //reunify parent part
    reunifyWholeTree();            //reunify detached part
       
    //because of the call to detachAllBonusSquaresOriginatingIn(parents) above
    //we have no choice but to invalidate all caches related to open leaves in 
    //the whole tree.
    masterRoot.invalidateOpenLeavesCachesWholeTreeDownwards(); //invalidate in parent part
    invalidateOpenLeavesCachesWholeTreeDownwards();   //invalidate in detached part
    
    //because leaf counts are wrong in parents of the detached tree we have to
    //invalidate related caches.
    oldRootConnection.invalidateLeavesCachesUpwards();
    
    return true;
  };
  
  
  /**
  * <p>Invalidates leaf count caches (but not open leaves counting caches)
  * of <code>this</code> and all parents.</p>
  */
  protected void invalidateLeavesCachesUpwards() //cache-clean, TESTED
  {
    FormulaDominoTree tree=this;
    do
    {
      tree.numOfLeavesCache_=-1;
      for (int i=0; i<tree.numOfLeavesCacheForConnection_.length; ++i)
      {
        tree.numOfLeavesCacheForConnection_[i]=-1;
      };
      tree=tree.rootConnection_;
    } 
    while(tree!=null);  
  };
  
  /**
  * <p>Invalidates open leaves counting caches
  * of <code>this</code> and all parents.</p>
  */
  protected void invalidateOpenLeavesCachesUpwards() //cache-clean, TESTED
  {
    FormulaDominoTree tree=this;
    do
    {
      tree.numOfOpenLeavesCache_=-1;
      for (int i=0; i<tree.numOfOpenLeavesCacheForConnection_.length; ++i)
      {
        tree.numOfOpenLeavesCacheForConnection_[i]=-1;
      };
      tree=tree.rootConnection_;
    } 
    while(tree!=null);  
  };
  
  
  /**
  * <p>Invalidates leaf count caches (but not open leaves counting caches) 
  * in <code>this</code> and all of
  * its subtrees (but not in parents
  * and siblings of <code>this</code>).</p>
  */
  protected void invalidateLeavesCachesWholeTreeDownwards() //cache-clean, TESTED
  {
    numOfLeavesCache_=-1;
    for (int i=0; i<numOfLeavesCacheForConnection_.length; ++i)
    {
      numOfLeavesCacheForConnection_[i]=-1;
    };
    
    for (int i=0; i<leafConnections_.length; ++i)
    {
      if (leafConnections_[i]!=null) leafConnections_[i].invalidateLeavesCachesWholeTreeDownwards();
    };
  };
  
  
  /**
  * <p>Invalidates caches related to open leaves in <code>this</code> and all of
  * its subtrees (but not in parents
  * and siblings of <code>this</code>).</p>
  */
  protected void invalidateOpenLeavesCachesWholeTreeDownwards() //cache-clean, TESTED
  {
    numOfOpenLeavesCache_=-1;
    for (int i=0; i<numOfOpenLeavesCacheForConnection_.length; ++i)
    {
      numOfOpenLeavesCacheForConnection_[i]=-1;
    };
    
    for (int i=0; i<leafConnections_.length; ++i)
    {
      if (leafConnections_[i]!=null) leafConnections_[i].invalidateOpenLeavesCachesWholeTreeDownwards();
    };
  };
  
  
  /**
  * <p>Invalidates caches related to open leaves in the whole tree (including parents
  * and siblings of <code>this</code>).</p>
  */
  protected void invalidateOpenLeavesCachesWholeTree() //cache-clean, TESTED
  {
    FormulaDominoTree tree=this;
    while(tree.rootConnection_!=null) tree=tree.rootConnection_;
    tree.invalidateOpenLeavesCachesWholeTreeDownwards();
  };
  
  /**
  * <p>Invalidates all caches in the whole tree (including parents
  * and siblings of <code>this</code>).</p>
  */
  protected void invalidateAllCachesWholeTree() //cache-clean, TESTED
  {
    FormulaDominoTree tree=this;
    while(tree.rootConnection_!=null) tree=tree.rootConnection_;
    
    //not very efficient, the following could be done in one merged function but
    //that would add maintenance overhead due to code duplication
    tree.invalidateOpenLeavesCachesWholeTreeDownwards();
    tree.invalidateLeavesCachesWholeTreeDownwards();
  };
  
  
  /**
  * <p>Connects the root of <code>this</code> to <code>leaf</code>
  * of <code>DominoTree</code> <code>to</code>.</p>
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;to.numberOfLeaves() && to.isOpenLeaf(leaf)</code>.</p>
  *
  * @throws IllegalArgumentException if <code>to</code> is a subtree of 
  *         <code>this</code>.
  *         The trees remain unchanged.
  * @throws IllegalNodeTypeException if <code>this</code> is already connected, or is
  *         a bonus square that has been detached or if the leaf to attach to is not
  *         open.
  *         The trees remain unchanged.
  * @throws SquareDoesNotExistException if <code>to==null</code> or 
  *         <code>leaf</code> is not a valid leaf 
  *         number. The trees remain unchanged.
  * @throws ConflictingBindException if the root square does not match the leaf to
  *         connect with. The trees remain unchanged.
  * @throws ClassCastException if <code>this</code> and <code>to</code>
  *         have incompatible types. The trees remain unchanged.
  *
  * @see #attachToSubtreeConnection(DominoTree, int)
  * @see #detachFromParent()
  * @see #numberOfLeaves()
  * @see #isOpenLeaf(int)
  */
  public void attachTo(DominoTree to, int leaf) //NOT cache-clean
  {
    if (to==null) throw new SquareDoesNotExistException();
    
    FormulaDominoTree fto=(FormulaDominoTree)to;
    LeafLocation l=fto.locateLeaf(leaf);
    
    if (l.index<0) throw new IllegalNodeTypeException();
    
    attachToSubtreeConnection(l.tree,l.index);
  };  
  
  
  /**
  * <p>Like {@link #attachTo(DominoTree,int)} with the difference that instead of a leaf
  * index, a subtree connection index is used.</p>
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;to.numberOfSubtreeConnections() && 
  * to.isOpenLeaf(leafIndexForSubtreeConnection(i))</code>.</p>
  *
  * @throws IllegalArgumentException if <code>to</code> is a subtree of 
  *         <code>this</code>.
  *         The trees remain unchanged.
  * @throws IllegalNodeTypeException if <code>this</code> is already connected, or is
  *         a bonus square that has been detached or if the leaf to attach to is not
  *         open.
  *         The trees remain unchanged.
  * @throws SquareDoesNotExistException if <code>to==null</code> or 
  *         <code>i</code> is not a valid subtree connection index. 
  *         The trees remain unchanged.
  * @throws ConflictingBindException if the root square does not match the leaf to
  *         connect with. The trees remain unchanged.
  * @throws ClassCastException if <code>this</code> and <code>to</code>
  *         have incompatible types. The trees remain unchanged.
  *
  * @see #attachTo(DominoTree, int)
  * @see #detachFromParent()
  * @see #numberOfLeaves()
  * @see #isOpenLeaf(int)
  */
  public void attachToSubtreeConnection(DominoTree to, int i) //NOT cache-clean
  {  
    if (to==null || i<0 || i>=to.numberOfSubtreeConnections()) throw new SquareDoesNotExistException();
    
    FormulaDominoTree tree=(FormulaDominoTree)to;
    int treeleaf=i;
    
    FormulaDominoTree masterTop=tree;
    while (masterTop.rootConnection_!=null) masterTop=masterTop.rootConnection_;
    if (masterTop==this) throw new IllegalArgumentException();
    
    if (rootConnection_!=null || !tree.isOpenConnection(treeleaf)) throw new IllegalNodeTypeException();
    
    //ATTENTION! The following may throw and we promise that the trees are unchanged
    //if an exception is thrown. So before this line there must not be any modifying
    //instructions.
    FormulaSyntaxNet.unify(root_, tree.leaves_[treeleaf]);
    
    //because attaching a new tree messes up the leaves count we have to invalidate
    //all related caches
    tree.invalidateLeavesCachesUpwards();
    tree.invalidateOpenLeavesCachesUpwards();
    
    tree.leafConnections_[treeleaf]=this;
    rootConnection_=tree;
    rootConnectionLeafIndex_=treeleaf;
  };
  
  
  /**
  * <p>Returns the number of subtrees that may be connected to <code>this</code>.</p>
  * <p>Every <code>DominoTree</code> consists of a root tree with connections
  * for subtrees. This function returns the number of connections regardless of 
  * whether actual subtrees are connected to them or not.</p>
  *
  * @see #subtree(int)
  */
  public int numberOfSubtreeConnections() //cache-clean, TESTED
  {
    return leafConnections_.length;
  };
  
  /**
  * <p>If a subtree is connected to subtree connection <code>i</code>, then that
  * subtree is returned. Otherwise this function returns <code>null</code>.</p>
  * 
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;to.numberOfSubtreeConnections()</code>.</p>
  *
  * @throws IllegalArgumentException if <code>i</code> has an illegal value.
  *
  * @see #numberOfSubtreeConnections()
  */
  public DominoTree subtree(int i) //cache-clean, TESTED
  {
    if (i<0 || i>=leafConnections_.length) throw new IllegalArgumentException();
    return leafConnections_[i];
  };
  
  
  /**
  * <p>Returns the number of leaves this tree has.</p>
  */
  public int numberOfLeaves() //NOT cache-clean, TESTED
  {
    if (numOfLeavesCache_<0)
    {
      numOfLeavesCache_=0;
      for (int i=0; i<leafConnections_.length; ++i)
        numOfLeavesCache_+=numOfLeavesForConnection(i);
    };
    
    return numOfLeavesCache_;
  };
  
  
  /**
  * <p>Returns the number of leaves this tree has that are not bonus squares.</p>
  */
  public int numberOfOpenLeaves() //NOT cache-clean, TESTED
  {
    if (numOfOpenLeavesCache_<0)
    {
      numOfOpenLeavesCache_=0;
      for (int i=0; i<leafConnections_.length; ++i)
        numOfOpenLeavesCache_+=numOfOpenLeavesForConnection(i);
    };
    
    return numOfOpenLeavesCache_;
  };
  
  
  /**
  * <p>Returns the number of new bonus {@link DominoSquare}s the player may use in a 
  * subtree attached to subtree connection <code>i</code>. This does not count bonus 
  * squares coming from parent nodes of <code>this</code>.
  * <p>A <code>FormulaDominoTree</code> may provide the same bonus square multiple
  * times and the same bonus square may be provided by <code>this</code> and/or
  * several parents at the same time, so the count returned by this function may
  * include squares that are also provided by parent trees.</p>
  *
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;numberOfSubtreeConnections()</code>.</p>
  *
  * @throws IllegalArgumentException if <code>i</code> is not a valid number.
  * 
  * @see #numberOfSubtreeConnections()
  * @see #subtree(int)
  */
  public int numberOfNewBonusSquaresForSubtreeConnection(int i) //TESTED
  {
    return bonus_.length;
  };
  
  
  /**
  * <p>Returns the number of bonus {@link DominoSquare}s the player may use in a 
  * subtree attached to <code>leaf</code>. This counts bonus squares coming from
  * parent nodes of <code>this</code> as well as new bonus squares not available in
  * parent nodes.</p>
  * <p>A <code>FormulaDominoTree</code> may provide the same bonus square multiple
  * times and the same bonus square may be provided by <code>this</code> and/or
  * several parents at the same time. These duplicates will be included in the count
  * returned by this function!</p>
  * <p>This method may be called even on non-open leaves although bonus squares
  * cannot be used on them. In this case the return value is the number of bonus
  * squares that could be used if the leaf were open.</p>
  *
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;numberOfLeaves()</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  */
  public int numberOfBonusSquares(int leaf) //NOT cache-clean, TESTED
  {
    LeafLocation l=locateLeaf(leaf);
    FormulaDominoTree tree=l.tree;
    int n=tree.bonus_.length;
    while (tree.rootConnection_!=null)
    {
      tree=tree.rootConnection_;
      n+=tree.bonus_.length;
    };
    return n;
  };
  
  
  /**
  * <p>Returns a copy of the {@link FormulaSyntaxNet} rooted at <code>n</code>
  * treating forward connections as non-existent.</p>
  * <p>When a
  * forward node is encountered it is copied as an unbound leaf and its forward
  * connections are not followed. Because of this
  * the returned {@link FormulaSyntaxNet} is always a proper tree with only branch
  * and leaf nodes.</p>
  *
  * @param copyowner the {@link FormulaSyntaxNet#owner() owner} of the returned 
  *        {@link FormulaSyntaxNet}.
  * @param n the node to start copying at.
  * @param m If you pass a {@link Map} that is empty, then
  *        after this function has finished
  *        <code>m</code> will map {@link FormulaSyntaxNet} nodes in the original net
  *        to their corresponding nodes in the copy.
  *        By passing the same {@link Map} (without clearing it) to later calls of
  *        this function you can carry over the mapping which is useful when
  *        copying several {@link FormulaSyntaxNet}s that share nodes.
  */
  protected static FormulaSyntaxNet noForwardCrossingCopy(Object copyowner, FormulaSyntaxNet n, Map m) //cache-clean, TESTED
  {
    FormulaSyntaxNet node=(FormulaSyntaxNet)m.get(n);
    if (node==null)
    {
      if (n.isBound())
      {
        node=FormulaSyntaxNet.newLeaf(copyowner, n.id());
      }
      else
      if (n.isUnboundOrForward())
      {
        node=FormulaSyntaxNet.newLeaf(copyowner, null);
      }
      else
      if (n.isBranch())
      {
        node=FormulaSyntaxNet.newBranch(copyowner,
                                           noForwardCrossingCopy(copyowner,n.left(),m),
                                           noForwardCrossingCopy(copyowner,n.right(),m)
                                       );
      };
      
      m.put(n,node);
    };
    return node;
  };
  
  /**
  * <p>Attaches to <code>leaf</code> the <code>num</code>-th bonus square that 
  * may be used in subtrees attached to <code>leaf</code>.</p>
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;numberOfLeaves()</code>.</p>
  * <p>Constraints for <code>num</code>: 
  * <code>0&lt;=num&lt;numberOfBonusSquares(leaf)</code>.</p>
  *
  * @return the {@link DominoTree} that was attached.
  *
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  * @throws ConflictingBindException if the bonus square and the leaf are connected
  *                                  in a way that would cause recursion or are
  *                                  otherwise incompatible.
  * @throws IllegalNodeTypeException <code>leaf</code> is not open.
  * @throws IllegalArgumentException if <code>num</code> has an illegal value.
  *
  * @see #rootSquare()
  * @see #leafSquare(int)
  * @see #bonusSquare(int,int)
  */
  public DominoTree useBonusSquare(int leaf, int num) //NOT cache-clean
  {
    int n=numberOfBonusSquares(leaf);
    if (num<0 || num>=n) throw new IllegalArgumentException();
    
    LeafLocation l=locateLeaf(leaf);
    
    if (l.index<0) throw new IllegalNodeTypeException();
    
    FormulaDominoTree bonustree=l.tree;
    while(true)
    {
      n-=bonustree.bonus_.length;
      if (num>=n) break;
      bonustree=bonustree.rootConnection_;
    };
        
    num-=n;
    
    //passing 0 for subtree because it doesn't matter as this class doesn't support
    //different bonuses for different subtrees
    return l.tree.useBonusSquare(l.index,bonustree,0,num);
  };
  

  /**
  * <p>Attaches to subtree connection <code>i</code> the bonus square returned by
  * <code>tree.newBonusSquareForSubtreeConnection(subtree,newBonusForSubtreeIndex)</code>.</p>
  * <p><code>tree</code> must be <code>this</code> or a direct or indirect parent of 
  * <code>this</code>.</p>
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;numberOfSubtreeConnections()</code>.</p>
  * <p>Constraints for <code>subtree</code>: 
  * <code>0&lt;=subtree&lt;tree.numberOfSubtreeConnections()</code>.</p>
  * <p>Constraints for <code>newBonusForSubtreeIndex</code>: 
  * <code>0&lt;=newBonusForSubtreeIndex&lt;tree.numberOfNewBonusSquaresForSubtreeConnection(subtree)</code>.</p>
  *
  * @return the {@link DominoTree} that was attached.
  *
  * @throws SquareDoesNotExistException if <code>i</code> is not a valid subtree
  *                                     connection index or 
  *                                     <code>newBonusForSubtreeIndex</code> 
  *                                     has an illegal value.
  * @throws ConflictingBindException if the bonus square and the leaf are connected
  *                                  in a way that would cause recursion or are
  *                                  otherwise incompatible, or if <code>tree</code>
  *                                  is not a direct or indirect parent of 
  *                                  <code>leaf</code>.
  * @throws IllegalNodeTypeException a subtree is connected to connection 
  *                                  <code>i</code> (i.e. it is not open).
  * @throws IllegalArgumentException if <code>subtree</code> is not a valid index.
  *
  * @see #rootSquare()
  * @see #leafSquare(int)
  * @see #bonusSquare(int,int)
  * @see #newBonusSquareForSubtreeConnection(int,int)
  */
  public DominoTree useBonusSquare(int i, DominoTree tree, int subtree, int newBonusForSubtreeIndex) 
  { //NOT cache-clean
   
    //subtree is checked for validity but it is not used as this class doesn't support
    //different bonuses for different subtrees
    if (subtree<0 || subtree>=tree.numberOfSubtreeConnections()) throw new IllegalArgumentException();
    if (newBonusForSubtreeIndex<0 || newBonusForSubtreeIndex>=tree.numberOfNewBonusSquaresForSubtreeConnection(subtree)) throw new SquareDoesNotExistException();
    
    if (i<0 || i>=numberOfSubtreeConnections()) throw new SquareDoesNotExistException();
    
    if (!isOpenConnection(i)) throw new IllegalNodeTypeException();
   
    FormulaDominoTree bonustree=this;
    while (bonustree!=tree && bonustree.rootConnection_!=null) 
      bonustree=bonustree.rootConnection_;
    if (bonustree!=tree) throw new ConflictingBindException();
    
    FormulaSyntaxNet bsNet=bonustree.bonus_[newBonusForSubtreeIndex];
    
    BonusFormulaDominoTree bsTree=new BonusFormulaDominoTree();
    FormulaSyntaxNet newBsNet=noForwardCrossingCopy(bsTree,bsNet,new HashMap());
    
    try{
      //because bonus squares may be wholly or partially unbound, we have to
      //unify with the original so that the new one keeps in sync, rather than
      //leaving or newBsNet independent.
      FormulaSyntaxNet.unify(bsNet,newBsNet);
    }catch(Exception x)
    {
      throw new RuntimeException("Unexpected exception: "+x.toString());
    };  
    
    bsTree.init(bonustree,newBonusForSubtreeIndex,newBsNet);
    
    try{
      FormulaSyntaxNet.unify(newBsNet, this.leaves_[i]);
    }catch(RuntimeException x)
    {
      HashSet owners=new HashSet();
      owners.add(bsTree);
      //using extractByOwner is safe here because the above unify() does not 
      //leave traces if it fails so that the only thing we undo here is the earlier
      //unify of the newBsNet with bsNet. And that unify() couldn't add connections
      //that this extractByOwner would not remove.
      FormulaSyntaxNet.extractByOwner(newBsNet,owners);
      throw x;
    };  
    
    //ATTENTION! If an exception-throwing command is added after the following line,
    //the exception will have to be caught so that bsTree can be removed again
    //from bonusSquaresInAction_ before propagating the exception.
    bonustree.bonusSquaresInAction_[newBonusForSubtreeIndex].add(bsTree);
    
    this.leafConnections_[i]=bsTree;
    bsTree.rootConnection_=this;
    bsTree.rootConnectionLeafIndex_=i;
    
    //because attaching bonus square messes up open leaves counts, we have to
    //invalidate the caches.
    this.invalidateOpenLeavesCachesUpwards();
    
    return bsTree;
  };


  
  /**
  * <p>Creates and returns an incomplete copy of <code>this</code> and its subtrees.</p>
  * <p>The {@link FormulaSyntaxNet}s of the resulting tree will not be unified and the
  * caches are uninitialized (the cache arrays have the proper size but undefined
  * values).</p>
  * <p><b>This function does not work on {@link BonusFormulaDominoTree}s.</b></p>
  *
  * @return the clone.
  *
  * @param cloneTreeRoot the <code>rootConnection_</code> the copy should get.
  * @param cloneTreeRootConnectionLeafIndex the rootConnectionLeafIndex_ the copy
  *        should get.
  * @param mapBonusSquareToClone used internally. Pass <code>null</code>.
  * @param wcmap When <code>wcmap.orginal</code> is cloned, a reference to the clone
  *        is stored in <code>wcmap.clone</code>.
  */
  protected FormulaDominoTree recursiveNonUnifiedClone(FormulaDominoTree cloneTreeRootConnection, int cloneTreeRootConnectionLeafIndex,Map mapBonusSquareToClone, WantedCloneMap wcmap)
  {                                                    //cache-clean, TESTED
    FormulaDominoTree cloneTree=new FormulaDominoTree();
    
    if (wcmap.original==this) wcmap.clone=cloneTree;
    
    Map copyMap=new HashMap();
    cloneTree.root_=noForwardCrossingCopy(cloneTree,root_,copyMap);
    cloneTree.leaves_=new FormulaSyntaxNet[leaves_.length];
    for (int i=0; i<leaves_.length; ++i)
    {
      cloneTree.leaves_[i]=noForwardCrossingCopy(cloneTree,leaves_[i],copyMap);
    };
    
    cloneTree.bonus_=new FormulaSyntaxNet[bonus_.length];
    for (int i=0; i<bonus_.length; ++i)
    {
      cloneTree.bonus_[i]=noForwardCrossingCopy(cloneTree,bonus_[i],copyMap);
    };
    
    //mapBonusSquareToClone is used to map a bonus square from the original tree to
    //its copy. This is necessary because we copy bonus squares when we encounter
    //the originating FormulaDominoTree and we need to get a reference to the copy
    //later when we're copying a leaf.
    if (mapBonusSquareToClone==null) mapBonusSquareToClone=new HashMap();
    
    cloneTree.bonusSquaresInAction_=new Collection[bonusSquaresInAction_.length];
    for (int i=0; i<bonusSquaresInAction_.length; ++i)
    {
      Collection cloneColl=newCollectionUsedForBonusSquaresInAction();
      cloneTree.bonusSquaresInAction_[i]=cloneColl;

      Iterator iter=bonusSquaresInAction_[i].iterator();
      while (iter.hasNext())
      {
        BonusFormulaDominoTree bsTree=(BonusFormulaDominoTree)iter.next();
        BonusFormulaDominoTree bsTreeClone=new BonusFormulaDominoTree();
        cloneColl.add(bsTreeClone);

        //do not use copyMap in the following call to noForwardCrossingCopy()
        //because we want an independent copy
        bsTreeClone.init(cloneTree,i,noForwardCrossingCopy(bsTreeClone,cloneTree.bonus_[i],new HashMap()));
        mapBonusSquareToClone.put(bsTree,bsTreeClone);
      };
    };
    
    cloneTree.numOfOpenLeavesCacheForConnection_=new int[numOfOpenLeavesCacheForConnection_.length];
    cloneTree.numOfLeavesCacheForConnection_=new int[numOfLeavesCacheForConnection_.length];
    
    cloneTree.rootConnection_=cloneTreeRootConnection;
    cloneTree.rootConnectionLeafIndex_=cloneTreeRootConnectionLeafIndex;
    
    cloneTree.leafConnections_=new FormulaDominoTree[leafConnections_.length];
    for (int i=0; i<leafConnections_.length; ++i)
    {
      FormulaDominoTree leafCon=leafConnections_[i];
      if (leafCon==null) continue;
      if (leafCon.leaves_.length==0) //if bonus square
      {
        FormulaDominoTree leafConClone=(FormulaDominoTree)mapBonusSquareToClone.get(leafCon);
        if (Debugging.enabled)
        {
          if (leafConClone==null) throw new RuntimeException("Huh? This is not possible.");
        };
        if (wcmap.original==leafCon) wcmap.clone=leafConClone;
        leafConClone.rootConnection_=cloneTree;
        leafConClone.rootConnectionLeafIndex_=i;
        cloneTree.leafConnections_[i]=leafConClone;
      }
      else cloneTree.leafConnections_[i]=leafCon.recursiveNonUnifiedClone(cloneTree,i,mapBonusSquareToClone,wcmap);
    };
    
    return cloneTree;
  };
  
  
  protected static class WantedCloneMap
  {
    public FormulaDominoTree original;
    public FormulaDominoTree clone=null;
    public WantedCloneMap(FormulaDominoTree orig) {original=orig;};
  };
  
  /**
  * <p>Returns a copy of the <code>DominoTree</code>. The copy will have the same
  * structure as the original but will not share any physical data, nor will it
  * be connected with the original in any way. This means that the clone can be
  * used independent of the orginal as if the clone had been synthesized directly
  * with no involvement of the original tree.</p>
  * <p>Note that if this function is called on a subtree of a larger tree, the
  * whole tree will be cloned and the return value of this function will be the
  * subtree of the large clone tree that corresponds to the tree being cloned.</p>
  */
  public DominoTree cloneDT()  //cache-clean, TESTED
  {
    //find absolute root of tree
    FormulaDominoTree tree=this;
    while (tree.rootConnection_!=null) tree=tree.rootConnection_;
    
    WantedCloneMap wcmap=new WantedCloneMap(this);
    
    FormulaDominoTree cloneTree=tree.recursiveNonUnifiedClone(null,-1,null,wcmap);
    cloneTree.recreateAllForwardsDownwards();
    cloneTree.invalidateAllCachesWholeTree();
    
    if (Debugging.enabled)
    {
      if (wcmap.clone==null) throw new RuntimeException("Failed to clone desired tree!");
    };
    
    return wcmap.clone;
  };
  
  /**
  * <p>See {@link #cloneDT()}.</p>
  */ 
  public Object clone(){ return cloneDT();}; //cache-clean, TESTED
  
  /**
  * <p>Returns the root square.</p>
  * @see #leafSquare(int)
  * @see #bonusSquare(int,int)
  */
  public DominoSquare rootSquare() //cache-clean, TESTED
  {
    return new RootFormulaDominoSquare();
  };
  
  /**
  * <p>Returns the square for <code>leaf</code>.</p>
  *
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;numberOfLeaves()</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  *
  * @see #rootSquare()
  * @see #bonusSquare(int,int)
  */
  public DominoSquare leafSquare(int leaf) //NOT cache-clean, TESTED
  {
    LeafLocation l=locateLeaf(leaf);
    if (l.index<0) return l.tree.rootSquare();
    return l.tree.subtreeConnectionSquare(l.index);
  };
  
  /**
  * <p>Returns the {@link DominoSquare} for subtree connection <code>i</code>.</p>
  * <p>This method works even for subtree connections with no attached subtree.</p>
  *
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;numberOfSubtreeConnections()</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>i</code> is not a valid
  *         subtree connection index.
  */
  public DominoSquare subtreeConnectionSquare(int i) //TESTED
  {
    if (i<0 || i>=numberOfSubtreeConnections()) throw new SquareDoesNotExistException();
    return new LeafFormulaDominoSquare(i);
  };
  
  /**
  * <p>Returns <code>num</code>-th bonus square usable for <code>leaf</code>.</p>
  *
  * <p>Constraints for <code>leaf</code>: 
  * <code>0&lt;=leaf&lt;numberOfLeaves()</code>.</p>
  * <p>Constraints for <code>num</code>: 
  * <code>0&lt;=num&lt;numberOfBonusSquares(leaf)</code>.</p>
  *
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  * @throws IllegalArgumentException if <code>num</code> has an illegal value;
  *
  * @see #rootSquare()
  * @see #leafSquare(int) 
  */
  public DominoSquare bonusSquare(int leaf, int num) //NOT cache-clean, TESTED
  {
    int n=numberOfBonusSquares(leaf);
    if (num<0 || num>=n) throw new IllegalArgumentException();
    
    LeafLocation l=locateLeaf(leaf);
    FormulaDominoTree bonustree=l.tree;
    while(true)
    {
      n-=bonustree.bonus_.length;
      if (num>=n) break;
      bonustree=bonustree.rootConnection_;
    };
        
    num-=n;
    return bonustree.bonusFormulaDominoSquare(num);
  };
  
  
  /**
  * <p>Returns <code>num</code>-th new bonus square usable for subtree connection
  * <code>i</code>.</p>
  *
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;numberOfSubtreeConnections()</code>.</p>
  * <p>Constraints for <code>num</code>: 
  * <code>0&lt;=num&lt;numberOfNewBonusSquaresForSubtreeConnection(i)</code>.</p>
  *
  * @throws IllegalArgumentException if <code>i</code> has an illegal value;
  * @throws SquareDoesNotExistException if <code>num</code> is not a valid number
  *         for subtree connection <code>i</code>.
  *
  * @see #rootSquare()
  * @see #leafSquare(int) 
  * @see #bonusSquare(int, int)
  * @see #subtreeConnectionSquare(int)
  * @see #numberOfSubtreeConnections()
  * @see #numberOfNewBonusSquaresForSubtreeConnection(int)
  */
  public DominoSquare newBonusSquareForSubtreeConnection(int i, int num) //TESTED
  {
    if (i<0 || i>=numberOfSubtreeConnections()) throw new IllegalArgumentException();
    if (num<0 || num>=numberOfNewBonusSquaresForSubtreeConnection(i)) throw new SquareDoesNotExistException();
    return bonusFormulaDominoSquare(num);
  };
  
  /**
  * <p>Returns bonus square for <code>bonus_[bonusnum]</code>.</p>
  */
  protected DominoSquare bonusFormulaDominoSquare(int bonusnum) //cache-clean, TESTED
  {
    if (Debugging.enabled)
    {
      if (bonusnum<0 || bonusnum>=bonus_.length) throw new RuntimeException();
    };
    return new BonusFormulaDominoSquare(bonusnum);
  };  
  
  /**
  * <p>Creates and returns a new <code>FormulaDominoTree</code> based on
  * infix formula {@link String}s.</p>
  * <p>A formula may contain "-&gt;", bound and unbound propositional variables.
  * Unbound propositional variables have the form X&lt;num&gt;, where &lt;num&gt;
  * is any non-negative integer. Bound propositional variables can be anything
  * (that doesn't include "(", ")" or "-&gt;").</p>
  * <p>No associativity is assumed for -&gt;, so A-&gt;B-&gt;C must be
  * written as either (A-&gt;B)-&gt;C or A-&gt;(B-&gt;C).</p>
  * <p>Note that there are no predefined symbols for false and true. Use any
  * bound propositional variable(s) you like.</p>
  * <p>If the same identifier for a propositional variable is used in more than
  * one place (in the same formula or in different formulas), it will represent the 
  * same variable. In the case of an unbound
  * propositional variable, any binding of one occurence will automatically bind
  * all other occurences. This works across all formulas passed in a single
  * invocation of this method but not across different invocations.<br>
  * Example (pseudo-code):
  * <pre>
  *    a=newFromInfix("X0",{"X0"},{});
  *    b=newFromInfix("X0",{"X0"},{});
  * </pre>
  * The 2 {@link DominoSquare}s of tree <code>a</code> will always be identical. 
  * If one is
  * bound, the other one will be bound with it. The same is true for tree 
  * <code>b</code>. However, there is no connection between tree <code>a</code> and
  * tree <code>b</code>. Binding a square in one of the 2 trees will not 
  * automatically bind a square in the other tree.</p>
  *
  * @param root the formula for the root {@link DominoSquare}.
  * @param leaves array of {@link String}s that may not be null or empty 
  *        and must not contain 
  *        <code>null</code>
  *        elements. Every element will add a corresponding 
  *        {@link DominoSquare} to the result tree as a leaf.
  * @param bonus array of {@link String}s that may be empty or <code>null</code>
  *        but must not contain
  *        <code>null</code> elements. 
  *        Every element will add a corresponding bonus square
  *        that may be {@link DominoTree#useBonusSquare(int,int) used}
  *        in subtrees attached to any leaves of the result tree.
  *
  * @throws IllegalArgumentException if <code>leaves</code> is empty or 
  *         <code>null</code> or 
  *         an element of any of the arrays is <code>null</code>
  *         or there is a syntax error in any of the strings.
  */
  public static FormulaDominoTree newFromInfix(String root, String[] leaves, String[] bonus)
  {                                                                //cache-clean, TESTED
    //having no leaves characterizes a bonus square and this test is used
    //in several places, so we can not allow it for regular FormulaDominoTrees
    if (leaves==null || leaves.length==0) throw new IllegalArgumentException();
    
    int blen=0;
    if (bonus!=null) blen=bonus.length;
    String[] formulas=new String[1+leaves.length+blen];
    formulas[0]=root;
    System.arraycopy(leaves,0,formulas,1,leaves.length);
    if (blen>0) System.arraycopy(bonus,0,formulas,1+leaves.length,blen);
    
    FormulaDominoTree tree=new FormulaDominoTree();
    
    FormulaSyntaxNet[] fsns=FormulaSyntaxNet.newFromInfixFormulaStrings(tree,formulas);
    FormulaSyntaxNet rootFSN=fsns[0];
    FormulaSyntaxNet[] leavesFSN=new FormulaSyntaxNet[leaves.length];
    System.arraycopy(fsns,1,leavesFSN,0,leaves.length);
    FormulaSyntaxNet[] bonusFSN=new FormulaSyntaxNet[blen];
    if (blen>0) System.arraycopy(fsns,1+leavesFSN.length,bonusFSN,0,blen);
    
    tree.init(rootFSN,leavesFSN,bonusFSN);
    
    return tree;
  };
  
  /**
  * <p>Creates a new <code>FormulaDominoTree</code> initialized with the respective
  * {@link FormulaSyntaxNet}s.</p>
  * <p>For internal use only. None of the arrays may be <code>null</code> and none
  * of the elements may be <code>null</code>.</p>
  */ 
  protected FormulaDominoTree(FormulaSyntaxNet root, FormulaSyntaxNet[] leaves, FormulaSyntaxNet[] bonus)
  {					//cache-clean, TESTED
    init(root, leaves, bonus);
  };  
  
  
  /**
  * <p>Initializes <code>this</code> with the respective
  * {@link FormulaSyntaxNet}s.</p>
  * <p>For internal use only. None of the arrays may be <code>null</code> and none
  * of the elements may be <code>null</code>.</p>
  * <p>This function may only be called on a <code>FormulaDominoTree</code> that
  * is not yet initialized.</p>
  */   
  protected void init(FormulaSyntaxNet root, FormulaSyntaxNet[] leaves, FormulaSyntaxNet[] bonus)
  { 				//cache-clean, TESTED
    root_=root;
    leaves_=leaves;
    bonus_=bonus;
    leafConnections_=new FormulaDominoTree[leaves_.length];
    bonusSquaresInAction_=new Collection[bonus_.length];
    numOfLeavesCacheForConnection_=new int[leaves_.length];
    numOfOpenLeavesCacheForConnection_=new int[leaves_.length];
    
    for (int i=0; i<bonusSquaresInAction_.length; ++i)
    {
      bonusSquaresInAction_[i]=newCollectionUsedForBonusSquaresInAction();
    };
    invalidateAllCachesWholeTree();
  };
  
  
  /**
  * <p>Creates an uninitialized <code>FormulaDominoTree</code>.
  * For internal use only.</p>
  */ 
  protected FormulaDominoTree() //TESTED
  {};
  
  /**
  * <p>Returns <code>leaf</code>'s {@link FormulaSyntaxNet}.</p>
  * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf number.
  */
  protected FormulaSyntaxNet leafNet(int leaf) //NOT cache-clean, TESTED
  {
    LeafLocation l=locateLeaf(leaf);
    if (l.index<0) return l.tree.root_; else return l.tree.leaves_[l.index];
  };

  /**
  * <p>Returns all the unbound nodes that occur in {@link FormulaSyntaxNet}s of 
  * leaves as a {@link Set} of {@link FormulaSyntaxNet}s.</p>
  */
  public Set unboundVariablesInLeaves() //NOT cache-clean, TESTED
  {
    HashSet unbound=new HashSet();
    for (int leaf=0; leaf<numberOfLeaves(); ++leaf)
    {
      FormulaSyntaxNet n=leafNet(leaf);
      FSNIterator iter=n.multiorderIterator(false,false,false);
      while (iter.hasNext()) 
      {
        FormulaSyntaxNet node=iter.nextFSN();
        if (node.isUnboundOrForward()) unbound.add(node);
      };  
    };
    return unbound;
  };
  
  /**
  * <p>Returns all the unbound nodes that occur in the root 
  * {@link FormulaSyntaxNet}.</p>
  */
  public Set unboundVariablesInRoot() //cache-clean, TESTED
  {
    HashSet unbound=new HashSet();
    FSNIterator iter=root_.multiorderIterator(false,false,false);
    while (iter.hasNext()) 
    {
      FormulaSyntaxNet node=iter.nextFSN();
      if (node.isUnboundOrForward()) unbound.add(node);
    };  
    return unbound;
  };
  
  /**
  * <p>Returns the number of unbound variables that occur in formulas in leaves.</p>
  */
  public int numberOfUnboundVariablesInLeaves() //NOT cache-clean, TESTED
  {
    return unboundVariablesInLeaves().size();
  };
  
  /**
  * <p>Returns the number of unbound variables that occur in the root formula.</p>
  */
  public int numberOfUnboundVariablesInRoot() //cache-clean, TESTED
  {
    return unboundVariablesInRoot().size();
  };
  
  /**
  * <p>Binds all unbound variables that occur in formulas in leaves to
  * identifiers from <code>ids</code>. The order in which unbound variables are
  * bound is unspecified. The owner of every new bound node will be the same as that of
  * the unbound node it is unified with.</p>
  *
  * @throws IllegalArgumentException if <code>ids==null</code> or
  *         <code>ids.length&lt;numberOfUnboundVariablesInLeaves()</code>. If
  *         this exception is thrown some variables may already be bound and will
  *         remain so.
  *
  */
  public void bindUnboundVariablesInLeaves(String[] ids) //NOT cache-clean, TESTED
  {
    bindUnbound(unboundVariablesInLeaves(),ids);
  };
  
  
  /**
  * <p>Binds all unbound variables that occur in the root formula to
  * identifiers from <code>ids</code>. The order in which unbound variables are
  * bound is unspecified. The owner of every new bound node will be the same as that of
  * the unbound node it is unified with.</p>
  *
  * @throws IllegalArgumentException if <code>ids==null</code> or
  *         <code>ids.length&lt;numberOfUnboundVariablesInRoot()</code>. If
  *         this exception is thrown some variables may already be bound and will
  *         remain so.
  *
  */
  public void bindUnboundVariablesInRoot(String[] ids) //NOT cache-clean, TESTED
  {
    bindUnbound(unboundVariablesInRoot(),ids);
  };  
    
  
  /**
  * <p>Unifies all {@link FormulaSyntaxNet}s in <code>unbound</code> with new 
  * leaf nodes with ids from
  * <code>ids</code>. The order in which they are unified is unspecified.</p>
  * <p>All elements of <code>unbound</code> must be unbound or forward nodes.
  * The owner of every new bound node will be the same as that of
  * the unbound node it is unified with.</p>
  *
  * @throws IllegalArgumentException if <code>ids==null</code> or
  *         <code>ids.length&lt;unbound.size()</code>. If
  *         this exception is thrown some variables may already be bound and will
  *         remain so.
  *
  */
  protected static void bindUnbound(Set unbound, String[] ids) //cache-clean, TESTED
  {
    if (ids==null || ids.length<unbound.size()) throw new IllegalArgumentException();
    
    int i=0;
    Iterator iter=unbound.iterator();
    while(iter.hasNext())
    {
      FormulaSyntaxNet unode=(FormulaSyntaxNet)iter.next();
      FormulaSyntaxNet.unify(unode,FormulaSyntaxNet.newLeaf(unode.owner(),ids[i++]));
    };
  };
 
  /**
  * <p>Tests the functions of FormulaDominoTree.</p>
  */
  public static void main(String args[]) //TESTED
  {
    if (Debugging.enabled)
    {
      java.util.Random rand0=new java.util.Random(0);
      domino.logic.TautologyGenerator tgen=new domino.logic.TautologyGenerator(rand0);
      Vector prob=new Vector();
      prob.add(new Double(1));
      prob.add(new Double(2));
      prob.add(new Double(3));
      prob.add(new Double(4));
      FormulaDominoTree randomTautProof=tgen.randomTautologyProof(2,prob,true);
      FormulaDominoTree toBeDetachedParent=(FormulaDominoTree)randomTautProof.subtree(0).subtree(0);
      FormulaDominoTree toBeDetached=(FormulaDominoTree)toBeDetachedParent.subtree(0);
      System.out.println("randomTautProof.numberOfLeaves()="+randomTautProof.numberOfLeaves());
      System.out.println("randomTautProof.numberOfOpenLeaves()="+randomTautProof.numberOfOpenLeaves());
      System.out.println("toBeDetached.numberOfLeaves()="+toBeDetached.numberOfLeaves());
      System.out.println("toBeDetached.numberOfOpenLeaves()="+toBeDetached.numberOfOpenLeaves());
      System.out.println("toBeDetached.detachFromParent():");
      toBeDetached.detachFromParent();
      System.out.println("randomTautProof.isOpenLeaf(0)="+randomTautProof.isOpenLeaf(0));
      System.out.println("toBeDetached.isOpenLeaf(0)="+toBeDetached.isOpenLeaf(0));
      System.out.println("randomTautProof.numberOfLeaves()="+randomTautProof.numberOfLeaves());
      System.out.println("randomTautProof.numberOfOpenLeaves()="+randomTautProof.numberOfOpenLeaves());
      System.out.println("toBeDetached.numberOfLeaves()="+toBeDetached.numberOfLeaves());
      System.out.println("toBeDetached.numberOfOpenLeaves()="+toBeDetached.numberOfOpenLeaves());
      System.out.println("toBeDetached.attachTo(...):");
      toBeDetached.attachTo(toBeDetachedParent,0);
      System.out.println("randomTautProof.numberOfLeaves()="+randomTautProof.numberOfLeaves());
      System.out.println("randomTautProof.numberOfOpenLeaves()="+randomTautProof.numberOfOpenLeaves());
      System.out.println("toBeDetached.numberOfLeaves()="+toBeDetached.numberOfLeaves());
      System.out.println("toBeDetached.numberOfOpenLeaves()="+toBeDetached.numberOfOpenLeaves());
      
      System.out.println("toBeDetached.root_="+toBeDetached.root_.toInfixFormulaString());
      System.out.println("toBeDetached.leaves_[0]="+toBeDetached.leaves_[0].toInfixFormulaString());
      
      HashSet allReach=new HashSet();
      toBeDetached.root_.getAllReachableNodes(allReach);
      System.out.println("toBeDetached.root_.getAllReachableNodes().size()="+allReach.size());
      
      System.out.println("cloneTree=toBeDetached.cloneDT():");
      FormulaDominoTree cloneTree=(FormulaDominoTree)toBeDetached.cloneDT();
      System.out.println("cloneTree.root_="+cloneTree.root_.toInfixFormulaString());
      System.out.println("cloneTree.leaves_[0]="+cloneTree.leaves_[0].toInfixFormulaString());
      
      toBeDetached.root_.getAllReachableNodes(allReach);
      System.out.println("toBeDetached.root_.getAllReachableNodes().size()="+allReach.size());
      cloneTree.root_.getAllReachableNodes(allReach);
      System.out.println("cloneTree.root_.getAllReachableNodes().size()="+allReach.size());
      
      System.out.println("cloneTree.numberOfLeaves()="+cloneTree.numberOfLeaves());
      System.out.println("cloneTree.numberOfOpenLeaves()="+cloneTree.numberOfOpenLeaves());
      
      FormulaDominoTree t=(FormulaDominoTree)randomTautProof;
      while (t!=null && t.numberOfSubtreeConnections()!=0) t=(FormulaDominoTree)t.subtree(0);
      if (t!=null) t.detachFromParent();
      t.root_.getAllReachableNodes(allReach);
      System.out.println("t.root_.getAllReachableNodes().size()="+allReach.size());
    };
  };
  
  /**
  * <p>Parent class of {@link FormulaDominoSquare}s.</p>
  */
  protected abstract class AbstractFormulaDominoSquare implements FormulaDominoSquare
  {
    /** <p>Returns this square's {@link FormulaSyntaxNet}.</p>*/
    protected abstract FormulaSyntaxNet myNet();
    public Object toDefaultRepresentation()
    { return toInfixRepresentation(null); };

    /**
    * <p>Returns an infix formula string for the square's formula.</p>
    */
    public String toInfixRepresentation() //TESTED
    {
      return toInfixRepresentation(null);
    };

    /**
    * <p>Returns an infix formula string for the square's formula.</p>
    * @param assign If this is <code>null</code>, unbound nodes will be numbered
    *        starting from 0. Otherwise, the assigner will be used to determine the
    *        numbers. The assigner stores the assignments so that by reusing the
    *        same assigner in different calls to this function you can achieve
    *        a consistent numbering of unbound nodes.
    */
    public String toInfixRepresentation(UnboundIdAssigner assign) //TESTED
    {
      StringBuffer s=new StringBuffer();
      FSNIterator i=myNet().multiorderIterator(true,true,true);
      java.util.Map map=null;
      if (assign==null) map=new java.util.HashMap(); //maps FormulaSyntaxNets to String names
      int unboundCount=0; //number of unbound leaves encountered so far
      while (i.hasNext())
      {
        FormulaSyntaxNet node=i.nextFSN();
        if (node.isBranch()) 
        {
          if (i.preorder()) s.append('(');
          if (i.inorder()) s.append("->");
          if (i.postorder()) s.append(')');
        }  
        else if (node.isBound()) s.append(node.id());
        else if (node.isUnboundOrForward())
        {
          String id;
          if (assign!=null) 
            id="X"+((UnboundIdAssignerImpl)assign).getIdFor(node);
          else
          {  
            id=(String)map.get(node);
            if (id==null)
            {
              id="X"+(unboundCount++);
              map.put(node,id);
            };
          };
          s.append(id);
        }
        else throw new RuntimeException("Huh? What's this node?"); //just to be safe
      };
      return s.toString();
    };
    
    /**
    * <p>Returns the prefix representation of the <code>FormulaDominoSquare</code>.</p>
    */
    public String toPrefixRepresentation()
    {
      return toPrefixRepresentation(null);
    };

    /**
    * <p>Returns a prefix formula string for the square's formula.</p>
    * @param assign If this is <code>null</code>, unbound nodes will be numbered
    *        starting from 0. Otherwise, the assigner will be used to determine the
    *        numbers. The assigner stores the assignments so that by reusing the
    *        same assigner in different calls to this function you can achieve
    *        a consistent numbering of unbound nodes.
    */
    public String toPrefixRepresentation(FormulaDominoTree.UnboundIdAssigner assign)
    { //TESTED
      return toPrePostfixRepresentation(assign,myNet().preorderIterator());
    };
    
    /**
    * <p>Returns the postfix representation of the <code>FormulaDominoSquare</code>.</p>
    */
    public String toPostfixRepresentation()
    {
      return toPostfixRepresentation(null);
    };

    /**
    * <p>Returns a postfix formula string for the square's formula.</p>
    * @param assign If this is <code>null</code>, unbound nodes will be numbered
    *        starting from 0. Otherwise, the assigner will be used to determine the
    *        numbers. The assigner stores the assignments so that by reusing the
    *        same assigner in different calls to this function you can achieve
    *        a consistent numbering of unbound nodes.
    */
    public String toPostfixRepresentation(FormulaDominoTree.UnboundIdAssigner assign)
    { //TESTED
      return toPrePostfixRepresentation(assign,myNet().postorderIterator());
    };  
    
    /**
    * <p>If <code>i</code> is a preorder iterator, this function performs as
    * {@link #toPrefixRepresentation(FormulaDominoTree.UnboundIdAssigner)} and if
    * <code>i</code> is a postorder iterator, it performs as
    * {@link #toPostfixRepresentation(FormulaDominoTree.UnboundIdAssigner)}.</p>
    */  
    protected String toPrePostfixRepresentation(FormulaDominoTree.UnboundIdAssigner assign, FSNIterator i)
    {
      StringBuffer s=new StringBuffer();
      java.util.Map map=null;
      if (assign==null) map=new java.util.HashMap(); //maps FormulaSyntaxNets to String names
      int unboundCount=0; //number of unbound leaves encountered so far
      while (i.hasNext())
      {
        FormulaSyntaxNet node=i.nextFSN();
        if (node.isBranch()) 
          s.append("->");
        else if (node.isBound()) 
          s.append("("+node.id()+")");
        else if (node.isUnboundOrForward())
        {
          String id;
          if (assign!=null) 
            id="X"+((UnboundIdAssignerImpl)assign).getIdFor(node);
          else
          {  
            id=(String)map.get(node);
            if (id==null)
            {
              id="X"+(unboundCount++);
              map.put(node,id);
            };
          };
          s.append("("+id+")");
        }
        else throw new RuntimeException("Huh? What's this node?"); //just to be safe
      };
      return s.toString();
    };
  };
  
  
  protected class RootFormulaDominoSquare extends AbstractFormulaDominoSquare
  {
    public RootFormulaDominoSquare() {}; //TESTED
    protected FormulaSyntaxNet myNet() {return root_;} //TESTED
  };
  
  protected class LeafFormulaDominoSquare extends AbstractFormulaDominoSquare
  {
    private int leafindex_;
    public LeafFormulaDominoSquare(int leafindex) {leafindex_=leafindex;}; //TESTED
    protected FormulaSyntaxNet myNet() {return leaves_[leafindex_];}; //TESTED
  };
  
  protected class BonusFormulaDominoSquare extends AbstractFormulaDominoSquare
  {
    private int bonusindex_;
    public BonusFormulaDominoSquare(int bonusindex) {bonusindex_=bonusindex;}; //TESTED
    protected FormulaSyntaxNet myNet() {return bonus_[bonusindex_];}; //TESTED
  };
  
  /**
  * <p>Used to assign numbers to unbound leaves of formulas.
  * <code>UnboundIdAssigner</code>s are used to make sure ids assigned to
  * unbound leaves are unique in the whole tree and the same ids are used for the
  * same nodes for independent calls to functions that generate a representation 
  * of the whole tree or a part of it for displaying.</p>
  */
  public static class UnboundIdAssigner
  {
    /**
    * <p>Returns a new <code>UnboundIdAssigner</code>.</p>
    */
    public static UnboundIdAssigner newAssigner() //TESTED
    {
      return new UnboundIdAssignerImpl();
    };
    
    /**
    * <p>Clears all mappings.</p>
    */
    public void reset(){}; //TESTED
    
    /**
    * <p>Removes mappings for the unbound nodes in formulas in <code>tree</code>.</p>
    * <p>If <code>tree</code> is attached to another tree, mappings in that tree
    * that affect <code>tree</code> through unification, will also be removed.
    * If this is not desirable, {@link #removeMappingsStrictlyFor(FormulaDominoTree)}
    * should be used instead.</p>
    */
    public void removeMappingsFor(FormulaDominoTree tree){}; //TESTED
    
    /**
    * <p>Removes mappings for the unbound nodes in formulas in <code>tree</code> if
    * they are unneeded because other mappings or unifications with bound or
    * branch nodes supercede them.</p>
    * <p>If <code>tree</code> is attached to another tree, mappings in that tree
    * that affect <code>tree</code> through unification, will also be removed.</p>
    */
    public void removeUnneededMappingsFor(FormulaDominoTree tree){};
    
    /**
    * <p>Removes mappings for the unbound nodes in formulas in <code>tree</code>.</p>
    * <p>Only mappings for nodes connected through branch connections will be affected.
    * Mappings for nodes connected through forward connections, even if they affect 
    * <code>tree</code>, will not be touched. 
    * If these mappings should be removed, too, 
    * {@link #removeMappingsFor(FormulaDominoTree)} should be used instead.</p>
    */
    public void removeMappingsStrictlyFor(FormulaDominoTree tree){};
    
    /**
    * <p><code>UnboundIdAssigner</code> may not be instantiated directly.</p>
    */
    protected UnboundIdAssigner(){}; //TESTED
    //
    // WHY DID I NOT MAKE THIS CLASS ABSTRACT ????  I AM SURE I HAD A REASON.
    //
    
  };
  
  /**
  * <p>Implements {@link FormulaDominoTree.UnboundIdAssigner}. This is a separate class in order to
  * hide the {@link #getIdFor(FormulaSyntaxNet)} function from the user, so that
  * s/he can't affect the mappings directly.</p>
  */  
  protected static class UnboundIdAssignerImpl extends UnboundIdAssigner
  {
    private Map myMap_=new HashMap();
    private Set freeIds_=new HashSet();
    private int maxIdAssignedSoFar_=-1;
    
    public void reset()  //TESTED
    {
      maxIdAssignedSoFar_=-1;
      myMap_.clear();
      freeIds_.clear();
    };
    
    
    /**
    * <p>Returns the distance of <code>tree</code> to the root of the large
    * {@link FormulaDominoTree} it is part of.</p>
    * <p>A return value of 0 means that <code>tree</code> is the root.</p>
    * <p>The maximum number returned by this function will be
    * <code>max{0,maxdist}</code>. This can be used for optimization because
    * the search for the root will stop after <code>maxdist</code> parent levels
    * have been found.</p>
    */
    protected int distanceToRoot(FormulaDominoTree tree, int maxdist) //TESTED
    {
      int n=0;
      while (tree.rootConnection_!=null && n<=maxdist) 
      {
        ++n;
        tree=tree.rootConnection_;
      };
      return n;
    };
    
    /**
    * <p>Returns the highest (=closest to the root) node in the 
    * {@link FormulaDominoTree} that is equivalent to <code>node</code>.</p>
    * <p><code>node must me an unbound or forward node that is owned by a
    * {@link FormulaDominoTree}.</p>
    */
    protected FormulaSyntaxNet highestEquivalent(FormulaSyntaxNet node)
    {  
      FormulaDominoTree tree=(FormulaDominoTree)node.owner();
      int rootdist=distanceToRoot(tree,Integer.MAX_VALUE);
      
      FormulaSyntaxNet nodeToAssign=node;
      FSNIterator i=node.forwards();
      while (i.hasNext())
      {
        FormulaSyntaxNet n=i.nextFSN();
        if (n.isUnboundOrForward())
        {
          int rdist=distanceToRoot((FormulaDominoTree)n.owner(),rootdist);
          if (rdist<rootdist || (rdist==rootdist && n.uid()<nodeToAssign.uid()))
          {
            rootdist=rdist;
            nodeToAssign=n;
          };
        };
      };
      return nodeToAssign;
    };
    
    
    /**
    * <p>Returns the mapping for <code>node</code> after creating one if necessary.
    * <code>node</code> must be an unbound or forward node whose owner is
    * a {@link FormulaDominoTree}.</p>
    * <p>This function is not part of {@link FormulaDominoTree.UnboundIdAssigner}, because users
    * must not affect the mappings directly 
    * (i.e. without {@link FormulaDominoTree}s being involved).</p>
    */
    protected int getIdFor(FormulaSyntaxNet node) //TESTED
    {
      FormulaSyntaxNet nodeToAssign=highestEquivalent(node);
      
      Integer id=(Integer)myMap_.get(nodeToAssign);
      if (id==null)
      {
        if (freeIds_.isEmpty())
        {
          freeIds_.add(new Integer(++maxIdAssignedSoFar_));
        };
        Iterator iter=freeIds_.iterator();
        id=(Integer)iter.next();
        iter.remove();
        myMap_.put(nodeToAssign,id);
      };
      
      return id.intValue();
    };
    
    public void removeMappingsFor(FormulaDominoTree tree) //TESTED
    {
      removeMappingsFor(tree, new HashSet(), true);
    };
    
    public void removeMappingsStrictlyFor(FormulaDominoTree tree)
    {
      removeMappingsFor(tree, new HashSet(), false);
    };
    
    protected void removeMappingsFor(FormulaDominoTree tree, Set haveVisited, boolean followForwards) 
    { //TESTED
      removeMappingsFor(tree.root_, haveVisited, followForwards);
      for (int i=0; i<tree.bonus_.length; ++i)
        removeMappingsFor(tree.bonus_[i], haveVisited, followForwards);
        
      for (int i=0; i<tree.leaves_.length; ++i)
      {
        removeMappingsFor(tree.leaves_[i], haveVisited, followForwards);
        if (tree.leafConnections_[i]!=null) removeMappingsFor(tree.leafConnections_[i], haveVisited, followForwards);
      };
    };
    
    protected void removeMappingsFor(FormulaSyntaxNet n, Set haveVisited, boolean followForwards) 
    { //TESTED
      if (haveVisited.contains(n)) return;
      haveVisited.add(n);
      
      if (n.isUnboundOrForward())
      {
        removeSingleMappingFor(n);
        if (followForwards)
        {
          FSNIterator i=n.forwards();
          while (i.hasNext()) removeMappingsFor(i.nextFSN(),haveVisited, followForwards);
        };  
      }
      else
      if (n.isBranch())
      {
        removeMappingsFor(n.left(),haveVisited, followForwards);
        removeMappingsFor(n.right(),haveVisited, followForwards);
      };
    };
    
    public void removeUnneededMappingsFor(FormulaDominoTree tree)
    {
      removeUnneededMappingsFor(tree, new HashSet());
    };
    
    protected void removeUnneededMappingsFor(FormulaDominoTree tree, Set haveVisited) 
    {
      removeUnneededMappingsFor(tree.root_, haveVisited);
      for (int i=0; i<tree.bonus_.length; ++i)
        removeUnneededMappingsFor(tree.bonus_[i], haveVisited);
        
      for (int i=0; i<tree.leaves_.length; ++i)
      {
        removeUnneededMappingsFor(tree.leaves_[i], haveVisited);
        if (tree.leafConnections_[i]!=null) removeUnneededMappingsFor(tree.leafConnections_[i], haveVisited);
      };
    };
    
    protected void removeUnneededMappingsFor(FormulaSyntaxNet n, Set haveVisited) 
    {
      if (haveVisited.contains(n)) return;
      haveVisited.add(n);
      
      if (n.isUnboundOrForward())
      {
        FormulaSyntaxNet nodeToAssign=highestEquivalent(n);
        
        if (n!=nodeToAssign) removeSingleMappingFor(n);
        
        boolean foundBranchOrBound=false;
        FSNIterator i=n.forwards();
        while (i.hasNext())
        {
          FormulaSyntaxNet n2=i.nextFSN();
          if (n2.isUnboundOrForward()) 
            { if (n2!=nodeToAssign) removeSingleMappingFor(n2); }
          else 
            { foundBranchOrBound=true; removeUnneededMappingsFor(n2,haveVisited); }
        };  

        if (foundBranchOrBound) removeSingleMappingFor(nodeToAssign);
      }
      else
      if (n.isBranch())
      {
        removeUnneededMappingsFor(n.left(),haveVisited);
        removeUnneededMappingsFor(n.right(),haveVisited);
      };
    };
    
    protected void removeSingleMappingFor(FormulaSyntaxNet n) //TESTED
    {
      Integer i=(Integer)myMap_.remove(n);
      if (i!=null)
      {
        freeIds_.add(i);
        if (i.intValue()==maxIdAssignedSoFar_)
        {
          while (freeIds_.contains(new Integer(maxIdAssignedSoFar_)))
          {
            freeIds_.remove(new Integer(maxIdAssignedSoFar_));
            --maxIdAssignedSoFar_;
          };
        }; 
      };
    };
  };
};



/**
* <p>This class is a specialized variant of {@link FormulaDominoTree}.</p>
*/
class BonusFormulaDominoTree extends FormulaDominoTree
{
  /** 
  * <p><code>originator_.bonusSquaresInAction[bonusIndex_].contains(this)==true</code>.</p> 
  */
  private FormulaDominoTree originator_;
  /** 
  * <p><code>originator_.bonusSquaresInAction[bonusIndex_].contains(this)==true</code>.</p> 
  */
  private int bonusIndex_;
  
  /**
  * <p>Creates an uninitialized <code>BonusFormulaDominoTree</code>.
  * For use by {@link FormulaDominoTree} only.</p>
  */ 
  protected BonusFormulaDominoTree()
  {
  };
  
  /**
  * <p>Initializes an uninitialized <code>BonusFormulaDominoTree</code>.</p> 
  * <p>This function must not be called on an already initialized tree.</p>
  * <p>This function does not add <code>this</code> to <code>originator</code>'s
  * <code>bonusSquaresInAction_</code> list, nor do other changes in
  * <code>originator</code> take place.</p>
  *
  * @see FormulaDominoTree#init(FormulaSyntaxNet, FormulaSyntaxNet[], FormulaSyntaxNet[])
  */
  protected void init(FormulaDominoTree originator, int bonusIndex, FormulaSyntaxNet root)
  { //TESTED
    FormulaSyntaxNet[] leaves=new FormulaSyntaxNet[0];
    FormulaSyntaxNet[] bonus=new FormulaSyntaxNet[0];
    super.init(root,leaves,bonus);
    originator_=originator;
    bonusIndex_=bonusIndex;
  };
  
  /** <p>Throws {@link IllegalNodeTypeException}.</p>*/
  public void attachTo(DominoTree to, int leaf)
  {
    throw new IllegalNodeTypeException();
  };
  
  /** <p>Returns the {@link FormulaDominoTree} thanks to which this bonus
  * square was used.</p> */
  public FormulaDominoTree originator() {return originator_;}
  
  /**
  * <p>Finds and detaches all bonus squares that originate in a
  * <code>FormulaDominoTree</code> contained in <code>parents</code>.</p>
  * <p>This function does not deunify or reunify the tree or 
  * invalidate any caches.
  * This means that after calling this function the <code>FormulaDominoTree</code>
  * is likely to be in an inconsistent state.</p>
  */ 
  protected void detachAllBonusSquaresOriginatingIn(Collection parents) //TESTED
  {
    if (parents.contains(originator()))
    {
      doDetach();
    };  
  }; 
  
  /**
  * <p>Clears the pointers between <code>this</code> and its parent, and removes
  * <code>this</code> from {@link #originator()}'s 
  * {@link FormulaDominoTree#bonusSquaresInAction_ bonusSquaresInAction_} list.</p>
  */
  protected void doDetach() //TESTED
  {
    if (Debugging.enabled)
    {
      if (rootConnection_==null) throw new RuntimeException("This call should not have happened!");
      if (!originator().bonusSquaresInAction_[bonusIndex_].contains(this)) 
        throw new RuntimeException("*sniff* *sniff* My father has disowned me :~-(");
    };
    
    FormulaDominoTree oldRootConnection=rootConnection_;
    rootConnection_=null;
    oldRootConnection.leafConnections_[rootConnectionLeafIndex_]=null;
    rootConnectionLeafIndex_=-1;
    originator().bonusSquaresInAction_[bonusIndex_].remove(this);
  };  
  
  /**
  * <p>If the <code>DominoTree</code> is a subtree of a larger tree, it is
  * detached from the parent tree. If the tree is not a subtree, nothing happens.</p>
  * <p>All bonus squares in the subtree to be detached, that originate from
  * <code>DominoTree</code>s not part of the detached subtree will be detached
  * from the subtree and cannot be reattached.</p>
  *
  * @return <code>true</code> if subtree was detached; <code>false</code> if tree
  *         had no parent tree to begin with.
  *
  * @see #attachTo(DominoTree,int)
  */
  public boolean detachFromParent() //TESTED
  {
    if (rootConnection_==null) return false;
    
    FormulaDominoTree oldConnection=rootConnection_;
    doDetach();
    
    //it would be great if we could just use FormulaSyntaxNet.extractByOwner() to
    //detach the tree. However (see doc for extractByOwner()) there are some
    //unifications that might remain. So we have no choice but to reunify the whole
    //tree.
    oldConnection.reunifyWholeTree(); //reunify parent part
    //the child part doesn't need to be reunified as the above will automatically
    //kill all forwards to this and there are none that need to be recreated
       
    //because detaching bonus square messes up open leaves caches, we have to
    //invalidate.
    oldConnection.invalidateOpenLeavesCachesUpwards(); //invalidate in parent part
    
    //no need to invalidate leaves caches in parent because a bonus square counts
    //as one leaf just as an open connection
    
    //no need to invalidate caches in this because this does not have any leaves
    return true;
  };
  
  /** <p>Throws {@link RuntimeException}.</p>*/
  protected FormulaDominoTree recursiveNonUnifiedClone(FormulaDominoTree cloneTreeRootConnection, int cloneTreeRootConnectionLeafIndex,Map mapBonusSquareToClone, WantedCloneMap wcmap)
  {
    throw new RuntimeException("This method should not be called on BonusFormulaDominoTrees");
  };
  
  
};

