/* Copyright (C) 2002 Matthias S. Benkmann <m.s.b@gmx.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License (ONLY THIS VERSION).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package domino.logic;

import java.util.List;
import java.util.LinkedList;
import java.util.Stack;
import java.util.Vector;
import java.util.Iterator;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import domino.core.Debugging;
import java.util.ConcurrentModificationException;
import java.util.NoSuchElementException;
import java.lang.UnsupportedOperationException; //to help javadoc


/**
* This class represents logical formulae with shared propositional variables
* as a net (a forest of interconnected trees). 
* <p><b>Neither this class nor its iterators are thread-safe! Note that
* synchronizing on a FormulaSyntaxNet does not help, because it will only
* protect one node of the net. You will need an external semaphore to control
* access to the whole net, if you want to use it from multiple threads.
* </b></p>
* <p>
* The following node
* types may be present in a <code>FormulaSyntaxNet</code>:
* <dl>
* <dt>bound leaf</dt>
*   <dd>
*     A bound leaf is a leaf node (a node with no outgoing edges) tagged with
*     the name of a propositional variable.
*   </dd>
*
* <dt>unbound leaf</dt>
*   <dd>
*     An unbound leaf is a placeholder for a subformula. When a 
*     <code>FormulaSyntaxNet</code> is unified with another one, unbound leaves are linked
*     to the corresponding nodes in the other <code>FormulaSyntaxNet</code>.
*   </dd>
*
* <dt>branch</dt>
*   <dd>
*     A branch is a node with 2 outgoing edges (left and right). The branch
*     node represents an implication. The left edge points at the premise and
*     the right edge at the conclusion.
*   </dd>
*
* <dt>forward</dt>
*   <dd>
*     A forward node is created during unification when an unbound leaf is
*     identified with a node in another <code>FormulaSyntaxNet</code>. A forward node
*     points to one or more <code>FormulaSyntaxNet</code>s that are all equivalent. For
*     the purpose of processing the formula(e) represented by a given
*     <code>FormulaSyntaxNet</code>, a forward node is to be replaced by one of the 
*     <code>FormulaSyntaxNet</code>s it points to that is not a forward node.<br>
*     Note that a forward node never points to itself.
*   </dd>
* <dt></dt>
* </dl>
* </p>
* <p><b><code>FormulaSyntaxNet</code>s must never be recursive (i.e. no branch node
* may point to a node that is equivalent to a direct or indirect parent node of
* the branch node). For such a <code>FormulaSyntaxNet</code>, all methods have
* undefined behaviour.</b><p>
*/
public class FormulaSyntaxNet
{
  /**
  * A list of {@link ForwardConnection} objects that point to <code>FormulaSyntaxNet</code>s
  * that are all equivalent, one of which (that is not a forward node) must 
  * be picked and substituted for the forward node, when processing 
  * a <code>FormulaSyntaxNet</code>.
  *  
  * <p>
  * Together with the variables {@link #id_}, {@link #left_} and {@link #right_}, 
  * <code>forwards_</code> determines the type of node 
  * (see {@link FormulaSyntaxNet class description}) as follows:
  * </p>
  * <ul>
  *   <li>A node with <code>id_!=null</code> is a propositional variable with
  *     name id_  (a bound leaf). </li>
  *   <li>A node with <code>forwards_!=null</code> is a forward node. </li>
  *   <li>A node with <code>left_!=null && right_!=null</code> is a branch. </li>
  *   <li>A node with all 4 variables <code>==null</code> is an unbound leaf. </li>
  * </ul>
  *
  *  <p>
  *  <b>Invariants:</b>
  *  <dl>
  *  <dt><code>	if (forwards_!=null)				</code></dt>
  *  <dd><code>	forwards_.isEmpty()==false &&			</code></dd>
  *  <dd><code>	id_==null && left_==null && right_==null	</code></dd>
  *  </dl>
  *  </p>
  *
  * @see #id_
  * @see #left_
  * @see #right_
  * @see FormulaSyntaxNet <code>FormulaSyntaxNet</code> description
  */
  private List /*of ForwardConnection*/ forwards_=null;
  
  /**
  * The <code>backwards_</code> list stores references to all 
  * <code>FormulaSyntaxNet</code>s 
  * that point to this <code>FormulaSyntaxNet</code>.
  *
  * @see #forwards_
  * @see #left_
  * @see #right_
  */
  private List /*of FormulaSyntaxNet*/ backwards_=new LinkedList();
  
  /**
  * Points to the <code>FormulaSyntaxNet</code> that specifies the 
  * premise of an implication (branch node).
  *
  * <p>See {@link #forwards_} for details on how this variable influences the
  * type of a <code>FormulaSyntaxNet</code> node.</p>
  *
  *  <p>
  *  <b>Invariants:</b>
  *  <dl>
  *  <dt><code>  if (left_!=null)  		</code></dt>
  *  <dd><code>  right_!=null &&   		<br>
  *              left_!=right_ &&               <br>
  *              id_==null && forwards_==null   </code></dd>
  *  </dl>
  *  </p>
  * 
  * @see #forwards_
  * @see #id_
  * @see #right_
  * @see FormulaSyntaxNet <code>FormulaSyntaxNet</code> description
  */
  private FormulaSyntaxNet left_=null;
  
  /**
  * Points to the <code>FormulaSyntaxNet</code> that specifies the 
  * conclusion of an implication (branch node).
  *
  * <p>See {@link #forwards_} for details on how this variable influences the
  * type of a <code>FormulaSyntaxNet</code> node.</p>
  *
  *  <p>
  *  <b>Invariants:</b>
  *  <dl>
  *  <dt><code>  if (right_!=null)  		</code></dt>
  *  <dd><code>  left_!=null &&   		<br>
  *              left_!=right_ &&               <br>
  *              id_==null && forwards_==null   </code></dd>
  *  </dl>
  *  </p>
  * 
  * @see #forwards_
  * @see #id_
  * @see #left_
  * @see FormulaSyntaxNet <code>FormulaSyntaxNet</code> description
  */
  private FormulaSyntaxNet right_=null;
  
  /**
  * Name of the propositional variable a bound leaf represents.
  *
  * <p>See {@link #forwards_} for details on how this variable influences the
  * type of a <code>FormulaSyntaxNet</code> node.</p>
  *
  *  <p>
  *  <b>Invariants:</b>
  *  <dl>
  *  <dt><code>  if (id_!=null)  		</code></dt>
  *  <dd><code>  left_==null && right_==null && forwards_==null   </code></dd>
  *  </dl>
  *  </p>
  * 
  * @see #forwards_
  * @see #left_
  * @see #right_
  * @see FormulaSyntaxNet <code>FormulaSyntaxNet</code> description
  */
  private String id_=null;
  
  /**
  * The owner of this <code>FormulaSyntaxNet</code>. Every 
  * <code>FormulaSyntaxNet</code> has an owner. The owner can be any
  * {@link Object}.
  *
  * <p><b>Invariants:</b> <code>owner_!=null</code></p>
  *
  * @see #owner()
  */
  private Object owner_;
  
  /**
  * A unique identifier. No 2 <code>FormulaSyntaxNet</code>s have the same
  * <code>uid_</code> value.
  */
  private final long uid_=newUID();
  
  /**
  * The number that will be assigned to the <code>uid_</code> of the next
  * <code>FormulaSyntaxNet</code> that is created.
  */
  private static long uidsource_=0;
  
  /** <p>Creates and returns a new unique identifier.</p>*/
  synchronized private static long newUID() {return uidsource_++;};
  
  /**
  * <p>Incremented by 1 for every modification that invalidates iterators over
  * {@link #forwards_} (for forward nodes) and iterators over 
  * {@link #left_} and {@link #right_} (for branch nodes).</p>
  *
  * <p><b>This variable is not used by all iterators. Some rely on the
  * underlying java.util.* collection for throwing 
  * {@link java.util.ConcurrentModificationException}</b></p>
  */
  private int forwardsDirty_=0;
  
  /**
  * <p>Incremented by 1 for every modification that invalidates iterators over
  * {@link #backwards_}.</p>
  *
  * <p><b>This variable is not used by all iterators. Some rely on the
  * underlying java.util.* collection for throwing 
  * {@link java.util.ConcurrentModificationException}</b></p>
  */
  private int backwardsDirty_=0;
  
  /**
  * Iterators over {@link #forwards_} (for forward nodes) or 
  * {@link #left_} and {@link #right_} (for branch nodes)
  * should store this number on their creation and 
  * invalidate themselves when it changes.
  * 
  * <p><b>This variable is not used by all iterators. Some rely on the
  * underlying java.util.* collection for throwing 
  * {@link java.util.ConcurrentModificationException}</b></p>
  *
  * @see #dirtyNumberBackwards()
  * @see #forwards_
  */
  int dirtyNumberForwards() {return forwardsDirty_;};
  
  /**
  * Iterators over {@link #backwards_} should store this number on their creation and 
  * invalidate themselves when it changes.
  *
  * <p><b>This variable is not used by all iterators. Some rely on the
  * underlying java.util.* collection for throwing 
  * {@link java.util.ConcurrentModificationException}</b></p>
  *
  *
  * @see #dirtyNumberForwards()
  * @see #backwards_
  */
  int dirtyNumberBackwards() {return backwardsDirty_;};
  
  /**
  * Returns the owner of this <code>FormulaSyntaxNet</code>. Every 
  * <code>FormulaSyntaxNet</code> has an owner. The owner can be any
  * {@link Object}.
  */
  public Object owner() {return owner_;};
  
  /** 
  * Tests the integrity of the object for which it is called.
  * If debugging is enabled, this method is called after every mutator method
  * and every constructor.
  * <p>Some tests are only done if 
  * <code>{@link Debugging#expensiveTestsEnabled}==true</code>.</p>
  *
  * @see #checkForwardConsistency()
  *
  * @throws RuntimeException if any of the object's invariants are violated.
  */
  public void checkInvariant()
  { //next invariant violation count is 12
    if (forwards_!=null)
    {
      if (forwards_.isEmpty() || left_!=null || right_!=null || id_!=null) throw new InvariantViolation("1");
    }
    else if (id_!=null)
    {
      if (left_!=null || right_!=null || forwards_!=null) throw new InvariantViolation("2");
    }
    else if (left_!=null || right_!=null)
    {
      if (left_==null || right_==null || forwards_!=null || id_!=null) throw new InvariantViolation("3");
      //hmm why did I forbid the following? fromInfixFormulaString() produces this : if (left_==right_) throw new InvariantViolation("9");
    };
   
    if (owner_==null) throw new InvariantViolation("4");
   
    if (Debugging.expensiveTestsEnabled)
    try{
      if (this.isForward())
      { //test if all forward connections have exactly 1 backwards connection 
        //to this, also test if this has a forward connection to itself (illegal)
        FSNIterator i=forwards();
        while (i.hasNext())
        {
          FormulaSyntaxNet n=i.nextFSN();
          if (n==this) throw new InvariantViolation("11");
          FSNIterator j=n.backwards();
          int found=0;
          while (j.hasNext()) { if (j.nextFSN()==this) ++found; };
          if (found!=1) throw new InvariantViolation("10-"+found);
        };
      }
      else if (this.isBranch())
      { //test if left and right branch both have exactly one backwards 
        //connections to this
        int found=0;
        for(int i=0; i<2; ++i)
        {
          FormulaSyntaxNet n;
          if (i==0) n=left(); else n=right();
          FSNIterator j=n.backwards();
          while (j.hasNext()) { if (j.nextFSN()==this) ++found; };
        };
        if (found!=2) throw new InvariantViolation("6-"+found);
      };
      
      //test if all backward connections have exactly 1 forward or 
      //at least 1 branch connection to this
      FSNIterator i=backwards();
      while (i.hasNext())
      {
        int found=0;
        FormulaSyntaxNet n=i.nextFSN();
        if (n.isBranch())
        {
          if (n.left()==this||n.right()==this) ++found;
        }
        else
        {
          FSNIterator j=n.forwards();
          while (j.hasNext())
          {
            if (j.nextFSN()==this) ++found;
          };
        };  
        if (found!=1) throw new InvariantViolation("7-"+found);
      };
      
      if (this.isForward())
      { //test if all forward nodes pointed to by this forward node point to
        //the same set of nodes. This includes the test that a forward node
        //must not point to an unbound node (because that node should be a
        //forward node pointing back).
        java.util.Collection targetSet=new java.util.HashSet();
        i=forwards();
        while (i.hasNext()) targetSet.add(new Long(i.nextFSN().uid()));
        targetSet.add(new Long(this.uid()));
        
        i=forwards();
        while (i.hasNext())
        {
          FormulaSyntaxNet n=i.nextFSN();
          if (n.isUnboundOrForward())
          {
            java.util.Set compareSet=new java.util.HashSet();
            FSNIterator j=n.forwards();
            while (j.hasNext()) compareSet.add(new Long(j.nextFSN().uid()));
            compareSet.add(new Long (n.uid()));
            if (!compareSet.equals(targetSet)) throw new InvariantViolation("8");
          };  
        };
      };
    }catch(InvariantViolation e)
    {
      throw e;
    }  
    catch(Exception e) 
    {
      throw new RuntimeException("unexpected exception in FormulaSyntaxNet.checkInvariant(): "+e);
    };
  };
 
  /**
  * <p>Tests that all forward connections from forward nodes are equivalent (have 
  * been properly unified).</p>
  * <p>If <code>{@link Debugging#enabled}==true</code> and 
  * <code>{@link Debugging#expensiveTestsEnabled}==true</code>, this method is 
  * called after {@link #unify(FormulaSyntaxNet,FormulaSyntaxNet)} and
  * {@link #extractByOwner(FormulaSyntaxNet,Collection)}.</p>
  *
  * @see #checkInvariant()
  * @see #isEquivalent(FormulaSyntaxNet,FormulaSyntaxNet)
  *
  * @throws RuntimeException if non-equivalent forward connections are found.
  */
  public void checkForwardConsistency()
  {
    if (!(isEquivalent(this,this))) throw new RuntimeException("non-equivalent forward connections detected");
  };


  /**
  * <p>Tests whether <code>a</code> and <code>b</code> are equivalent.</p>
  * <p>This method always walks through the whole net, even if <code>a</code>
  * and <code>b</code> are the same {@link Object}. This allows using this
  * function to check the consistency of a <code>FormulaSyntaxNet</code>.</p>
  *
  * @see #checkForwardConsistency()
  *
  * @return <code>true</code> if <code>a</code> and <code>b</code> are
  *         equivalent.
  */
  public static boolean isEquivalent(FormulaSyntaxNet a,FormulaSyntaxNet b)
  {
    return isEquivalent(a,b,new HaveVisitedPair());
  };
  
  
  /**
  * <p>See {@link #isEquivalent(FormulaSyntaxNet,FormulaSyntaxNet)}.</p>
  *
  * @param a <code>FormulaSyntaxNet</code> to check for equivalence
  * @param b <code>FormulaSyntaxNet</code> to check for equivalence
  * @param haveVisited used internally - pass <code>new HaveVisitedPair()</code>
  * 
  * @return <code>true</code> if <code>a</code> and <code>b</code> are
  *         equivalent.
  */
  protected static boolean isEquivalent(FormulaSyntaxNet a,FormulaSyntaxNet b, HaveVisitedPair haveVisited)
  {
    //do !!NOT!! short-cut if a==b (see documentation)
    if (haveVisited.includesUnordered(a,b)) return true;
    
    /* Note regarding haveVisited: Infinite loops can only occur if a
       group of equivalent forward nodes point to each other. All of these
       forward nodes are located on the same level of the net (i.e. the same
       distance to the tree roots if only branch nodes are counted). Because
       forward nodes on different levels can never point to each other, we
       only need to keep track of the forward nodes visited since the last
       pair of branch/bound nodes. 
       So with every pair of branch/bound nodes we encounter on our recursive
       walk through the net we reset haveVisited and we only add pairs to
       haveVisited that have at least one forward or unbound node.
    */
    
    //if we have at least one bound node or branch, make sure it is a
    if (b.isBoundOrBranch()) {FormulaSyntaxNet c=a; a=b; b=c;};
    
    if (a.isBoundOrBranch())
    {
      if (b.isBoundOrBranch())
      {
        if ( (a.isBound()!=b.isBound()) || 
             (a.isBound() && !a.id().equals(b.id()))
           ) return false;
           
        if (a.isBranch()) 
        {
          if ( !(
                 isEquivalent(a.left(),b.left(),new HaveVisitedPair()) &&
                 isEquivalent(a.right(),b.right(),new HaveVisitedPair())
                )
             ) return false;
        }; 
      }
      else //if b.isUnboundOrForward()
      {
        haveVisited.addUnordered(a,b);
        FSNIterator i=b.forwards();
        boolean foundNonUnboundForward=false;
        while (i.hasNext()) 
        {
          FormulaSyntaxNet n=i.nextFSN();
          if (!n.isUnboundOrForward()) foundNonUnboundForward=true;
          if (!isEquivalent(a,n,haveVisited)) return false;
        };
        
        //Because of haveVisited, all isEquivalent() calls above would return true
        //if all of the forward connections are just to other forward nodes. Therefore
        //we test if there was at least one real forward.
        //This also catches the error that b is an unbound node with no forward
        //connections.
        if (!foundNonUnboundForward) return false;
      };
    }
    else if (a.isUnboundOrForward()) 
    { //because of the swapping at the beginning of this method, b is neither
      //bound nor branch node, so b.isUnboundOrForward() is true.
      
      haveVisited.addUnordered(a,b);
      
      //if we have at least one forward node, make sure it is a
      if (b.isForward()) {FormulaSyntaxNet c=a; a=b; b=c;};
      
      if (b.isUnbound())
      {
        //unbound nodes are only equivalent if they are the same Object
        //(because otherwise they could get out of sync)
        if (! (a==b) ) return false;
      } 
      else //if (b.isForward())
      {
        FSNIterator i=a.forwards();
        FSNIterator j=b.forwards();
        if (!(i.hasNext() && j.hasNext())) return false; //check for buggy iterators
        while (j.hasNext()) if (!isEquivalent(a,j.nextFSN(),haveVisited)) return false;
        while (i.hasNext()) 
        {
          FormulaSyntaxNet c=i.nextFSN();
          if (!isEquivalent(c,b,haveVisited)) return false;
          j=b.forwards();
          while (j.hasNext()) 
            if (!isEquivalent(c,j.nextFSN(),haveVisited)) return false;
        };  
      };
    };
    return true;
  };
  
  
  /** 
  * Returns <code>true</code> if the node is a forward node 
  * (see {@link FormulaSyntaxNet class description}).
  *
  * @see #isForward()
  * @see #isBound()
  * @see #isBoundOrBranch()
  * @see #isBranch()
  * @see #isUnbound()
  * @see #isUnboundOrForward()
  */
  public boolean isForward() { return forwards_!=null;};
  
  /** 
  * Returns <code>true</code> if the node is a branch node 
  * (see {@link FormulaSyntaxNet class description}).
  *
  * @see #isForward()
  * @see #isBound()
  * @see #isBoundOrBranch()
  * @see #isBranch()
  * @see #isUnbound()
  * @see #isUnboundOrForward()
  */
  public boolean isBranch() {return left_!=null;};
  
  /** 
  * Returns <code>true</code> if the node is an unbound leaf node
  * (see {@link FormulaSyntaxNet class description}).
  *
  * @see #isForward()
  * @see #isBound()
  * @see #isBoundOrBranch()
  * @see #isBranch()
  * @see #isUnbound()
  * @see #isUnboundOrForward()
  */
  public boolean isUnbound() {return !(isBranch() || isBound() || isForward());};
  
  /** 
  * Returns <code>true</code> if the node is an unbound leaf node or a 
  * forward node (see {@link FormulaSyntaxNet class description}).
  *
  * @see #isForward()
  * @see #isBound()
  * @see #isBoundOrBranch()
  * @see #isBranch()
  * @see #isUnbound()
  * @see #isUnboundOrForward()
  */
  public boolean isUnboundOrForward() {return isUnbound()||isForward();};
  
  /** 
  * Returns <code>true</code> if the node is a bound leaf node or a 
  * branch node (see {@link FormulaSyntaxNet class description}).
  *
  * @see #isForward()
  * @see #isBound()
  * @see #isBoundOrBranch()
  * @see #isBranch()
  * @see #isUnbound()
  * @see #isUnboundOrForward()
  */
  public boolean isBoundOrBranch() {return isBound()||isBranch();};
  
  /** 
  * Returns <code>true</code> if the node is a bound leaf node
  *
  * (see {@link FormulaSyntaxNet class description}).
  * @see #isForward()
  * @see #isBound()
  * @see #isBoundOrBranch()
  * @see #isBranch()
  * @see #isUnbound()
  * @see #isUnboundOrForward()
  */
  public boolean isBound() {return id_!=null;};
  
  /**
  * Adds a forward connection to an unbound leaf node or a forward node.
  * <p>If <code>newForward</code> is already in the list of forward connections,
  * it will not be added a second time.</p>
  * <p>Calling this method will invalidate iterators that iterate over the
  * list of forward connections for <code>this</code> and iterators that
  * iterate over the list of backward connections of <code>newForward</code>.</p>
  *
  * <p><b>This method does not call {@link #checkInvariant()}, so the
  * <code>FormulaSyntaxNet</code> structure may be violated after calling
  * this method.</b></p>
  *
  * @param newForward the <code>FormulaSyntaxNet</code> to connect with. 
  *          {@link #unify(FormulaSyntaxNet,FormulaSyntaxNet) unify()} should be used to
  *          make sure that <code>newForward</code> is equivalent to all existing 
  *          forward connections (which is required for proper forward nodes).
  * @return <code>true</code> if a new connection was added.
  * @throws IllegalNodeTypeException if <code>this</code> is neither unbound
  *         leaf nor forward node.
  */
  protected boolean addForwardUnsafe(FormulaSyntaxNet newForward)
  {
    if (!isUnboundOrForward()) throw new IllegalNodeTypeException();
    if (isUnbound()) forwards_=new Vector();
    FSNIterator i=forwards();
    while(i.hasNext())
    {
      if (newForward==i.nextFSN()) return false;
    };
    ForwardConnection newCon=new ForwardConnection(this,newForward);
    ++forwardsDirty_; //invalidate iterators
    forwards_.add(newCon);
    ++newForward.backwardsDirty_; //invalidate iterators
    newForward.backwards_.add(this);
    return true;
  };
  
  
  /**
  * Removes a forward connection from an unbound leaf node or a forward node.
  * <p>If <code>oldForward</code> is not in the list of forward connections,
  * nothing is done.</p>
  * <p>Calling this method will invalidate iterators that iterate over the
  * list of forward connections for <code>this</code> and iterators that
  * iterate over the list of backward connections of <code>oldForward</code>.</p>
  *
  * <p><b>This method does not call {@link #checkInvariant()}, so the
  * <code>FormulaSyntaxNet</code> structure may be violated after calling
  * this method.</b></p>
  *
  * @param oldForward the <code>FormulaSyntaxNet</code> to remove connection to. 
  * @return <code>true</code> if a connection was removed.
  * @throws IllegalNodeTypeException if <code>this</code> is neither unbound
  *         leaf nor forward node.
  */
  protected boolean removeForwardUnsafe(FormulaSyntaxNet oldForward)
  {
    if (!isUnboundOrForward()) throw new IllegalNodeTypeException();
    
    if (isUnbound()) return false;
    
    FConnIterator i=forwardConns();
    while(i.hasNext())
    {
      if (oldForward==i.nextFConn().to()) 
      { 
        i.remove();
        return true;  
      };
    };
    
    return false;
  };
  
  
  /**
  * <p>Removes the forward connection <code>oldForward</code> which must be
  * the {@link ForwardConnection#to() to()} part of a {@link ForwardConnection} returned 
  * by a preceding call to <code>i.next()</code>.</p>
  * <p>Calling this method will invalidate iterators that iterate over the
  * list of forward connections for <code>this</code> and iterators that
  * iterate over the list of backward connections of <code>oldForward</code>.</p>
  *
  * <p><b>This method does not call {@link #checkInvariant()}, so the
  * <code>FormulaSyntaxNet</code> structure may be violated after calling
  * this method.</b></p>
  *
  * @param i an iterator that returned a {@link ForwardConnection} pointing to
  *        <code>oldForward</code> on a preceding call to <code>next()</code>.
  * @param oldForward the <code>FormulaSyntaxNet</code> to remove connection to. 
  * @throws IllegalStateException if <code>i.remove()</code> would throw it.
  * @throws ConcurrentModificationException if <code>i.remove()</code> would throw it.
  * @throws UnsupportedOperationException if <code>i.remove()</code> would throw it.
  */
  protected void removeForwardUnsafe(Iterator i, FormulaSyntaxNet oldForward)
  {
    i.remove(); //must be before any other commands that modify the structure
                //because this iterator could throw IllegalState.. or ConcurrentMod..
    ++forwardsDirty_; //invalidate iterators
    if (forwards_.isEmpty()) forwards_=null;
    ++oldForward.backwardsDirty_; //invalidate iterators
    oldForward.backwards_.remove(this);
  };  
  
  
  /**
  * <p>Returns <code>true</code> if <code>this</code> is a descendant of or equivalent
  * to a descendant of <code>b</code>.</p>
  * <p>If <code>this==b</code> or <code>this</code> is equivalent to <code>b</code>
  * the return value is <code>false</code>.</p>
  */
  public boolean isDescendantOf(FormulaSyntaxNet b)
  {
    return (this.isDescendantOf(b,new java.util.HashSet(),true)>0);
  };
  
  /**
  * <p>Determines whether <code>this</code> is a descendant of or equivalent
  * to a descendant of <code>b</code>.</p>
  *
  * @param b ancestor to check for
  * @param haveVisited pass a new {@link HashSet}.
  * @param equiv pass <code>true</code>.
  *
  * @return <code>-1</code> if <code>this</code> is equivalent to <code>b</code>,
  *         <code>+1</code> if <code>this</code> is a descendant of <code>b</code>,
  *         <code>0</code> if <code>this</code> is not a descendant of <code>b</code>.
  *
  * @see #isDescendantOf(FormulaSyntaxNet)
  */
  protected int isDescendantOf(FormulaSyntaxNet b, Collection haveVisited, boolean equiv)
  {
    if (this==b) 
    {
      if (equiv) return -1; else return +1;
    };  
    
    if (!haveVisited.contains(this))
    {
      haveVisited.add(this);
      
      FSNIterator i=backwards();
      while (i.hasNext()) 
      {
        FormulaSyntaxNet n=i.nextFSN();
        int ret=n.isDescendantOf(b,haveVisited,equiv && n.isForward());
        if (ret!=0) return ret;
      };
        
      //the following part is not redundant because a forward node can point to a
      //node that is not a forward node, so that it will not be in backwards()
      if (this.isForward())
      {
        i=forwards();
        while (i.hasNext()) 
        {
          FormulaSyntaxNet n=i.nextFSN();
          int ret=n.isDescendantOf(b,haveVisited,equiv);
          if (ret!=0) return ret;
        };  
      };
    };
    
    return 0;
  };
  
  
  /**
  * Checks <code>a</code> and <code>b</code> for compatibility and crosslinks
  * unbound and forward nodes with their corresponding nodes from the respective
  * other net. 
  *
  * <p>After a successful call to <code>unify()</code>,
  * <code>a</code> and <code>b</code> are equivalent.</p>
  *
  * <p><code>unify</code> either succeeds completely or throws an exception and 
  * leaves <code>a</code> and <code>b</code> (and other nodes visited during
  * the unification) unchanged.</p>
  *
  * <p>This method may invalidate iterators related to <code>this</code> and
  * to nodes reachable from <code>this</code> by following forward connections.
  * Do not use any such iterators after calling this method.</p>
  *
  * <p>Unifying a <code>FormulaSyntaxNet</code> with one of its
  * subnets would create an infinite recursion in the formula represented by the
  * net, so it is disallowed.</p>
  * 
  * @throws ConflictingBindException if <code>a</code> and <code>b</code>
  *         can not be unified because at least one bound leaf in <code>a</code>
  *         or <code>b</code> does not match a corresponding (bound or unbound)
  *         leaf in the other one, or because a node would have to be unified with
  *         a direct or indirect parent.
  */
  public static void unify(FormulaSyntaxNet a, FormulaSyntaxNet b) 
  {
    ConflictingBindException conflict=null;
    
    List madeConnections=new Vector();
    try
    {
      unify(a,b,madeConnections,new HaveVisitedPair());

      /*
        Check the new net for recursion by walking through net a and keeping track of
        all branches visited on every path. It is an error if the same branch is
        encountered twice on the same path (it's okay on different paths).
        It doesn't matter whether a or b is chosen as starting point as unify() makes
        them equivalent.
      */
      Collection haveVisitedOnThisPath=new HashSet();
      FSNIterator iter=a.multiorderIterator(true,false,true);
      while (iter.hasNext())
      {
        FormulaSyntaxNet next=iter.nextFSN();
        
        if (next.isBranch())
        {
          if (iter.preorder())
          {
            Long uid=new Long(next.uid());
            if (haveVisitedOnThisPath.contains(uid)) throw new ConflictingBindException();
            haveVisitedOnThisPath.add(uid);
          }
          else //if (iter.postorder())
          {
            Long uid=new Long(next.uid());
            haveVisitedOnThisPath.remove(uid);
          };
        };  
      };
    }
    catch(ConflictingBindException x)
    {
      conflict=x;
      Iterator i=madeConnections.iterator();
      while (i.hasNext())
      {
        FormulaSyntaxNet from=(FormulaSyntaxNet)i.next();
        FormulaSyntaxNet to=(FormulaSyntaxNet)i.next();
        from.removeForwardUnsafe(to);
      };
    };
    
    if (Debugging.enabled)
    {
      Iterator i=madeConnections.iterator();
      while (i.hasNext())
      {
        FormulaSyntaxNet from=(FormulaSyntaxNet)i.next();
        i.next(); //throw away "to" because it either doesn't change 
                  //or reappears as "from" later.
        from.checkInvariant();
      };
      
      if (Debugging.expensiveTestsEnabled) 
      {
        a.checkForwardConsistency(); 
        b.checkForwardConsistency();
        if (conflict==null && !isEquivalent(a,b)) throw new RuntimeException("unify() failed abnormally");
      };  
    };
    
    if (conflict!=null) throw conflict;
  };
  
  /**
  * <p>Like {@link #unify(FormulaSyntaxNet,FormulaSyntaxNet)} but leaves net in a
  * broken state in the case of an exception.</p>
  * <p><b>Unlike {@link #unify(FormulaSyntaxNet,FormulaSyntaxNet)}
  * this method does never perform debugging tests.</b></p>
  *
  *
  * @param a <code>FormulaSyntaxNet</code> to be unified
  * @param b <code>FormulaSyntaxNet</code> to be unified
  * @param conns must be initialized with an empty {@link List}.
  *        On successful return <code>conns</code> contains the connections that were
  *        made to make <code>a</code> and <code>b</code> equivalent.
  *        If an exception is thrown
  *        this list contains the connections that were made before the exception, i.e.
  *        the connections that need to be undone to restore the original nets.
  *        The 1st, 3rd, 5th,... entry is the <code>FormulaSyntaxNet</code>
  *        (a forward or unbound node) where the connection originates and
  *        the 2nd, 4th, 6th entry is the <code>FormulaSyntaxNet</code> the
  *        connection points to.<br>
  *        This list may contain duplicate entries.
  * @param haveVisited contains unordered 
  *        <code>(FormulaSyntaxNet,FormulaSyntaxNet)</code> pairs of nodes
  *        that have already been unified. This is for internal use by this
  *        method only and must be initialized with an empty 
  *        <code>HaveVisitedPair</code> object by the caller.
  *
  * @throws ConflictingBindException if <code>a</code> and <code>b</code>
  *         can not be unified because at least one bound leaf in <code>a</code>
  *         or <code>b</code> does not match a corresponding (bound or unbound)
  *         leaf in the other one.
  *
  * @see #unify(FormulaSyntaxNet,FormulaSyntaxNet)
  */
  protected static void unify(FormulaSyntaxNet a, FormulaSyntaxNet b, List conns, HaveVisitedPair haveVisited)
  {
    if (a==b || haveVisited.includesUnordered(a,b)) return;
    
    /* Note regarding haveVisited: Infinite loops can only occur if a
       group of equivalent forward nodes point to each other. All of these
       forward nodes are located on the same level of the net (i.e. the same
       distance to the tree roots if only branch nodes are counted). Because
       forward nodes on different levels can never point to each other, we
       only need to keep track of the forward nodes visited since the last
       pair of branch/bound nodes. 
       So with every pair of branch/bound nodes we encounter on our recursive
       walk through the net we reset haveVisited and we only add pairs to
       haveVisited that have at least one forward or unbound node.
       
       UPDATE: The above is outdated now, because the code tests for already
       existing connections and doesn't recurse into them. I still keep haveVisited
       around because it could improve performance. However, this would probably
       require a hashing scheme in HaveVisited and would only have an effect on
       pathological cases with many crosslinked forward nodes.
       Anyway, keeping haveVisited doesn't hurt that much.
       
       UPDATE: Hashing scheme for HaveVisitedPair now implemented.
    */
    
    //if we have at least one bound node or branch, make sure it is a
    if (b.isBoundOrBranch()) {FormulaSyntaxNet c=a; a=b; b=c;};
    
    if (a.isBoundOrBranch())
    {
      if (b.isBoundOrBranch())
      {
        if ( (a.isBound()!=b.isBound()) || 
             (a.isBound() && !a.id().equals(b.id()))
           ) throw new ConflictingBindException();
           
        if (a.isBranch()) 
        {
          unify(a.left(),b.left(),conns,new HaveVisitedPair());
          unify(a.right(),b.right(),conns,new HaveVisitedPair());
        }; 
      }
      else //if b.isUnboundOrForward()
      {
        haveVisited.addUnordered(a,b);

        if (b.addForwardUnsafe(a))  //if a and b not yet unified
        {
          conns.add(b);
          conns.add(a);
        
          /**
            Recursive unification can create a recursive FormulaSyntaxNet
            (this is an illegal condition that will be detected later in 
            unify(FormulaSyntaxNet,FormulaSyntaxNet), a b.forwards() iterator could
            be invalidated due to a call to unify(b, whatever) or unify(whatever,b)
            (where whatever != a because unify(b,a) and unify(a,b) are caught because
            the connection already exists).
            Because of the above problem we copy b's forward connections for
            iteration.
          */
          FormulaSyntaxNet[] bforwards=new FormulaSyntaxNet[b.forwards_.size()];
          FSNIterator i=b.forwards();
          int j=0;
          while (i.hasNext()) bforwards[j++]=i.nextFSN();
          while (--j>=0) unify(a,bforwards[j],conns,haveVisited);
        };  
      };
    }
    else //if (a.isUnboundOrForward()) 
    { //because of the swapping at the beginning of this method, b is neither
      //bound nor branch node, so b.isUnboundOrForward() is true.
      
      haveVisited.addUnordered(a,b);
      
      if (a.addForwardUnsafe(b))  //if a and b not yet unified
      {
        b.addForwardUnsafe(a);
        conns.add(a);
        conns.add(b);
        conns.add(b);
        conns.add(a);
      
        //Unifying a with the forward connections of b may lead to the adding of
        //new connections to b and/or a. The same is true for unifying b with the 
        //forward connections of a. This will cause forwards() iterators to be
        //invalidated. For that reason, we have to add the necesarry connections
        //to a and b before we call unify in iterator loops.
        List aForwardsWithoutB=new Vector();
        List bForwardsWithoutA=new Vector();
        
        FSNIterator i=a.forwards();
        while (i.hasNext()) 
        {
          FormulaSyntaxNet n=i.nextFSN();
          if (n!=b) aForwardsWithoutB.add(n);
        };  
        i=b.forwards();
        while (i.hasNext()) 
        {
          FormulaSyntaxNet n=i.nextFSN();
          if (n!=a) bForwardsWithoutA.add(n);
        };
          
        /*
          The following block adds all of b's connections to a and all of a's
          connections to b. All of a's connections that are forward nodes will also
          get connected to b and all of b's connections that are forward nodes
          will also get connected to a.
        */
        final FormulaSyntaxNet[] fsna={a,b};
        final List[] la={bForwardsWithoutA, aForwardsWithoutB};
        for (int idx=0; idx<2; ++idx)
        {
          Iterator iter=la[idx].iterator();
          while (iter.hasNext())
          {
            FormulaSyntaxNet n=(FormulaSyntaxNet)iter.next();
            if (fsna[idx].addForwardUnsafe(n))
            {
              conns.add(fsna[idx]);
              conns.add(n);
            };
              
            if (n.isUnboundOrForward()) 
            {
              if (n.addForwardUnsafe(fsna[idx]))
              {
                conns.add(n);
                conns.add(fsna[idx]);
              };  
              
              Iterator iter2=la[1-idx].iterator();
              while (iter2.hasNext())
              {
                FormulaSyntaxNet n2=(FormulaSyntaxNet)iter2.next();
                if (n.addForwardUnsafe(n2))
                {
                  conns.add(n);
                  conns.add(n2);
                };  
              };
            };  
          };
        };
          
        /* 
          At this point a, b and all of their forward connections are
          completely interconnected.
          All that has to be done now is to unify those connections.
          Because forward nodes are completely connected, we can limit the
          unification to bound and branch nodes.
        */  
        Iterator iter=aForwardsWithoutB.iterator();
        while (iter.hasNext())
        {
          FormulaSyntaxNet n=(FormulaSyntaxNet)iter.next();
          
          if (n.isUnboundOrForward()) continue;
          
          Iterator iter2=bForwardsWithoutA.iterator();
          while (iter2.hasNext())
          {
            FormulaSyntaxNet n2=(FormulaSyntaxNet)iter2.next();
            
            if (n2.isUnboundOrForward()) continue;
            
            unify(n,n2,conns,new HaveVisitedPair());
          };
        };
      };
    };  
  };
  
  
  /**
  * <p>Extracts a <code>FormulaSyntaxNet</code> that contains 
  * only nodes owned by {@link Object}s contained in <code>owners</code>.</p>
  * <p>If <code>n</code> is not owned by an {@link Object} found 
  * in <code>owners</code>, then <code>null</code> is returned and nothing else
  * happens.</p>
  * <p>Otherwise, beginning with <code>n</code> all reachable 
  * <code>FormulaSyntaxNet</code> nodes (not just children and forward connections of
  * <code>n</code> but also parents and backward connections) will be scanned
  * recursively for nodes with an owner found in <code>owners</code>. From these nodes
  * all connections (forward and backward) that are not branches are removed if they
  * connect with a node not owned by an {@link Object} from <code>owners</code>.</p>
  * <p>The scan stops at nodes owned by {@link Object}s not in <code>owners</code>,
  * which means that after this function returns there may still remain some nodes 
  * that are owned by {@link Object}s from <code>owners</code> and have non-branch
  * connections to nodes owned by {@link Object}s not in <code>owners</code>. The
  * nodes in question are not extracted because they are not reachable from 
  * <code>n</code> through a path of nodes all owned by {@link Object}s from
  * <code>owners</code>.</p>
  * <p><b>Note that this method can not always be used to undo a
  * {@link #unify(FormulaSyntaxNet,FormulaSyntaxNet) unify()}.</b><br>
  * Example:</p>
  * <p><code>FormulaSyntaxNet a=newLeaf(owner1,null);<br>
  * FormulaSyntaxNet b=newLeaf(owner1,null);<br>
  * FormulaSyntaxNet c=newLeaf(owner2,null);<br>
  * unify(a,c); unify(c,b); <br>
  * HashSet owners=new HashSet(); owners.add(c);<br>
  * extractByOwner(c,owners);</code></p>
  * <p><code>a</code> and <code>b</code> will be unified after the example commands.
  * </p>
  *
  * @param n node where the scanning starts.
  * @param owners {@link Collection} of {@link Object}s permissible as owners of nodes
  *        in the extracted net.
  *
  * @return <code>null</code> if <code>n</code>'s owner is not in <code>owners</code>.
  *          Otherwise returns <code>n</code>.
  *
  * @see #removeAllForwards(FormulaSyntaxNet) 
  */
  public static FormulaSyntaxNet extractByOwner(FormulaSyntaxNet n, Collection owners)
  {
    if (!owners.contains(n.owner())) return null;
    Collection haveVisited=new java.util.HashSet();
    splitByOwner(n,owners,haveVisited);
    
    if (Debugging.enabled)
    {
      Iterator i=haveVisited.iterator();
      while (i.hasNext()) ((FormulaSyntaxNet)i.next()).checkInvariant();
      if (Debugging.expensiveTestsEnabled) n.checkForwardConsistency();
    };
    
    return n;
  };
  
  /**
  * <p>see {@link #extractByOwner(FormulaSyntaxNet,Collection)}.</p>
  * <p><b>This method does not call {@link #checkInvariant()}!</b></p>
  *
  * @param Pass an empty {@link Collection}. After <code>splitByOwner()</code> returns
  *        this contains all nodes that have been visited by this method.
  */
  protected static void splitByOwner(FormulaSyntaxNet n, Collection owners, Collection haveVisited)
  {
    if (haveVisited.contains(n) || !owners.contains(n.owner())) return;
    haveVisited.add(n);
    
    /*
      Gather all nodes that point to n via forward connections and are not owned
      by a permissible owner. Ignore those that point here via branches because
      we can't chop off branches (would violate the left_!=null!=right_ invariant).
    */
    List backForwards=new Vector();
    Iterator i=n.backwards();
    while (i.hasNext())
    {
      FormulaSyntaxNet back=(FormulaSyntaxNet)i.next();
      if (back.isForward() && !owners.contains(back.owner())) backForwards.add(back);
    };

    /*
      Now remove the forward connections from the nodes determined above. We have
      to do this separately through iterating over backForwards instead of putting
      it directly in the above loop because removeForwardUnsafe() would invalidate
      our n.backwards() iterator.
    */
    i=backForwards.iterator();
    while(i.hasNext()) ((FormulaSyntaxNet)i.next()).removeForwardUnsafe(n);
    
    if (n.isBranch())
    {
      /* now split left and right child recursively */
      splitByOwner(n.left(), owners, haveVisited);
      splitByOwner(n.right(), owners, haveVisited);
    }
    else if (n.isForward())
    {
      /* Remove all forward connections to nodes with no permissible owner. */
      FConnIterator j=n.forwardConns();
      while (j.hasNext())
      {
        ForwardConnection conn=j.nextFConn();
        if (!owners.contains(conn.to().owner())) j.remove();
      };
      
      /* 
        Now we split remaining forward connections recursively. Because n has a
        permissible owner, this recursive splitting can not cause any further removes
        from n.forwards_ that would invalidate iterator i.
      */
      i=n.forwards();
      while (i.hasNext()) splitByOwner((FormulaSyntaxNet)i.next(), owners, haveVisited);
    }; 
    
    /* 
      Now split remaing backwards connections recursively. Because n has a
      permissible owner and all backwards connections to nodes with non-permissible
      owners have either been removed already or are (non-removable) branches 
      the recursive split can not cause further removes from n.backwards_ that would
      invalidate iterator i.
    */
    i=n.backwards();
    while (i.hasNext()) splitByOwner((FormulaSyntaxNet)i.next(), owners, haveVisited);
  };
  
  
  /**
  * <p>Recursively removes all forward connections from <code>n</code> and all
  * nets connected to <code>n</code> via branches or forward connections,
  * backwards or forwards.</p>
  *
  * @see #extractByOwner(FormulaSyntaxNet,Collection)
  */
  public static void removeAllForwards(FormulaSyntaxNet n)
  {
    Collection haveVisited=new java.util.HashSet();
    Stack todo=new Stack();
    todo.add(n);
    removeAllForwards(todo,haveVisited);
    
    if (Debugging.enabled)
    {
      //very incomplete test as removeAllForwards(Stack, haveVisited) only'
      //stores branch nodes in haveVisited.
      Iterator i=haveVisited.iterator();
      while (i.hasNext()) ((FormulaSyntaxNet)i.next()).checkInvariant();
    };
  };
  
  
  /**
  * <p>See {@link #removeAllForwards(FormulaSyntaxNet)}.</p>
  * @param todo contains all {@link FormulaSyntaxNet}s to process.
  * @param haveVisited Pass an empty {@link Collection} (preferably a {@link HashSet}).
  */
  protected static void removeAllForwards(Stack todo,Collection haveVisited)
  {
    while(!todo.isEmpty())
    {
      FormulaSyntaxNet n=(FormulaSyntaxNet)todo.pop();
      if (haveVisited.contains(n)) continue;
     
      FSNIterator i=n.backwards();
      while (i.hasNext()) todo.add(i.nextFSN());
    
      if (n.isBranch())
      {
        haveVisited.add(n);
        todo.add(n.left());
        todo.add(n.right());
      }
      else if (n.isForward())
      {
        FConnIterator j=n.forwardConns();
        while (j.hasNext()) 
        {
          todo.add(j.nextFConn().to());
          j.remove();
        };  
      };
    };  
  };
  
  
  /**
  * <p>Traverses the whole net (forwards and backwards) and stores all of its nodes in
  * <code>nodes</code>.</p>
  * <p>After this method returns, <code>nodes</code> will contain <code>this</code> 
  * and all
  * the nodes necessary so that all nodes in <code>nodes</code> point to and are
  * pointed to only by nodes in <code>nodes</code>.</p>
  *
  * @param nodes (should be a {@link Collection} with good 
  *        {@link Collection#add(Object) add} and 
  *        {@link Collection#contains(Object) contains} performace) collects all 
  *        nodes of the whole <code>FormulaSyntaxNet</code>. 
  *        <code>nodes</code> is cleared before the execution of this method, so it
  *        need not be initially empty.
  */
  public void getAllReachableNodes(Collection nodes)
  {
    nodes.clear();
    getAllReachableNodesNoClear(nodes);
  };  
    
  /**
  * <p>Like {@link #getAllReachableNodes(Collection)} but does not clear 
  * <code>nodes</code>. Nodes contained in <code>nodes</code> will not be visited, i.e
  * nodes only reachable through these nodes will not be part of the final 
  * collection.</p>
  */
  protected void getAllReachableNodesNoClear(Collection nodes)
  {
    if (nodes.contains(this)) return;
    nodes.add(this);
    
    if (this.isBranch())
    {
      left().getAllReachableNodesNoClear(nodes);
      right().getAllReachableNodesNoClear(nodes);
    }
    else if (this.isForward())
    {
      FSNIterator i=forwards();
      while (i.hasNext()) i.nextFSN().getAllReachableNodesNoClear(nodes);
    }; 
    
    FSNIterator i=backwards();
    while (i.hasNext()) i.nextFSN().getAllReachableNodesNoClear(nodes);
  };
    
  /**
  * <p>Returns <code>true</code> if <code>this</code> and <code>obj</code> are physically the same
  * {@link Object}.</p>
  */
  public boolean equals(Object obj) {return this==obj;};
  
  /**
  * <p>Returns the left branch of a branch node.</p>
  *
  * @throws IllegalNodeTypeException if <code>this</code> is not a branch node.
  *
  * @see #right()
  */
  public FormulaSyntaxNet left() 
  {
    if (!isBranch()) throw new IllegalNodeTypeException();
    return left_;
  };
  
  /**
  * <p>Returns the right branch of a branch node.</p>
  *
  * @throws IllegalNodeTypeException if <code>this</code> is not a branch node.
  *
  * @see #left()
  */
  public FormulaSyntaxNet right() 
  {
    if (!isBranch()) throw new IllegalNodeTypeException();
    return right_;
  };
  
  /**
  * Returns the unique id of this <code>FormulaSyntaxNet</code>.
  * No 2 <code>FormulaSyntaxNet</code>s return the same number for 
  * <code>uid()</code>
  */
  public long uid() {return uid_;};
  
  /**
  * <p>Returns the id of the propositional variable for a bound node.</p>
  *
  * @throws IllegalNodeTypeException if <code>this</code> is not a bound node.
  *
  */
  public String id() 
  {
    if (!isBound()) throw new IllegalNodeTypeException();
    return id_;
  };
  
  /**
  * Returns an iterator to visit all the nodes that a forward node points to.
  * This function may be called on forward nodes and unbound leaves.
  *
  * @throws IllegalNodeTypeException if <code>this</code> is neither unbound
  *         leaf nor forward node.
  *
  * @see #backwards()
  * @see #preorderIterator()
  * @see #inorderIterator()
  * @see #postorderIterator()
  * @see #multiorderIterator(boolean,boolean,boolean)
  */
  public FSNIterator forwards() {return new FSNIteratorFromFConnIterator(forwardConns());};
  
  /**
  * Returns an iterator to visit all the (branch and forward) nodes that 
  * point to <code>this</code>.
  *
  * @see #forwards()
  * @see #preorderIterator()
  * @see #inorderIterator()
  * @see #postorderIterator()
  * @see #multiorderIterator(boolean,boolean,boolean)
  */
  public FSNIterator backwards() {return new FSNIteratorFromIterator(backwards_.iterator());};
  
  /**
  * <p>Returns an iterator that visits bound and unbound leaves and branch nodes 
  * in preorder sequence.</p> 
  * <p>Forward nodes will not be visited unless all nodes
  * equivalent to the forward node are also forward nodes. In that case,
  * the forward node that is visited is the one with the lowest {@link #uid()}.
  * </p>
  * <p>For every set of equivalent nodes at the same position in the net
  * only one representative will be visited. Equivalent nodes in different
  * branches of the net will be visited multiple times.</p>
  *
  * @see #forwards()
  * @see #inorderIterator()
  * @see #postorderIterator()
  * @see #multiorderIterator(boolean,boolean,boolean)
  */
  public FSNIterator preorderIterator() {return new FSNMultiorderIterator(this,true,false,false);};
  
  /**
  * <p>Returns an iterator that visits bound and unbound leaves and branch nodes 
  * in inorder sequence.</p> 
  * <p>Forward nodes will not be visited unless all nodes
  * equivalent to the forward node are also forward nodes. In that case,
  * the forward node that is visited is the one with the lowest {@link #uid()}.
  * </p>
  * <p>For every set of equivalent nodes at the same position in the net
  * only one representative will be visited. Equivalent nodes in different
  * branches of the net will be visited multiple times.</p>
  *
  * @see #forwards()
  * @see #preorderIterator()
  * @see #postorderIterator()
  * @see #multiorderIterator(boolean,boolean,boolean)
  */
  public FSNIterator inorderIterator() {return new FSNMultiorderIterator(this,false,true,false);};
  
  /**
  * <p>Returns an iterator that visits bound and unbound leaves and branch nodes 
  * in postorder sequence.</p> 
  * <p>Forward nodes will not be visited unless all nodes
  * equivalent to the forward node are also forward nodes. In that case,
  * the forward node that is visited is the one with the lowest {@link #uid()}.
  * </p>
  * <p>For every set of equivalent nodes at the same position in the net
  * only one representative will be visited. Equivalent nodes in different
  * branches of the net will be visited multiple times.</p>
  *
  * @see #forwards()
  * @see #inorderIterator()
  * @see #postorderIterator()
  * @see #multiorderIterator(boolean,boolean,boolean)
  */
  public FSNIterator postorderIterator() {return new FSNMultiorderIterator(this,false,false,true);};
  
  /**
  * <p>Returns an iterator that visits bound and unbound leaves and branch nodes 
  * in pre/in/postorder sequence.</p> 
  * <p>Forward nodes will not be visited unless all nodes
  * equivalent to the forward node are also forward nodes. In that case,
  * the forward node that is visited is the one with the lowest {@link #uid()}.
  * </p>
  * <p>For every set of equivalent nodes at the same position in the net
  * only one representative will be visited. Equivalent nodes in different
  * branches of the net will be visited multiple times.</p>
  *
  * <p><code>pre</code>, <code>in</code> and <code>post</code> determine
  * whether and when the iterator will return branch nodes. If all 3
  * values are <code>false</code>, only leaf nodes will be returned. If
  * more than one is <code>true</code>, branch nodes will be visited
  * multiple times.</p>
  * 
  * @param pre stop at branch node before visiting left child
  * @param in stop at branch node before visiting right child
  * @param post stop at branch node after visiting right child
  *
  * @see #forwards()
  * @see #inorderIterator()
  * @see #postorderIterator()
  * @see #preorderIterator()
  */
  public FSNIterator multiorderIterator(boolean pre, boolean in, boolean post) {return new FSNMultiorderIterator(this,pre,in,post);};
  
  /**
  * <p>Returns an iterator to visit all {@link FormulaSyntaxNet.ForwardConnection}s that a 
  * forward node has. This function may be called on forward nodes and 
  * unbound leaves.</p>
  *
  * @throws IllegalNodeTypeException if <code>this</code> is neither unbound
  *         leaf nor forward node.
  *
  * @see #forwards()
  * @see #preorderIterator()
  * @see #inorderIterator()
  * @see #postorderIterator()
  * @see #multiorderIterator(boolean,boolean,boolean)
  */
  protected FConnIterator forwardConns()//DO NOT MAKE this method public because this exposes the FSN to unchecked modification throw FConnIterator.remove()!
  {
    if (!isUnboundOrForward()) throw new IllegalNodeTypeException();
    if (isUnbound()) return new EmptyFConnIterator();
    return new FConnIteratorForFSN(this);
  };
  
  
  /**
  * <p>Creates a new <code>FormulaSyntaxNet</code> that is a (bound or unbound)
  * leaf node.</p>
  *
  * @param owner the owner of the new object
  * @param id the name of the bound leaf or <code>null</code> if the new leaf
  *           is to be unbound.
  *
  * @return a new <code>FormulaSyntaxNet</code> 
  */
  public static FormulaSyntaxNet newLeaf(Object owner, String id)
  {
    FormulaSyntaxNet newNode=new FormulaSyntaxNet(id,null,null,owner);
    if (Debugging.enabled) newNode.checkInvariant();
    return newNode;
  };
  
  /**
  * <p>Creates a new <code>FormulaSyntaxNet</code> that is a branch node.</p>
  *
  * @param owner the owner of the new object
  * @param left the left branch
  * @param right the right branch
  *
  * @return a new <code>FormulaSyntaxNet</code> 
  */
  public static FormulaSyntaxNet newBranch(Object owner, FormulaSyntaxNet left, FormulaSyntaxNet right)
  {
    FormulaSyntaxNet newNode=new FormulaSyntaxNet(null,left,right,owner);
    ++left.backwardsDirty_; //invalidate iterators
    left.backwards_.add(newNode);
    if (left!=right)
    {
      ++right.backwardsDirty_; //invalidate iterators
      right.backwards_.add(newNode);
    };  
    
    if (Debugging.enabled) 
    {
      newNode.checkInvariant();
      left.checkInvariant();
      right.checkInvariant();
    };  
    return newNode;
  };
  
  /**
  * <p>Returns the formula represented by the <code>FormulaSyntaxNet</code>
  * as a {@link String} in infix format, like (A-&gt;B) .</p>
  * <p>Unbound leaves will be represented as X0, X1,...</p>
  */
  public String toInfixFormulaString()
  {
    StringBuffer s=new StringBuffer();
    FSNIterator i=multiorderIterator(true,true,true);
    java.util.Map map=new java.util.HashMap(); //maps uid()s to String names
    int unboundCount=0; //number of unbound leaves encountered so far
    while (i.hasNext())
    {
      FormulaSyntaxNet node=i.nextFSN();
            if (node.isBranch()) 
      {
        if (i.preorder()) s.append('(');
        if (i.inorder()) s.append("->");
        if (i.postorder()) s.append(')');
      }  
      else if (node.isBound()) s.append(node.id());
      else if (node.isUnboundOrForward())
      {
        String id=(String)map.get(new Long(node.uid()));
        if (id==null)
        {
          id="X"+(unboundCount++);
          map.put(new Long(node.uid()),id);
        };
        s.append(id);
      }
      else throw new RuntimeException("Huh? What's this node?"); //just to be safe
    };
    return s.toString();
  };
  
  
  /**
  * <p>Returns a new FormulaSyntaxNet that corresponds to <code>s</code>,
  * which must be a formula string in infix format, like (A-&gt;B) .</p>
  * <p>Unbound leaves must be represented as X&lt;num&gt;, where &lt;num&gt;
  * is any non-negative integer.</p>
  * <p>The net that is returned will not contain forward nodes. Leaf nodes with 
  * the same id will
  * be identical (i.e. the same object).</p>
  * <p>No associativity is assumed for -&gt;, so A-&gt;B-&gt;C must be
  * written as either (A-&gt;B)-&gt;C or A-&gt;(B-&gt;C).</p>
  *
  * @param owner the owner of the <code>FormulaSyntaxNet</code> to create.
  * @param s the infix formula string to interpret
  *
  * @throws IllegalArgumentException if there is a syntax error in <code>s</code> or
  *        <code>s==null</code>.
  *
  * @see #newFromInfixFormulaStrings(Object, String[])
  */
  public static FormulaSyntaxNet newFromInfixFormulaString(Object owner, String s)
  {
    return newFromInfixFormulaString(owner,s,new java.util.HashMap());
  };
  
  /**
  * <p>Like {@link #newFromInfixFormulaString(Object,String)} but with an
  * extra parameter to take influence on leaf creation.</p>
  *
  * @param leaves must map {@link String}s to <code>FormulaSyntaxNet</code>s.
  *        Whenever a leaf is about to be created, this mapping is consulted. If
  *        a leaf already exists it is reused. Otherwise a new leaf is created and
  *        a mapping from its name to the newly created leaf is added to 
  *        <code>leaves</code>. This parameter is useful when creating several
  *        <code>FormulaSyntaxNet</code>s that are to share leaves.
  */
  protected static FormulaSyntaxNet newFromInfixFormulaString(Object owner, String s, Map leaves)
  {
    if (s==null) throw new IllegalArgumentException();
    
    Stack operandStack=new Stack(); //stores FormulaSyntaxNets
    Stack operatorStack=new Stack(); //stores Integers that carry the 
                                    //parenthese level of the operator
    
    boolean error=false; //if a syntax error is found, this is set to true and the while loop is broken
    boolean parseOperand=false; //if true, currentOperand will be converted to a FormylaSyntaxNet and put on the operandStack at the end of the while loop
    boolean applyOperator=false; //if true, the top operator will be applied at the end of the while loop
    int i=0; //index of the current character to parse
    int state=0; //0=>waiting for operand, 1=>parsing operand, 2=>waiting for ')' to apply operator
    int parentheses=0; //current parentheses level
    int lastOperatorLevel=-1; //parentheses level of the top (most recent) unapplied operator
    StringBuffer currentOperand=new StringBuffer();
    while (i<s.length())
    {
      if (s.charAt(i)=='(')
      {
        ++parentheses;
        if (state!=0) {error=true; break;};
      }
      else if (s.charAt(i)==')')
      {
        if (state==0) {error=true; break;};
        if (state==1) parseOperand=true;
        if (parentheses==lastOperatorLevel) applyOperator=true;
        --parentheses;
        if (parentheses<0) {error=true; break;};
        state=2;
      }
      else
      if (s.regionMatches(i,"->",0,2))
      {
        if (state==0 || lastOperatorLevel==parentheses) {error=true; break;};
        if (state==1) parseOperand=true;

        state=0;
        lastOperatorLevel=parentheses;
        operatorStack.push(new Integer(parentheses));
        
        ++i; //extra increment because "->" has 2 characters
      }
      else //if s.charAt(i) part of operand
      {
        if (state==2) {error=true; break;};
        state=1;
        currentOperand.append(s.charAt(i));
      };
      
      ++i;
      
      if (i==s.length())
      {
        if (parentheses!=0) {error=true; break;};
        
        if (state==1) parseOperand=true;
        
        //determine the number of operands expected to be on operandStack
        //we start with 1 because we expect the final result to be the only thing
        //on the stack
        int expectedOperands=1;
         
        //if we still have to put one operand on the operand stack, we expect one
        //operand less to be there now
        if (parseOperand) --expectedOperands;
        
        if (!operatorStack.isEmpty())
        {
          //we expect one more operand per operator
          expectedOperands+=operatorStack.size();
          applyOperator=true;
        };
        
        if (operandStack.size()!=expectedOperands) {error=true; break;};
      };
      
      if (parseOperand)
      {
        String operand=currentOperand.toString();
        currentOperand.delete(0,currentOperand.length());
        FormulaSyntaxNet net=(FormulaSyntaxNet)leaves.get(operand);
        if (net==null)
        {
          boolean unbound=true;
          if (operand.charAt(0)!='X' || operand.length()<2) 
            unbound=false;
          else
            for (int j=1; j<operand.length(); ++j)
              if (!Character.isDigit(operand.charAt(j))) {unbound=false; break;};
          
          if (unbound) net=newLeaf(owner, null);
          else         net=newLeaf(owner, operand);
          
          leaves.put(operand, net);
        };
        operandStack.push(net);
        parseOperand=false;
      };
      
      while (applyOperator)
      {
        /*
          We don't need to test if operandStack has at least 2 elements.
          The i==s.length() block above tests explicitly for this before setting
          applyOperator=true and the if (s.charAt(i)==')') block tests this
          implicitly via the state variable.
        */
        FormulaSyntaxNet right=(FormulaSyntaxNet)operandStack.pop();
        FormulaSyntaxNet left=(FormulaSyntaxNet)operandStack.pop();
        operandStack.push(newBranch(owner, left, right));
        
        operatorStack.pop();
        
        if (operatorStack.isEmpty()) 
        {
          lastOperatorLevel=-1;
        }
        else
        {  
          lastOperatorLevel=((Integer)operatorStack.peek()).intValue();
        };
        
        if (i==s.length()) 
          applyOperator=!operatorStack.isEmpty();
        else 
          applyOperator=false;  
      };
    };
    
    if (error) throw new IllegalArgumentException(s.substring(0,i)+" ERROR=>"+s.substring(i));

    if (operandStack.isEmpty()) throw new IllegalArgumentException("empty expression");
    return (FormulaSyntaxNet)operandStack.pop();
  };
  
  /**
  * <p>Like {@link #newFromInfixFormulaString(Object, String)} but creates several
  * nets at the same time.</p>
  * <p>Leaf nodes with the same id will
  * be identical (i.e. the same object) in all returned nets.</p>
  *
  * @param owner the owner of the <code>FormulaSyntaxNet</code> to create.
  * @param s the infix formula strings to interpret.
  *
  * @throws IllegalArgumentException if <code>s==null</code> or 
  *        <code>s.length==0</code> or an element of <code>s</code> is 
  *        <code>null</code> or
  *        there is a syntax error in an element of <code>s</code>.
  *
  * @return array whose elements correspond to those of <code>s</code>, i.e. the 1st
  *         element is the <code>FormulaSyntaxNet</code> that corresponds to the
  *         1st element of <code>s</code>.
  *
  * @see #newFromInfixFormulaString(Object, String)
  */
  public static FormulaSyntaxNet[] newFromInfixFormulaStrings(Object owner, String[] s)
  {
    if (s==null || s.length==0) throw new IllegalArgumentException();
    
    Map leaves=new java.util.HashMap();
    FormulaSyntaxNet[] result=new FormulaSyntaxNet[s.length];
    for (int i=0; i<result.length; ++i)
    {
      if (s[i]==null) throw new IllegalArgumentException();
      result[i]=newFromInfixFormulaString(owner,s[i],leaves);
    };  
      
    return result;
  };
  
  /**
  * <p>Constructor for internal use only. Use the <code>new*()</code> methods
  * to create <code>FormulaSyntaxNet</code>s.</p>
  *
  * <p><b>Does not call {@link #checkInvariant()}</b>.</p>
  *
  * @see #newBranch(Object, FormulaSyntaxNet, FormulaSyntaxNet)
  * @see #newLeaf(Object, String)
  * @see #newFromInfixFormulaString(Object, String)
  */
  private FormulaSyntaxNet(String id,FormulaSyntaxNet left,FormulaSyntaxNet right,Object owner)
  {
    id_=id;
    left_=left;
    right_=right;
    owner_=owner;
  };
  
  
  /**
  * Connection between 2 <code>FormulaSyntaxNets</code>.
  */  
  protected static class ForwardConnection
  {
    private FormulaSyntaxNet from_;
    private FormulaSyntaxNet to_;
    
    /**
    * Returns the target node of the connection.
    */
    public FormulaSyntaxNet to() {return to_;};
    
    /**
    * Returns the source node of the connection.
    */
    public FormulaSyntaxNet from() {return from_;};
    
    /**
    * Creates a <code>ForwardConnection</code> for connecting <code>from</code>
    * with <code>to</code>. Creating the <code>ForwardConnection</code> object
    * does not actually make a connection between the 2 
    * <code>FormulaSyntaxNet</code>s. Use {@link FormulaSyntaxNet#addForwardUnsafe(FormulaSyntaxNet)}
    * for that.
    *
    * @see FormulaSyntaxNet#addForwardUnsafe(FormulaSyntaxNet)
    */
    public ForwardConnection(FormulaSyntaxNet from, FormulaSyntaxNet to)
    {
      from_=from;
      to_=to;
    };
  };
  
  
  /**
  * Iterates over a collection of <code>FormulaSyntaxNet.ForwardConnections</code>.
  * An <code>FConnIterator</code> fails with a 
  * {@link ConcurrentModificationException} if the underlying collection is
  * modified. Note that this does not prevent all modifications anywhere in
  * the <code>FormulaSyntaxNet</code>. Only modifications that add to or remove
  * from the list of <code>ForwardConnection</code>s that the 
  * <code>FConnIterator</code> iterates over will invalidate the iterator.
  */
  public interface FConnIterator extends java.util.Iterator
  {
    /** Returns true if the iteration has more elements. */
    public boolean hasNext();
  
    /** 
    * Returns the next element in the iteration.
    *
    * @return the next element in the iteration. 
    * @throws NoSuchElementException if the iteration has no more elements.
    * @see #nextFConn()
    */
    public Object next();
  
    /** 
    * Returns the next element in the iteration.
    * @return the next element in the iteration. 
    * @throws NoSuchElementException if the iteration has no more elements.
    * @see #next()
    */
    public ForwardConnection nextFConn();
    
    /**
    * <p>Removes from the underlying collection the last element returned by 
    * the iterator (optional operation). This method can be called only once 
    * per call to next. The behavior of an iterator is unspecified if the 
    * underlying collection is modified while the iteration is in progress in 
    * any way other than by calling this method.
    *
    * @throws UnsupportedOperationException if the remove operation is not 
    *         supported by this Iterator.
    * @throws IllegalStateException if the next method has not yet been 
    *         called, or the remove method has already been called after 
    *         the last call to the next method.
    */
    public void remove();
  };

  
  /**
  * <p>Iterates over the {@link FormulaSyntaxNet.ForwardConnection}s of a
  * {@link FormulaSyntaxNet}.</p>
  */
  protected static class FConnIteratorForFSN implements FConnIterator
  {
    private Iterator myIterator_;
    private FormulaSyntaxNet myFSN_;
    private ForwardConnection current=null;
    
    /**
    * <p>Creates an iterator to iterate over the <code>ForwardConnection</code>s
    * of <code>n</code>.</p>
    */
    public FConnIteratorForFSN(FormulaSyntaxNet n)
    {
      myFSN_=n;
      myIterator_=myFSN_.forwards_.iterator();
    };
    
    /** <p>Returns true if the iteration has more elements.</p> */
    public boolean hasNext() {return myIterator_.hasNext();};
    
    /** 
    * <p>Returns the next element in the iteration.</p>
    * @return the next element in the iteration. 
    * @throws NoSuchElementException if the iteration has no more elements.
    * @see #next()
    */
    public ForwardConnection nextFConn() 
    {
      current=(ForwardConnection)myIterator_.next();
      return current;
    };
    
    /** 
    * <p>Returns the next element in the iteration.</p>
    *
    * @return the next element in the iteration. 
    * @throws NoSuchElementException if the iteration has no more elements.
    * @see #nextFConn()
    */
    public Object next() {return nextFConn();};
    
    /**
    * <p>Removes from the underlying collection the last element returned by 
    * the iterator (optional operation). This method can be called only once 
    * per call to next. The behavior of an iterator is unspecified if the 
    * underlying collection is modified while the iteration is in progress in 
    * any way other than by calling this method.</p>
    *
    * @throws UnsupportedOperationException if the remove operation is not 
    *         supported by this Iterator.
    * @throws IllegalStateException if the next method has not yet been 
    *         called, or the remove method has already been called after 
    *         the last call to the next method.
    */
    public void remove() { myFSN_.removeForwardUnsafe(myIterator_,current.to()); };
  };
  
    
  /**
  * <p>Iterator that never returns an element.
  */
  public class EmptyFConnIterator implements FConnIterator
  {
    /** <p>Returns <code>false</code>.</p> */
    public boolean hasNext() {return false;};
  
    /** 
    * <p>Throws {@link NoSuchElementException}.</p>
    * @see #nextFConn()
    */
    public Object next() {throw new NoSuchElementException();};
  
    /** 
    * <p>Throws {@link NoSuchElementException}.</p>
    * @see #next()
    */
    public ForwardConnection nextFConn() {throw new NoSuchElementException();};
  
    /** 
    * <p>Throws {@link IllegalStateException} .</p>
    *
    * @throws IllegalStateException
    */
    public void remove() {throw new IllegalStateException();};
  };
    

  
  /**
  * A <code>HaveVisitedPair</code> object keeps a list of object pairs
  * that have already been visited by an algorithm.
  */
  protected static class HaveVisitedPair
  {
    private HashSet visited_=new HashSet();
    
    /**
    * Returns true if the pair <code>(o1,o2)</code> or the pair 
    * <code>(o2,o1)</code> has been visited.
    */
    public boolean includesUnordered(Object o1,Object o2)
    {
      return visited_.contains(new Pair(o1,o2));
    };
    
    /**
    * Adds the pairs <code>(o1,o2)</code> and <code>(o2,o1)</code> to the
    * list of visited pairs.
    */
    public void addUnordered(Object o1,Object o2)
    {
      visited_.add(new Pair(o1,o2));
    };
    
    /**
    * Clears the list of visited pairs.
    */
    public void clear()
    {
      visited_.clear();
    };
    
    protected static class Pair
    {
      private Object o1_, o2_;
      public Pair(Object o1,Object o2) {o1_=o1; o2_=o2;};
      public boolean equals(Object o)
      {
        try{
          Pair p=(Pair)o;
          return ((o1_==p.o1_ && o2_==p.o2_)||(o2_==p.o1_ && o1_==p.o2_));
        }catch(Exception x){return false;}
      };
      public int hashCode()
      {
        return o1_.hashCode()^o2_.hashCode();
      };
    };
  };
  
  /**
  * <p>Iterates over the nodes of a {@link FormulaSyntaxNet}. It visits bound
  * and unbound leaves and branch nodes 
  * in pre/in/postorder sequence.</p> 
  * <p>Forward nodes will not be visited unless all nodes
  * equivalent to the forward node are also forward nodes. In that case,
  * the forward node that is visited is the one with the lowest {@link #uid()}.
  * </p>
  * <p>For every set of equivalent nodes at the same position in the net
  * only one representative will be visited. Equivalent nodes in different
  * branches of the net will be visited multiple times.</p>
  * 
  * An <code>FSNMultiorderIterator</code> fails with a 
  * {@link ConcurrentModificationException} if the underlying 
  * {@link FormulaSyntaxNet} is modified. 
  *
  * @see FormulaSyntaxNet#preorderIterator()
  * @see FormulaSyntaxNet#inorderIterator()
  * @see FormulaSyntaxNet#postorderIterator()
  * @see FormulaSyntaxNet#multiorderIterator(boolean,boolean,boolean)
  */
  public static class FSNMultiorderIterator implements FSNIterator
  {
    private Stack branchStack_=new Stack();
    
    /** <p>0: visiting current node for the 1st time <br>
    *   1: visiting again after visiting left child  <br>
    *   2: visiting again after visiting right child <br>
    *   3: node is leaf node .<p>
    */
    private int order_=0;
    private boolean preorderStop_;
    private boolean postorderStop_;
    private boolean inorderStop_;
    
    /**
    * <p>Creates an iterator that visits bound 
    * and unbound leaves and branch nodes 
    * in pre/in/postorder sequence.</p> 
    * <p>Forward nodes will not be visited unless all nodes
    * equivalent to the forward node are also forward nodes. In that case,
    * the forward node that is visited is the one with the lowest {@link #uid()}.
    * </p>
    * <p>For every set of equivalent nodes at the same position in the net
    * only one representative will be visited. Equivalent nodes in different
    * branches of the net will be visited multiple times.</p>
    *
    * <p><code>pre</code>, <code>in</code> and <code>post</code> determine
    * whether and when the iterator will return branch nodes. If all 3
    * values are <code>false</code>, only leaf nodes will be returned. If
    * more than one is <code>true</code>, branch nodes will be visited
    * multiple times.</p>
    * 
    * @param n the {@link FormulaSyntaxNet} to iterate over. May be 
    *        <code>null</code>.
    * @param pre stop at branch node before visiting left child
    * @param in stop at branch node before visiting right child
    * @param post stop at branch node after visiting right child
    *
    * @see FormulaSyntaxNet#preorderIterator()
    * @see FormulaSyntaxNet#inorderIterator()
    * @see FormulaSyntaxNet#postorderIterator()
    * @see FormulaSyntaxNet#multiorderIterator(boolean,boolean,boolean)
    *
    */
    public FSNMultiorderIterator(FormulaSyntaxNet n,boolean pre,boolean in,boolean post)
    {
      preorderStop_=pre;
      inorderStop_=in;
      postorderStop_=post;
      if (n!=null) branchStack_.push(new stackElem(n,0,0));
    };
    
    /** <p>Returns true if the iteration has more elements.</p> */
    public boolean hasNext() {return !branchStack_.isEmpty();};
    
    /** 
    * <p>Returns the next element in the iteration.</p>
    * @return the next element in the iteration. 
    * @throws NoSuchElementException if the iteration has no more elements.
    * @see #next()
    */
    public FormulaSyntaxNet nextFSN()
    {
      FormulaSyntaxNet next_=null;
      boolean getNextFSN=true;
      while (getNextFSN)
      {
        getNextFSN=false;
      
        if (!hasNext()) throw new java.util.NoSuchElementException();
        stackElem e=(stackElem)branchStack_.pop();
      
        next_=e.node;
              
        if (next_.isBranch())
        {
          if (e.order==0)
          {
            branchStack_.push(new stackElem(next_,1,next_.dirtyNumberForwards()));
            branchStack_.push(new stackElem(next_.left(),0,0));
            order_=0;
            if (!preorderStop_) getNextFSN=true;
          } 
          else
          {
            if (e.dirtyNumber!=next_.dirtyNumberForwards())
              throw new ConcurrentModificationException();
        
            if (e.order==1)
            {
              if (postorderStop_) branchStack_.push(new stackElem(next_,2,e.dirtyNumber));
              branchStack_.push(new stackElem(next_.right(),0,0));
              order_=1;
              if (!inorderStop_) getNextFSN=true;
            } 
            else //if (e.order==2)
            {
              order_=2;
              //next_=next_;  
            };
          };
        }
        else
        if (next_.isForward())
        {
          /* equivalent forward nodes and the nodes they point to are always 
          fully connected by unify(), so if there's any non-forward node
          equivalent to next_, it's in next_'s forwards() list */
      
          FormulaSyntaxNet equivalentNodeWithLowestUID=next_;
          long lowestUID=next_.uid();
          FSNIterator i=next_.forwards();
          while (i.hasNext())
          {
            FormulaSyntaxNet n=i.nextFSN();
            if (!n.isForward()) { next_=n; break;};
            if (n.uid()<lowestUID) 
            {
              equivalentNodeWithLowestUID=n;
              lowestUID=n.uid();
            };
          };
      
          //if we found a non-forward node, we have to process it in place 
          //of the current forward node
          if (!next_.isForward()) 
          {
            branchStack_.push(new stackElem(next_,0,0));
            getNextFSN=true;
          } 
          else //otherwise we have to pick the forward node with lowest uid.
          {
            next_=equivalentNodeWithLowestUID;
            order_=3;
          };  
        }
        else //if (next_.isUnbound() || next_.isBound())
        {
          order_=3;
        };
      };
    
      return next_; 
    };
    
    /**
    * <p>If the node returned by the most recent call to
    * {@link #next()} or {@link #nextFSN()} is a branch, this method returns
    * true if it has to be 
    * processed now in a preorder iteration. If the node returned is a leaf,
    * this method always returns true.</p>
    * <p>If {@link #next()} or {@link #nextFSN()} was not called prior to
    * calling this method, its return value is unspecified.</p>
    */
    public boolean preorder() {return (order_==0||order_==3);};
    
    /**
    * <p>If the node returned by the most recent call to
    * {@link #next()} or {@link #nextFSN()} is a branch, this method returns
    * true if it has to be 
    * processed now in a postorder iteration. If the node returned is a leaf,
    * this method always returns true.</p>
    * <p>If {@link #next()} or {@link #nextFSN()} was not called prior to
    * calling this method, its return value is unspecified.</p>
    */
    public boolean postorder() {return (order_==2||order_==3);};
    
    /**
    * <p>If the node returned by the most recent call to
    * {@link #next()} or {@link #nextFSN()} is a branch, this method returns
    * true if it has to be 
    * processed now in an inorder iteration. If the node returned is a leaf,
    * this method always returns true.</p>
    * <p>If {@link #next()} or {@link #nextFSN()} was not called prior to
    * calling this method, its return value is unspecified.</p>
    */
    public boolean inorder() {return (order_==1||order_==3);};
    
    /** 
    * <p>Returns the next element in the iteration.</p>
    *
    * @return the next element in the iteration. 
    * @throws NoSuchElementException if the iteration has no more elements.
    * @see #nextFSN()
    */
    public Object next() {return nextFSN();};
    
    /** 
    * <p>Throws {@link UnsupportedOperationException} .</p>
    *
    * @throws UnsupportedOperationException
    */
    public void remove() {throw new UnsupportedOperationException();};
    
    private class stackElem
    {
      public int order;
      public FormulaSyntaxNet node;
      public int dirtyNumber;
      public stackElem(FormulaSyntaxNet n,int o,int num)
      {node=n; order=o; dirtyNumber=num;};
    };
  };
  
  
  
  /**
  * <p>Iterates over the {@link FormulaSyntaxNet}s pointed to by the
  * {@link FormulaSyntaxNet.ForwardConnection}s iterated over by an
  * {@link FormulaSyntaxNet.FConnIterator}</p>
  */
  protected static class FSNIteratorFromFConnIterator implements FSNIterator
  {
    private FConnIterator myFConnIterator_;
    
    /**
    * <p>Creates an iterator to iterate over the <code>FormulaSyntaxNet</code>s
    * pointed to by the {@link FormulaSyntaxNet.ForwardConnection}s iterated over by 
    * <code>i</code>.</p>
    *
    * <p>Note that operations on the iterator <code>i</code> affect the
    * <code>FSNIteratorFromIterator</code>.</p>
    */
    public FSNIteratorFromFConnIterator(FConnIterator i)
    {
      myFConnIterator_=i;
    };
    
    /** <p>Returns true if the iteration has more elements.</p> */
    public boolean hasNext() {return myFConnIterator_.hasNext();};
    
    /** 
    * <p>Returns the next element in the iteration.</p>
    * @return the next element in the iteration. 
    * @throws NoSuchElementException if the iteration has no more elements.
    * @see #next()
    */
    public FormulaSyntaxNet nextFSN() {return myFConnIterator_.nextFConn().to();};
    
    /** 
    * <p>Returns the next element in the iteration.</p>
    *
    * @return the next element in the iteration. 
    * @throws NoSuchElementException if the iteration has no more elements.
    * @see #nextFSN()
    */
    public Object next() {return nextFSN();};
    
    /** 
    * <p>Throws {@link UnsupportedOperationException} .</p>
    *
    * @throws UnsupportedOperationException
    */
    public void remove() {throw new UnsupportedOperationException();};
    
    /**
    * Always returns <code>true</code>.
    */
    public boolean inorder() {return true;};
    /**
    * Always returns <code>true</code>.
    */
    public boolean preorder() {return true;};
    /**
    * Always returns <code>true</code>.
    */
    public boolean postorder() {return true;};
  };
  
  /**
  * <p>Iterates over the {@link FormulaSyntaxNet}s iterated over by
  * an {@link java.util.Iterator Iterator}</p>
  */
  protected static class FSNIteratorFromIterator implements FSNIterator
  {
    private Iterator myIterator_;
    
    /**
    * <p>Creates an iterator to iterate over the <code>FormulaSyntaxNet</code>s
    * iterated over by <code>i</code>.</p>
    *
    * <p>Note that operations on the iterator <code>i</code> affect the
    * <code>FSNIteratorFromIterator</code>.</p>
    */
    public FSNIteratorFromIterator(Iterator i)
    {
      myIterator_=i;
    };
    
    /** <p>Returns true if the iteration has more elements.</p> */
    public boolean hasNext() {return myIterator_.hasNext();};
    
    /** 
    * <p>Returns the next element in the iteration.</p>
    * @return the next element in the iteration. 
    * @throws NoSuchElementException if the iteration has no more elements.
    * @see #next()
    */
    public FormulaSyntaxNet nextFSN() {return (FormulaSyntaxNet)myIterator_.next();};
    
    /** 
    * <p>Returns the next element in the iteration.</p>
    *
    * @return the next element in the iteration. 
    * @throws NoSuchElementException if the iteration has no more elements.
    * @see #nextFSN()
    */
    public Object next() {return nextFSN();};
    
    /** 
    * <p>Throws {@link UnsupportedOperationException} .</p>
    *
    * @throws UnsupportedOperationException
    */
    public void remove() {throw new UnsupportedOperationException();};
    
    /**
    * Always returns <code>true</code>.
    */
    public boolean inorder() {return true;};
    /**
    * Always returns <code>true</code>.
    */
    public boolean preorder() {return true;};
    /**
    * Always returns <code>true</code>.
    */
    public boolean postorder() {return true;};
  };
  
  
  /**
  * <p><code>InvariantViolation</code> is thrown when {@link #checkInvariant()}
  * detects an error.</p>
  */
  public static class InvariantViolation extends RuntimeException
  {
    /** Constructs an <code>InvariantViolation</code> with no detail message. */
    public InvariantViolation(){};
    /** Constructs an <code>InvariantViolation</code> with the specified 
    * detail message */
    public InvariantViolation(String s){super(s);};
  };

  
  /**
  * If <code>{@link Debugging#enabled}==true</code>, this function tests 
  * the <code>FormulaSyntaxNet</code> methods.
  * It is not meant to be called from application code.
  */
  public static void main(String args[])
  {
    if (Debugging.enabled)
    {
      System.out.println("Testing FormulaSyntaxNet");
      
      System.out.println("Creating ((A->B)->C) -> (D->(E->F))");
      
      String owner="owner";
      FormulaSyntaxNet A=newLeaf(owner,"A");
      FormulaSyntaxNet B=newLeaf(owner,"B");
      FormulaSyntaxNet C=newLeaf(owner,"C");
      FormulaSyntaxNet D=newLeaf(owner,"D");
      FormulaSyntaxNet E=newLeaf(owner,"E");
      FormulaSyntaxNet F=newLeaf(owner,"F");
      
      System.out.println("UIDs for A,B,C,D,E,F: "+A.uid()+", "+B.uid()+", "+C.uid()+", "+D.uid()+", "+E.uid()+", "+F.uid());
      
      FormulaSyntaxNet net=newBranch(owner, 
                             newBranch(owner, newBranch(owner,A,B), C),
                             newBranch(owner, D, newBranch(owner,E,F))
                           );
                           
      System.out.println("A.id()="+A.id());
      System.out.println("A.isBound()="+A.isBound()+"  "+
                         "A.isBranch()="+A.isBranch()+"  "+ 
                         "A.isBoundOrBranch()="+A.isBoundOrBranch()+"  "+
                         "A.isUnbound()="+A.isUnbound()+"  "+
                         "A.isForward()="+A.isForward()+"  "+
                         "A.isUnboundOrForward()="+A.isUnboundOrForward()
                         );
    
      System.out.println("infix string: "+net.toInfixFormulaString());
      
      final String[] orderStr={"in", "pre", "post"};
      FSNIterator itAr[]=new FSNIterator[3];
      itAr[0]=net.inorderIterator();
      itAr[1]=net.preorderIterator();
      itAr[2]=net.postorderIterator();
      for (int i=0; i<3; ++i)
      {
        System.out.print(orderStr[i]+"order walk: ");
        while (itAr[i].hasNext())
        {
          FormulaSyntaxNet n=itAr[i].nextFSN();
          if (n.isBranch()) System.out.print("->");
          else if (n.isBound()) System.out.print("("+n.id()+")");
          else if (n.isUnboundOrForward()) System.out.println("(unbound)");
          else throw new RuntimeException("Huh? What's this node?");
        };
        System.out.println();
      };  
      
      String nastyString1="(A->X0)->X0";
      String nastyString2="(((A))->(B))->(C)";
      FormulaSyntaxNet nasty1=newFromInfixFormulaString(owner,nastyString1);
      FormulaSyntaxNet nasty2=newFromInfixFormulaString(owner,nastyString2);
      
      System.out.println(nastyString1+" parsed to "+nasty1.toInfixFormulaString());
      System.out.println(nastyString2+" parsed to "+nasty2.toInfixFormulaString());
      System.out.print("unifying "+nastyString1+" with "+nastyString2+"...");
      try{
        unify(nasty1,nasty2);
        System.out.println("succeeds (This is BAD!)");
      }
      catch(ConflictingBindException x)
      {
        System.out.println("fails (This is GOOD!)");
      };
      
      System.out.println(nastyString1+" has become "+nasty1.toInfixFormulaString()+" (should be unchanged)");
      System.out.println(nastyString2+" has become "+nasty2.toInfixFormulaString()+" (should be unchanged)");
      
      FormulaSyntaxNet nasty3=newBranch(owner, newBranch(owner,A,B), newLeaf(owner,"B"));
      String nastyString3=nasty3.toInfixFormulaString();
      
      System.out.print("unifying "+nastyString1+" with "+nastyString3+"...");
      try{
        unify(nasty3,nasty1);
        System.out.println("succeeds (This is GOOD!)");
      }
      catch(ConflictingBindException x)
      {
        System.out.println("fails (This is BAD!)");
      };
      
      System.out.println(nastyString3+" has become "+nasty3.toInfixFormulaString()+" (should be unchanged)");
      System.out.println(nastyString1+" has become "+nasty1.toInfixFormulaString()+" (should be same as ^^^^^)");
      System.out.println("unifying last net with itself");
      unify(nasty1,nasty1);
      
      String testStr1="A->(X0->(B->X3))";
      String testStr2="X1->((C->D)->(B->X10))";
      FormulaSyntaxNet test1=newFromInfixFormulaString(owner,testStr1);
      FormulaSyntaxNet test2=newFromInfixFormulaString(owner,testStr2);
      System.out.println("T1:"+testStr1+" parsed to "+test1.toInfixFormulaString());
      System.out.println("T2:"+testStr2+" parsed to "+test2.toInfixFormulaString());
      unify(test1,test2);
      System.out.println("unify("+testStr1+","+testStr2+") == ("+test1.toInfixFormulaString()+","+test2.toInfixFormulaString()+")");
      
      String testStr3="X1->(X2->(B->C))";
      FormulaSyntaxNet test3=newFromInfixFormulaString(owner,testStr3);
      System.out.println("T3:"+testStr3+" parsed to "+test3.toInfixFormulaString());
      
      System.out.println("unifying T1 with T3");
      unify(test1,test3);
      System.out.println("T1 has become "+test1.toInfixFormulaString());
      System.out.println("T3 has become "+test3.toInfixFormulaString());
      System.out.println("T2 has become "+test2.toInfixFormulaString()+" (should be same as T1 and T3)");
      
      System.out.println("creating unbound leaves u1, u2, u3 and u4");
      FormulaSyntaxNet u1=newLeaf(owner,null);
      FormulaSyntaxNet u2=newLeaf(owner,null);
      FormulaSyntaxNet u3=newLeaf(owner,null);
      FormulaSyntaxNet u4=newLeaf(owner,null);
      System.out.println("unifying u1 and u2");
      unify(u1,u2);
      System.out.println("unifying u3 and u4");
      unify(u3,u4);
      System.out.println("unifying u1 and u3");
      unify(u1,u3);
      System.out.println("unifying u4 and A");
      unify(u4,A);
      System.out.println("u1, u2, u3, u4 should all be 'A' now: "+u1.toInfixFormulaString()+", "+u2.toInfixFormulaString()+", "+u3.toInfixFormulaString()+", "+u4.toInfixFormulaString());
      
      
      FormulaSyntaxNet[] cn=new FormulaSyntaxNet[10];
      cn[1]=newBranch(owner, newLeaf(owner,null),newLeaf(owner,null));
      cn[2]=newBranch(owner,
                                        newLeaf(owner, "A"),
                                        newBranch(owner,
                                                     newLeaf(owner,null),
                                                     newLeaf(owner,null)
                                                  )
                                     );
      cn[3]=newLeaf(owner,null);
      cn[7]=newLeaf(owner,null);
      cn[8]=newLeaf(owner,null);
      
      String owner2="owner2";
      cn[4]=newLeaf(owner2, null);
      cn[5]=newBranch(owner2,
                                        newLeaf(owner2, "A"),
                                        newBranch(owner2,
                                                     newLeaf(owner2,"B"),
                                                     newLeaf(owner2,"B")
                                                  )
                                     );
      cn[6]=newBranch(owner2,
                                        newLeaf(owner2, "A"),
                                        newBranch(owner2,
                                                     newLeaf(owner2,null),
                                                     newLeaf(owner2,"B")
                                                  )
                                     );
      
      for (int i=1; i<=8; ++i)
      {
        System.out.println("cn["+i+"]="+cn[i].toInfixFormulaString());
      };
      
      System.out.println("unify(cn[3],cn[7]);");
      unify(cn[3],cn[7]);
      System.out.println("unify(cn[8],cn[7]);");
      unify(cn[8],cn[7]);
      System.out.println("cn[3]="+cn[3].toInfixFormulaString());
      System.out.println("cn[7]="+cn[7].toInfixFormulaString());
      System.out.println("cn[8]="+cn[8].toInfixFormulaString());
      
      System.out.println("unify(cn[1].right(),cn[3])");
      unify(cn[1].right(),cn[3]);
      System.out.println("cn[1]="+cn[1].toInfixFormulaString());
      System.out.println("cn[3]="+cn[3].toInfixFormulaString());
      
      System.out.println("unify(cn[3],cn[2])");
      unify(cn[3],cn[2]);
      for (int i=1; i<=3; ++i)
      {
        System.out.println("cn["+i+"]="+cn[i].toInfixFormulaString());
      };
      
      System.out.println("unify(cn[1].left(),cn[2].right().left())");
      unify(cn[1].left(),cn[2].right().left());
      for (int i=1; i<=3; ++i)
      {
        System.out.println("cn["+i+"]="+cn[i].toInfixFormulaString());
      };
      
      System.out.println("unify(cn[4],cn[5])");
      unify(cn[4],cn[5]);
      for (int i=1; i<=5; ++i)
      {
        System.out.println("cn["+i+"]="+cn[i].toInfixFormulaString());
      };
      
      System.out.println("unify(cn[4],cn[6])");
      unify(cn[4],cn[6]);
      for (int i=1; i<=6; ++i)
      {
        System.out.println("cn["+i+"]="+cn[i].toInfixFormulaString());
      };
      
      Collection allNodesReachableFromCN4=new java.util.HashSet();
      cn[4].getAllReachableNodes(allNodesReachableFromCN4);
      System.out.println("number of elements reachable from cn[4]="+allNodesReachableFromCN4.size());
      Collection allNodesReachableFromCN1=new java.util.HashSet();
      cn[1].getAllReachableNodes(allNodesReachableFromCN1);
      System.out.println("number of elements reachable from cn[1]="+allNodesReachableFromCN1.size());
      
      System.out.println("unify(cn[4],cn[1].right())");
      unify(cn[4],cn[1].right());
      
      for (int i=1; i<=8; ++i)
      {
        System.out.println("cn["+i+"]="+cn[i].toInfixFormulaString());
      };
      
      Collection owners_owner2=new java.util.HashSet();
      owners_owner2.add(owner2);
      System.out.println("Splitting off cn[4],cn[5],cn[6]:");
      System.out.println("  extractByOwner(cn[4],owners_owner2)");
      extractByOwner(cn[4],owners_owner2);
      for (int i=1; i<=8; ++i)
      {
        System.out.println("cn["+i+"]="+cn[i].toInfixFormulaString());
      };
      
      Collection allNodesReachableFromCN4_2=new java.util.HashSet();
      cn[4].getAllReachableNodes(allNodesReachableFromCN4_2);
      System.out.println("number of elements reachable from cn[4]="+allNodesReachableFromCN4_2.size());
      Collection allNodesReachableFromCN1_2=new java.util.HashSet();
      cn[1].getAllReachableNodes(allNodesReachableFromCN1_2);
      System.out.println("number of elements reachable from cn[1]="+allNodesReachableFromCN1_2.size());
      
      if (
           !allNodesReachableFromCN4.equals(allNodesReachableFromCN4_2) ||
           !allNodesReachableFromCN1.equals(allNodesReachableFromCN1_2)
         ) 
         throw new RuntimeException("disentangling c[1] and c[4] FAILED!");
      
      
      System.out.print("unify(cn[2].right().right(),cn[1])...");
      try{
        unify(cn[2].right().right(),cn[1]);
        System.out.println("succeeds (This is BAD!)");
      }catch(ConflictingBindException x)
      {
        System.out.println("fails (This is GOOD!)");
      };
      
      
      System.out.println("removeAllForwards(cn[1])");
      removeAllForwards(cn[1]);
      for (int i=1; i<=8; ++i)
      {
        System.out.println("cn["+i+"]="+cn[i].toInfixFormulaString());
      };
      
      System.out.println();
      FormulaSyntaxNet sleaf=newLeaf(owner, null);
      FormulaSyntaxNet fakebranch=newBranch(owner, sleaf, sleaf);
      FormulaSyntaxNet weird=newBranch(owner, fakebranch, fakebranch);
      FormulaSyntaxNet apyra=newFromInfixFormulaString(owner,"(X0->A)->(X1->X0)");
      System.out.println("weird="+weird.toInfixFormulaString());
      System.out.println("apyra="+apyra.toInfixFormulaString());
      System.out.println("unify(apyra,weird)");
      unify(apyra,weird);
      System.out.println("weird="+weird.toInfixFormulaString());
      System.out.println("apyra="+apyra.toInfixFormulaString());
      
      String[] formulaStrings={"X0->X1","(X1->(X1->X2))","X2->X0"};
      FormulaSyntaxNet[] formulas=newFromInfixFormulaStrings(owner,formulaStrings);
      
      if (formulas.length!=formulaStrings.length) throw new RuntimeException("Oops!");
      
      for (int i=0; i<formulas.length; ++i)
      {
        System.out.println("formulaStrings["+i+"]="+formulaStrings[i]+" parsed to "+formulas[i].toInfixFormulaString());
      };
      
      System.out.println("unify(A,formulas[0].left())");
      unify(A,formulas[0].left());
      for (int i=0; i<formulas.length; ++i)
      {
        System.out.println("formulas["+i+"] has become "+formulas[i].toInfixFormulaString());
      };
      
      System.out.println("unify(B,formulas[2].left())");
      unify(B,formulas[2].left());
      for (int i=0; i<formulas.length; ++i)
      {
        System.out.println("formulas["+i+"] has become "+formulas[i].toInfixFormulaString());
      };
      
      System.out.println("unify(C,formulas[1].left())");
      unify(C,formulas[1].left());
      for (int i=0; i<formulas.length; ++i)
      {
        System.out.println("formulas["+i+"] has become "+formulas[i].toInfixFormulaString());
      };
      
    };
  };
};

