/* Copyright (C) 2002 Matthias S. Benkmann <m.s.b@gmx.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License (ONLY THIS VERSION).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package domino.logic;

import domino.core.Debugging;

/**
* <p>Formats proof for output.</p>
*/
public class ProofFormatter
{
  /**
  * <p>Spaces between 2 formulas on the same line.</p>
  */
  static final int SPACING=4;
  
  /**
  * <p>Converts <code>proof</code> into an array of strings to be printed
  * in ascending order to get a nice ASCII diagram.</p>
  */
  public static String[] naturalDeductionDiagram(FormulaDominoTree proof)
  {
    FormulaDominoTree.UnboundIdAssigner assigner=FormulaDominoTree.UnboundIdAssigner.newAssigner();
    return naturalDeductionDiagramD(proof,assigner).toStringArray();
  };
  
  /**
  * <p>Converts <code>proof</code> into an array of strings to be printed
  * in ascending order to get a nice ASCII diagram.</p>
  * <p><code>assigner</code> is used to assign names to unbound nodes.</p>
  */
  public static String[] naturalDeductionDiagram(FormulaDominoTree proof, FormulaDominoTree.UnboundIdAssigner assigner)
  {
    return naturalDeductionDiagramD(proof,assigner).toStringArray();
  };
    
  protected static Diagram naturalDeductionDiagramD(FormulaDominoTree proof, FormulaDominoTree.UnboundIdAssigner assigner)
  {
    FormulaDominoSquare root=(FormulaDominoSquare)proof.rootSquare();
    
    int nsub=proof.numberOfSubtreeConnections();
    
    String rootInfixRep=root.toInfixRepresentation(assigner);
    
    if (nsub==0)
    {
      return new FormulaDiagram("["+rootInfixRep+"]");
    };
    
    Diagram[] subs=new Diagram[nsub];
    String[][] bonusStrings=new String[nsub][];
    for (int i=0; i<nsub; ++i)
    {
      FormulaDominoTree subtree=(FormulaDominoTree)proof.subtree(i);
      if (subtree==null)
      {
        FormulaDominoSquare unconn=(FormulaDominoSquare)proof.subtreeConnectionSquare(i);
        subs[i]=new FormulaDiagram(unconn.toInfixRepresentation(assigner));
      } else
      {
        subs[i]=naturalDeductionDiagramD(subtree,assigner);
      };
      int nbon=proof.numberOfNewBonusSquaresForSubtreeConnection(i);
      bonusStrings[i]=new String[nbon];
      for (int j=0; j<nbon; ++j)
      {
        FormulaDominoSquare newbon=(FormulaDominoSquare)proof.newBonusSquareForSubtreeConnection(i,j);
        bonusStrings[i][j]="["+newbon.toInfixRepresentation(assigner)+"]";
      };
    };
    
    return new SeparatedDiagram(rootInfixRep,new SeparatorDiagram(subs,bonusStrings));
  };  
  
  
  protected abstract static class Diagram
  {
    protected Diagram parent_=null;
    
    protected abstract String[] subdiagramToStringArray();
    protected abstract String myLine();
    protected abstract int length();
    protected abstract int minLength();
    
    protected void setParent(Diagram parent) {parent_=parent;};
    
    public String[] toStringArray()
    {
      String[] sub=subdiagramToStringArray();
      String[] arr=new String[sub.length+1];
      if (sub.length>0) System.arraycopy(sub,0,arr,0,sub.length);
      arr[sub.length]=myLine();
      return arr;
    };
  };
  
  protected static class FormulaDiagram extends Diagram
  {
    private String formula_;
    public FormulaDiagram(String formula) 
    {
      formula_=formula;
    };
    
    protected String myLine() {return formula_;};
    protected int minLength() {return formula_.length();};
    protected int length() {return formula_.length();};
    protected String[] subdiagramToStringArray() {return new String[0];};
  };
 
  
  protected static class SeparatedDiagram extends Diagram
  {
    private String formula_;
    private Diagram separator_;
    private int lengthcache=-1;
    
    public SeparatedDiagram(String formula, Diagram separator)
    {
      formula_=formula;
      separator_=separator;
      separator.setParent(this);
    };
    
    protected int minLength() {return formula_.length();};
    
    protected int length()
    {
      if (lengthcache<0)
      {
        int l=separator_.length();
        if (l>minLength()) lengthcache=l; else lengthcache=minLength();
      };  
      return lengthcache;  
    };
    
    protected String myLine() {return formula_;};
    
    protected String[] subdiagramToStringArray() 
    {
      return separator_.toStringArray();
    };
  };
  
  protected static class SeparatorDiagram extends Diagram
  {
    private Diagram[] subdiagrams_;
    private int nsubs_;
    private boolean allHaveSameBonus_;
    private String[][] bonusStrings_;
    private int lengthcache=-1;
    private int separatorbarminlengthcache=-1;
    
    public SeparatorDiagram(Diagram[] subs,String[][] bonusStrings)
    {
      subdiagrams_=subs;
      nsubs_=subs.length;
      bonusStrings_=bonusStrings;
      
      allHaveSameBonus_=true;
      for (int i=1; i<nsubs_; ++i)
      {
        if (bonusStrings[0].length!=bonusStrings[i].length)
        {allHaveSameBonus_=false; break;};

        for (int j=0; j<bonusStrings[0].length; ++j)
        if (!bonusStrings[0][j].equals(bonusStrings[i][j]))
        {allHaveSameBonus_=false; break;};
      };
      
      for (int i=0; i<nsubs_; ++i) subs[i].setParent(this);
    };
    
    
    protected int separatorBarMinLength()
    {
      if (separatorbarminlengthcache<0)
      {
        int l=0;
        for (int i=0; i<nsubs_; ++i)
        {
          l+=subdiagrams_[i].length()+SPACING;
        };
        l-=SPACING; //no spacing after last
        
        int l2=parent_.minLength();
        if (l2>l) l=l2;
        separatorbarminlengthcache=l;
      };
      return separatorbarminlengthcache;
    };  
    
    
    protected int length() 
    {
      if (lengthcache<0)
      {
        int l=separatorBarMinLength();
      
        if (allHaveSameBonus_)
        {
          for (int i=0; i<bonusStrings_[0].length; ++i)
            l+=bonusStrings_[0][i].length();
        }
        else
        {
          throw new RuntimeException("not implemented");
        };
        lengthcache=l;
      };  
      
      return lengthcache;
    };
    
    protected int minLength()  {return length();};
    
    protected String myLine()
    {
      if (allHaveSameBonus_)  // ----------------[f1][f2]
      {
        int l=separatorBarMinLength();
        StringBuffer s=new StringBuffer(l);
        while(l-- > 0) s.append('-');
        
        for (int i=0; i<bonusStrings_[0].length; ++i) s.append(bonusStrings_[0][i]);

        return s.toString();
      }
      else  // ---[f11][f12]---[f21]---[f31][f32]---
      {
        throw new RuntimeException("not implemented");
      }
    };
    
    protected String[] subdiagramToStringArray() 
    {
      String[][] sub=new String[nsubs_][];
      int[] len=new int[nsubs_];
      
      for (int i=0; i<nsubs_; ++i)
      {
        sub[i]=subdiagrams_[i].toStringArray();
        len[i]=subdiagrams_[i].length()+SPACING;
      };
      len[nsubs_-1]-=SPACING; //no spacing after last
        
      int maxarraylen=sub[0].length;
      for (int i=1; i<nsubs_; ++i)
        if (sub[i].length>maxarraylen) maxarraylen=sub[i].length;
        
      String[] res=new String[maxarraylen];
      for (int line=0; line<maxarraylen; ++line)
      {
        StringBuffer sbuf=new StringBuffer();
        for (int i=0; i<nsubs_; ++i)
        {
          int space=len[i];
          int subline=line+sub[i].length-maxarraylen;
          if (subline>=0)
          {
            String s=sub[i][subline];
            sbuf.append(s);
            space-=s.length();
            if (Debugging.enabled)
            {
              if (space<0) throw new RuntimeException("subdiagrams_[i] has line that is longer than subdiagrams_[i].length()");
            };
          };
          while (space-->0) sbuf.append(' ');
        };
        res[line]=sbuf.toString();
      };
      return res;
    };
    
  };
};


