/* Copyright (C) 2002 Matthias S. Benkmann <m.s.b@gmx.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License (ONLY THIS VERSION).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package domino.logic;

import java.util.Vector;
import java.util.List;
import java.util.Random;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import domino.logic.TautologyGenerator;
import domino.logic.ProofFormatter;

/**
* <p>This class is a test program for testing {@link FormulaDominoTree},
* {@link TautologyGenerator} and related classes.</p>
*/
public class TextModeProver
{
  protected static DeductionRule[] rules={new TNDRule(),new PERule(),new MPRule(),new CloseRule()};
  
  public static void main(String[] args)
  {
    Vector prob=new Vector();
    prob.add(new Double(1)); //TND
    prob.add(new Double(3)); //->I
    prob.add(new Double(6)); //->E
    prob.add(new Double(5)); //[assumpt]
    Random rand0=new Random();
    long seed=rand0.nextLong();
    System.out.println("Random seed: "+seed);
    rand0=new Random(seed);
    System.out.print("Generating tautology to prove...");
    String tautRepresentation=((FormulaDominoSquare)new TautologyGenerator(rand0).randomTautologyProof(3,prob,true).rootSquare()).toInfixRepresentation();
    System.out.println(tautRepresentation);
    System.out.println();
    
    String[] leaves=new String[1];
    leaves[0]=tautRepresentation;
    FormulaDominoTree tree=FormulaDominoTree.newFromInfix(tautRepresentation,leaves,null);
    
    BufferedReader in=new BufferedReader(new InputStreamReader(System.in));
    
    FormulaDominoTree.UnboundIdAssigner assigner=
                                   FormulaDominoTree.UnboundIdAssigner.newAssigner();
    Vector option=new Vector(); //format of option: Integer rule, Integer leaf, Vector bsIdx
    while (tree.numberOfOpenLeaves()>0)
    {
      option.clear();
      assigner.reset();
      
      for (int leaf=0; leaf<tree.numberOfLeaves(); ++leaf)
      {
        if (!tree.isOpenLeaf(leaf)) continue;
        System.out.println();
        System.out.print("Goal: "+((FormulaDominoSquare)tree.leafSquare(leaf)).toInfixRepresentation(assigner)+"   ");
        int bnum=tree.numberOfBonusSquares(leaf);
        for (int j=0; j<bnum; ++j) 
          System.out.print("["+((FormulaDominoSquare)tree.bonusSquare(leaf,j)).toInfixRepresentation(assigner)+"] ");
        System.out.println();
        
        for (int rule=0; rule<rules.length; ++rule)
        {
          int needBS=rules[rule].requiredBonusSquares();
          if (needBS>1) throw new RuntimeException("Unsupported rule");
          int bs=0;
          Vector bsIdx=new Vector();
          do
          {
            bsIdx.clear();
            if (needBS>0) bsIdx.add(new Integer(bs));
            
            try{
              DominoTree t=rules[rule].applyTo(tree, leaf, bsIdx);
              option.add(new Integer(rule));
              option.add(new Integer(leaf));
              option.add(bsIdx);
              bsIdx=new Vector(); //so that we don't overwrite the Vector in option
              System.out.println(option.size()/3+". "+rules[rule].name(((FormulaDominoSquare)t.rootSquare()).toInfixRepresentation(assigner)));
              if (t.numberOfOpenLeaves()==0) System.out.println("\t<done>");
              else
              {
                for (int i=0; i<t.numberOfLeaves(); ++i)
                {
                  if (!t.isOpenLeaf(i)) continue;
                  
                  System.out.print("\t"+((FormulaDominoSquare)t.leafSquare(i)).toInfixRepresentation(assigner)+"   ");
                  bnum=t.numberOfBonusSquares(i);
                  for (int j=0; j<bnum; ++j) 
                                System.out.print("["+((FormulaDominoSquare)t.bonusSquare(i,j)).toInfixRepresentation(assigner)+"] ");
                  System.out.println();
                };
              };
              t.detachFromParent();
            }
            catch(ConflictingBindException x){};
            
            ++bs;
            
          }while(needBS>0 && bs<tree.numberOfBonusSquares(leaf));
        };
      };

      if (option.isEmpty()) throw new RuntimeException("No applicable rules!");
      
      while(true)
      {
        System.out.println();
        System.out.print("> ");
        try{
          int i=Integer.parseInt(in.readLine());
          i=(i-1)*3;
          if (i<0 || i>=option.size()) throw new NumberFormatException();
          int rule=((Integer)option.get(i)).intValue();
          int leaf=((Integer)option.get(i+1)).intValue();
          Vector bsIdx=(Vector)option.get(i+2);
          rules[rule].applyTo(tree, leaf, bsIdx);
          System.out.println();
          break;
        }
        catch(NumberFormatException x)
        {
          System.err.println("Illegal input!");
        }
        catch(IOException x)
        {
          System.err.println("Read error!");
          System.exit(1);
        };
      };  
    };
    
    String[] diagram=ProofFormatter.naturalDeductionDiagram(tree);
    System.out.println();
    for (int i=0; i<diagram.length; ++i) System.out.println(diagram[i]);
    System.out.println();
  };
  
  
  /**
  * <p>Classes that implement this interface are rules that can be applied to
  * a {@link DominoTree}'s leaf.</p>
  */
  protected interface DeductionRule
  {
    /**
    * <p>The number of bonus squares required to apply this rule.</p>
    */
    public int requiredBonusSquares();
    
    /**
    * <p>Applies the rule to <code>leaf</code> of <code>tree</code> using the
    * bonus squares whose indices are in <code>bsIdx</code>.</p>
    * <p><code>bsIdx</code> must contain {@link #requiredBonusSquares()} elements.</p>
    *
    * @return the new subtree attached to <code>leaf</code> if the rule was applied
    *         successfully. You can undo the application of the rule by calling
    *         this tree's {@link DominoTree#detachFromParent()} function.
    *
    * @throws ConflictingBindException if the rule is not applicable. The tree
    *         remains unchanged.
    * @throws SquareDoesNotExistException if <code>leaf</code> is not a valid leaf 
    *         number. The tree remains unchanged.
    * @throws ClassCastException if <code>this</code> rule can not be applied to
    *         <code>tree</code> because of an unsupported {@link DominoTree} subtype.
    */
    public DominoTree applyTo(DominoTree tree, int leaf, Vector bsIdx);
    
    /**
    * <p>Returns the name of the rule if the conclusion is <code>conclusion</code>.</p>
    */
    public String name(String conclusion);
  };
  
  protected static class CloseRule implements DeductionRule
  {
    public CloseRule() {};
    
    /**
    * <p>Returns 1.</p>
    */
    public int requiredBonusSquares() {return 1;};
    
    public DominoTree applyTo(DominoTree tree, int leaf, Vector bsIdx)
    {
      try{
        return tree.useBonusSquare(leaf,((Integer)bsIdx.get(0)).intValue());
      }
      catch(SquareDoesNotExistException X)
      {
        throw X;
      }
      catch(IllegalNodeTypeException X)
      {
        throw X;
      }
      catch(ConflictingBindException X)
      {
        throw X;
      }
      catch(IllegalArgumentException X)
      {
        throw new ConflictingBindException();
      }
    };
    
    public String name(String conclusion) {return "["+conclusion+"]";};
  };
  
  
  protected static class DominoTreeRule implements DeductionRule
  {
    private DominoTree myTree_;
    
    /**
    * <p><code>t</code> is copied so changing <code>t</code> after calling this
    * function has no effect on the <code>DominoTreeRule</code> object.</p>
    */
    public DominoTreeRule(DominoTree t)
    {
      init(t);
    };
    
    
    protected DominoTreeRule() {};
    
    protected void init(DominoTree t)
    {
      myTree_=t.cloneDT();
    };
    
    /**
    * <p>Returns 0.</p>
    */
    public int requiredBonusSquares() {return 0;};
    
    /**
    * @param bsIdx is not used.
    */
    public DominoTree applyTo(DominoTree tree, int leaf, Vector bsIdx)
    {
      DominoTree newTree=myTree_.cloneDT();
      newTree.attachTo(tree,leaf);
      return newTree;
    };
    
    public String name(String conclusion) {throw new UnsupportedOperationException("Can only be called for derived classes");};
  };
  
  protected static class TNDRule extends DominoTreeRule
  {
    public TNDRule()
    {
      String[] leaves={"_|_"};
      String[] bonus={"X0->_|_"};
      init(FormulaDominoTree.newFromInfix("X0",leaves,bonus));
    };
    
    public String name(String conclusion) {return "TND applied to "+conclusion;};
  };
  
  protected static class PERule extends DominoTreeRule
  {
    public PERule()
    {
      String[] leaves={"X1"};
      String[] bonus={"X0"};
      init(FormulaDominoTree.newFromInfix("X0->X1",leaves,bonus));
    };
    
    public String name(String conclusion) {return "->Introduction applied to "+conclusion;};
  };
  
  protected static class MPRule extends DominoTreeRule
  {
    public MPRule()
    {
      String[] leaves={"X0->X1","X0"};
      init(FormulaDominoTree.newFromInfix("X1",leaves,null));
    };
    
    public String name(String conclusion) {return "->Eliminiation applied to "+conclusion;};
  };
};

