/* Copyright (C) 2002 Matthias S. Benkmann <m.s.b@gmx.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License (ONLY THIS VERSION).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package domino.ui;

import java.util.Collection;
import java.util.LinkedList;
import java.util.Vector;
import java.util.Iterator;
import javax.swing.*;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Insets;
import javax.swing.SwingUtilities; //to help javadoc

import domino.logic.*;
import domino.core.Debugging;

/**
* <p>Representation of a {@link domino.logic.FormulaDominoTree} in an {@link IsoWorld}.</p>
*/
public class DominoTile implements IsoObject
{
  /**
  * <p>Used for consistent labelling of all <code>DominoTile</code>s.
  * After a <code>DominoTile</code> is removed from the world,
  * {@link #removeIdAssignments()} needs to be called to free the labels used
  * for that tile.</p>
  */
  protected static FormulaDominoTree.UnboundIdAssigner globalUnboundIdAssigner=FormulaDominoTree.UnboundIdAssigner.newAssigner();
  
  /** 
  * <p>Contains all {@link IsoBlock}s this tile consists of.</p> 
  * <p>The first entry is always the root block (which need not always be
  * a {@link RootDominoBlock}).</p>
  */
  private Vector myBlocks_;
  
  /** <p>If <code>this DominoTile</code> is connected to another, then this
  * variable is a reference to the connected tile.</p>
  */
  private DominoTile rootConnection_=null;
  /** 
  * <p><code>rootConnection_.leafConnections_[rootConnectionLeafIndex_]==this</code></p>
  */
  private int rootConnectionLeafIndex_=-1;
  /**
  * <p>For each block of <code>this</code> tile where another <code>DominoTile</code>
  * can be connected, there is one entry in this array, pointing to the connected
  * tile.</p>
  */
  private DominoTile[] leafConnections_;
  
  /**
  * <p>The {@link FormulaDominoTree}, <code>this DominoTile</code> represents.</p>
  */
  private FormulaDominoTree myTree_;
  
  /**
  * <p>Location of the reference block within the {@link IsoWorld}.</p>
  */
  private IsoLocation refBlockLocation_=null;
  
  /**
  * <p>Shift value to be added to the inividual shifts of all blocks of this tile.</p>
  */  
  private IsoShift shift_=new IsoShift(0,0,0);
  
  /**
  * <p>The {@link IsoWorld} this tile is in.</p>
  */
  private IsoWorld myWorld_=null;
  
  /**
  * <p>Determines if the {@link DominoTile.DominoBlock}s that form <code>this</code> tile return
  * <code>true</code> for {@link DominoTile.DominoBlock#isMassive()}.</p>
  */
  private boolean massive_=false;
  
  /**
  * <p>0: root block points north, 1: root block points east, 
  * 2: root block points south, 3 root block points west.</p>
  * <p><code>orientation_</code> does not affect shift!</p>
  */
  private int orientation_=0;
  
  /** 
  * <p>After a <code>DominoTile</code> is removed from the world, this method
  * has to be called to free the label colors used for that tile.</p>
  * <p>See {@link domino.logic.FormulaDominoTree.UnboundIdAssigner#removeMappingsStrictlyFor(FormulaDominoTree)}
  * for more information.</p>
  */
  public void removeIdAssignments()
  {
    if (rootTree()!=null)
      globalUnboundIdAssigner.removeMappingsStrictlyFor(rootTree());
  };
  
  
  /**
  * <p>Frees the label colors that are not used because they are superceded by
  * another color due to unification.</p>
  */
  public void removeUnneededIdAssignments()
  {
    globalUnboundIdAssigner.removeUnneededMappingsFor(myTree_);
  };
  
  
  /** 
  * <p>Removes all id assignments for all <code>DominoTile</code>s.</p> 
  * <p>Note that the tiles will not be redrawn.</p>
  */
  public static void removeIdAssignmentsForAllTiles()
  {
    globalUnboundIdAssigner.reset();
  };
  
  /**
  * <p>Attaches <code>this</code> tile to <code>t</code>, connecting with
  * leaf block <code>leafIndex</code>.</p>
  * <p>This method does NOT perform any unifications/connections of the
  * underlying {@link FormulaDominoTree}s.</p>
  *
  * @throws IllegalArgumentException if <code>this</code> is already attached to
  *         a <code>DominoTile</code> or if <code>leafIndex</code> is not a valid
  *         index.
  *
  * @see #detach()
  * @see #attachTo(DominoTile.SubtreeConnectionDominoBlock)
  */
  public void attachTo(DominoTile t,int leafIndex)
  {
    if (leafIndex<0 || leafIndex>=t.leafConnections_.length || 
        t.leafConnections_[leafIndex]!=null) throw new IllegalArgumentException();
    t.leafConnections_[leafIndex]=this;
    rootConnection_=t;
    rootConnectionLeafIndex_=leafIndex;
  };
  
  
  /**
  * <p>Attaches <code>this</code> tile to <code>block</code>.</p>
  * <p>This method does NOT perform any unifications/connections of the
  * underlying {@link FormulaDominoTree}s.</p>
  *
  * @throws IllegalArgumentException if <code>this</code> is already attached to
  *         a <code>DominoTile</code> or if <code>leafIndex</code> is not a valid
  *         index.
  *
  * @see #detach()
  */
  public void attachTo(DominoTile.SubtreeConnectionDominoBlock block)
  {
    DominoTile t=(DominoTile)block.parentObject();
    Iterator iter=t.blocksIterator();
    int i=0;
    while(true)
    {
      IsoBlock b=(IsoBlock)iter.next();
      if (b instanceof SubtreeConnectionDominoBlock)
      { if (b==block) break; else ++i; };
    };
    this.attachTo(t,i);
  };
  
  
  /**
  * <p>Detaches <code>this</code> tile from the <code>DominoTile</code> it is
  * attached to (if there is one).</p>
  * <p>This method does NOT perform any deunificating/detaching of the
  * underlying {@link FormulaDominoTree}s.</p>
  *
  * @see #attachTo(DominoTile,int)
  */
  public void detach()
  {
    if (rootConnection_!=null) 
    {
      rootConnection_.leafConnections_[rootConnectionLeafIndex_]=null;
      rootConnection_=null;
    };  
  };


  /**
  * <p>0: root block points north, 1: root block points east, 
  * 2: root block points south, 3 root block points west.</p>
  * <p>The orientation does not affect shift!</p>
  *
  * @see #setOrientation(int)
  */
  public int getOrientation() //TESTED
  {
    return orientation_;
  };
  
  /**
  * <p>0: root block points north, 1: root block points east, 
  * 2: root block points south, 3 root block points west.</p>
  * <p>The orientation does not affect shift!</p>
  * <p>If <code>this DominoTile</code> is already part of an {@link IsoWorld},
  * the new orientation will only be set if it does not cause a collision.</p>
  * <p>If the orientation is changed, the object is removed from and added again to
  * the {@link IsoWorld} it is part of (if any). This means that the 
  * <code>DominoTile</code>'s layer will be reset and 
  * {@link IsoWorld#putInTopLayer(IsoObject)} may need to be called again.</p>
  *
  * @param newOrientation is ANDed with 3 to get the new orientation.
  *
  * @see #getOrientation()
  */
  public void setOrientation(int newOrientation) //TESTED
  {
    newOrientation=newOrientation&3;
    if (newOrientation==orientation_) return;
    int oldOrientation=orientation_;
    orientation_=newOrientation;
    if (myWorld_!=null)
    {
      IsoLocation loc=getReferenceBlockLocation();
      IsoShift sh=getShift();
      
      //The following is a bit unclean: We test whether the new orientation is
      //permissible by calling IsoWorld.collision() for this with the new orientation
      //already set. In theory this could break horribly because the docs to 
      //IsoWorld.addObject() say that after being added to the IsoWorld, the
      //object must not change its shape. However, being the implementor I know that
      //collision() does not break and its unlikely it will ever be changed in a way
      //that the following would stop working.
      if (myWorld_.collision(this,loc,sh))
      {
        orientation_=oldOrientation;
      }
      else
      {
        //must restore orientation before calling removeObject() because removeObject()
        //WILL break if the object being removed has changed shape.
        orientation_=oldOrientation;
        //need to save reference to world because it will be nulled
        IsoWorld myOldWorld=myWorld_;
        myWorld_.removeObject(this);
        orientation_=newOrientation;
        myOldWorld.addObjectNoCollision(this,loc,sh);
      };
    };  
  };


  /**
  * <p>Puts the root block of the tile into inactive state so that no other tiles
  * can be connected to it.</p>
  */
  public void deactivateRootBlock()
  {
    ((DominoBlock)myBlocks_.get(0)).deactivate();
  };


  /**
  * <p>Returns the {@link FormulaDominoTree} this tile represents.</p>
  * <p>Note that the return value will be <code>null</code> if the tile represents a
  * bonus square that is not currently attached to a tree.</p>
  */
  public FormulaDominoTree rootTree()
  {
    DominoBlock rootBlock=rootBlock();
    if (rootBlock instanceof BonusDominoBlock)
    {
      return ((BonusDominoBlock)rootBlock).getBonusTree();
    }
    else return myTree_;
  };


  /**
  * <p>Returns the {@link DominoTile.DominoBlock} that can be connected to
  * a leaf of another <code>DominoBlock</code>.</p>
  */
  public DominoBlock rootBlock()
  {
    return (DominoBlock)myBlocks_.get(0);
  };
  
  /**
  * <p>Returns a string representation of this tile.</p>
  */
  public String toStringRepresentation()
  {
    if (rootBlock() instanceof BonusDominoBlock)
    {
      BonusDominoBlock b=(BonusDominoBlock)rootBlock();
      return ((FormulaDominoSquare)b.formulaDominoTree().newBonusSquareForSubtreeConnection(b.subtreeConnectionIndex(),b.newBonusForSubtreeConnectionIndex())).toInfixRepresentation();
    }
    else
    {
      StringBuffer rep=new StringBuffer();
      recursiveTreeToString(rootTree(),rep,FormulaDominoTree.UnboundIdAssigner.newAssigner());
      return rep.toString();
    }
  };
  
  /**
  * <p>Appends a string representation of <code>tree</code> to <code>rep</code>.</p>
  */
  protected void recursiveTreeToString(FormulaDominoTree tree,StringBuffer rep,FormulaDominoTree.UnboundIdAssigner assigner)
  {
    if (tree==null) 
      rep.append(";  ");
    else  
    {
      rep.append(((FormulaDominoSquare)tree.rootSquare()).toInfixRepresentation(assigner));
      /*
        NOTE: The following code assumes that all subtree connections of a
        FormulaDominoTree have the same bonus squares. This is true for the
        current implementation but is not specified in the interface. While it
        would be possible to write this function in a general way, this would
        break UITest.java:TileProducerManager.addProducerForStringRepresentation()
        because it does not support per-branch bonuses.
      */
      if (tree.numberOfSubtreeConnections()>0)
      {
        for (int i=0; i<tree.numberOfNewBonusSquaresForSubtreeConnection(0); ++i)
        {
          rep.append(" [");
          rep.append(((FormulaDominoSquare)tree.newBonusSquareForSubtreeConnection(0,i)).toInfixRepresentation(assigner));
          rep.append(']');
        };
  
        for (int i=0; i<tree.numberOfSubtreeConnections(); ++i)
        {
          rep.append(", ");
          rep.append(((FormulaDominoSquare)tree.subtreeConnectionSquare(i)).toInfixRepresentation(assigner));
        };
      };  
  
      rep.append(";  ");
  
      for (int i=0; i<tree.numberOfSubtreeConnections(); ++i)
      {
        recursiveTreeToString((FormulaDominoTree)tree.subtree(i),rep,assigner);
      };
    };
  };
  
  /**
  * <p>Returns the number of leaves <code>this</code> tile has for other tiles to
  * connect to.</p>
  */
  public int numberOfTileConnections()
  {
    return leafConnections_.length;
  };
  
  /**
  * <p>If a tile is connected to leaf <code>i</code>, it is returned, otherwise
  * the result is <code>null</code>.</p>
  * <p>Constraints for <code>i</code>: 
  * <code>0&lt;=i&lt;numberOfTileConnections()</code>.</p>
  *
  * @throws IllegalArgumentException if <code>i</code> is not a valid index.
  */
  public DominoTile tileConnection(int i)
  {
    if (i<0 || i>=numberOfTileConnections()) throw new IllegalArgumentException();
    return leafConnections_[i];
  };


  /**
  * <p>If <code>this</code> tile is connected to another tile, that tile is returned,
  * otherwise the result is <code>null</code>.</p>
  */
  public DominoTile connectedToTile()
  {
    return rootConnection_;
  };
  
  /**
  * <p>Creates a <code>DominoTile</code> to represent the bonus square with number
  * <code>newBonusForSubtreeIdx</code> that <code>tree</code> adds to the bonus squares that may
  * be used in the subtree of <code>tree</code> with index <code>subtreeIdx</code>.</p>
  * <p><code>tree</code> is not copied, it is linked to the <code>DominoTile</code>.
  * However, to make the <code>DominoTile</code> reflect changes made to the tree,
  * {@link #updateFormulas()} has to be called.</p>
  */
  public DominoTile(FormulaDominoTree tree, int subtreeIdx, int newBonusForSubtreeIdx)
  {
    myTree_=tree;
    leafConnections_=new DominoTile[0];
    myBlocks_=new Vector(1);
    IsoLocation location=new IsoLocation(0,0,0);
    myBlocks_.add(new BonusDominoBlock(location,true,tree,subtreeIdx,newBonusForSubtreeIdx));
  };


  /**
  * <p>A reference to a leaf (more precisely: a subtree connection) of a 
  * {@link FormulaDominoTree}.</p>
  */
  protected static class LeafRef
  {
    /** <p>The tree that owns the subtree connection.</p>*/
    public FormulaDominoTree owner;
    /** <p>The index of the subtree connection.</p> */
    public int index;
    /** 
    * <p>A {@link Collection} of {@link DominoTile.BonusRef}s with references to all bonus 
    * that may be attached to this leaf.</p>
    */
    public Collection bonus;
  };
    
    
  /**
  * <p>Reference to a bonus square in a {@link FormulaDominoTree}.</p>
  */  
  protected static class BonusRef
  {
    /** <p>The tree that owns the bonus square.</p> */
    public FormulaDominoTree owner;
    /** <p>The index of the subtree connection for which the bonus square is valid.</p> */
    public int subtreeIndex;
    /** <p>The index of the bonus square in the list of bonus squares valid for
    * the respective subtree connection.</p> */
    public int newBonusForSubtreeIndex;
    /** <p>The formula of the bonus square.</p> */
    public String formula;
  };
  
  
  /**
  * <p>Returns a {@link Collection} that contains a {@link DominoTile.LeafRef} for every
  * open leaf in <code>tree</code>.</p>
  *
  * @param tree the tree whose leaves to collect.
  * @param inheritedBonus pass a new {@link LinkedList}.
  * @param tempAssigner pass a new {@link domino.logic.FormulaDominoTree.UnboundIdAssigner}.
  */
  protected static Collection collectLeafRefs(FormulaDominoTree tree,Collection inheritedBonus, FormulaDominoTree.UnboundIdAssigner tempAssigner)
  { //TESTED
    Collection leafRefs=new LinkedList();

    for (int i=0; i<tree.numberOfSubtreeConnections(); ++i)
    {
      Collection accumulatedBonus=new LinkedList(inheritedBonus);

addNewBonusRefs:
      for (int j=0; j<tree.numberOfNewBonusSquaresForSubtreeConnection(i); ++j)
      {
        FormulaDominoSquare b=(FormulaDominoSquare)tree.newBonusSquareForSubtreeConnection(i,j);
        String bonusFormula=b.toPrefixRepresentation(tempAssigner);
        Iterator iter=accumulatedBonus.iterator();
        while (iter.hasNext())
        {
          if ( ((BonusRef)iter.next()).formula.equals(bonusFormula))
            continue addNewBonusRefs;
        };
        
        BonusRef bonusRef=new BonusRef();
        bonusRef.formula=bonusFormula;
        bonusRef.owner=tree;
        bonusRef.subtreeIndex=i;
        bonusRef.newBonusForSubtreeIndex=j;
        accumulatedBonus.add(bonusRef);
      };
      
      FormulaDominoTree sub=(FormulaDominoTree)tree.subtree(i);
      if (sub==null)
      {
        LeafRef leafRef=new LeafRef();
        leafRef.owner=tree;
        leafRef.index=i;
        leafRef.bonus=new Vector(accumulatedBonus);
        leafRefs.add(leafRef);
      }
      else
      {
        leafRefs.addAll(collectLeafRefs(sub,accumulatedBonus,tempAssigner));
      };
    };  
      
    return leafRefs;
  };
 
  /**
  * <p>Removes the {@link BonusRef}s common to all {@link LeafRef}s from
  * <code>leafRefs</code> and stores them in <code>commonBonus</code>.</p>
  */
  protected void extractCommonBonus(Collection leafRefs, Collection commonBonus)
  {
    if (leafRefs.size()==0) return;
    //if (leafRefs.size()==1) return; //activate this line to get back the longish tiles
    LeafRef lRef=(LeafRef)leafRefs.iterator().next();
    Iterator bRefIter=lRef.bonus.iterator();

moveAllCommonBonusRefsToCommonBonus:
    while(bRefIter.hasNext())
    {
      BonusRef bRef=(BonusRef)bRefIter.next();
      Iterator iter=leafRefs.iterator();
      iter.next(); //skip lRef

checkIfBrefPresentInAllLeafRefs:
      while(iter.hasNext())
      {
        LeafRef lRef2=(LeafRef)iter.next();
        Iterator bRefIter2=lRef2.bonus.iterator();
        while(bRefIter2.hasNext())
        {
          if (((BonusRef)bRefIter2.next()).formula.equals(bRef.formula))
          continue checkIfBrefPresentInAllLeafRefs;
        };
        continue moveAllCommonBonusRefsToCommonBonus;
      };
      
      commonBonus.add(bRef);
      bRefIter.remove(); //do this separately, not in the loop below, otherwise bRefIter would be invalidated
      iter=leafRefs.iterator();
      iter.next(); //skip lRef
      while(iter.hasNext())
      {
        LeafRef lRef2=(LeafRef)iter.next();
        Iterator bRefIter2=lRef2.bonus.iterator();
        while(bRefIter2.hasNext())
        {
          if (((BonusRef)bRefIter2.next()).formula.equals(bRef.formula))
            bRefIter2.remove();
        };
      };
    };
  };

  /**
  * <p>Creates a <code>DominoTile</code> to represent <code>tree</code>.</p>
  * <p><code>tree</code> is not copied, it is linked to the <code>DominoTile</code>.
  * However, to make the <code>DominoTile</code> reflect changes made to the tree,
  * {@link #updateFormulas()} has to be called.</p>
  */
  public DominoTile(FormulaDominoTree tree) //TESTED
  {
    myTree_=tree;
    
    Collection leafRefs=collectLeafRefs(tree,new LinkedList(),FormulaDominoTree.UnboundIdAssigner.newAssigner());
    leafConnections_=new DominoTile[leafRefs.size()];
    
    //move all BonusRefs that occur in all LeafRefs to commonBonus
    Collection commonBonus=new LinkedList();  //Collection of BonusRefs
    extractCommonBonus(leafRefs,commonBonus);
    
    /* 
      now we have to construct a DominoTile from the following:
    
      1) 1x RootDominoBlock  (tree root, active)
      2) commonBonus.size() x BonusDominoBlock  (inactive)
      3) commonBonus.size() x BonusDominoBlock  (active, on top of blocks from 2)
      4) leafRefs.size() x a series of BonusDominoBlocks (inactive)
      5) leafRefs.size() x a series of BonusDominoBlocks (active, on top of 4)
      6) leafRefs.size() x SubtreeConnectionDominoBlock (active)
      7) n x EmptyDominoBlock (inactive, filler)
      
      NOTE: It is possible that leafRefs.isEmpty(), i.e. there are no open leaves.
    */
    
    int numLeafBonusBlocks=0; //number of bonus blocks that are not common to all leaves
    Iterator iter=leafRefs.iterator();
    while (iter.hasNext()) numLeafBonusBlocks+=((LeafRef)iter.next()).bonus.size();
    
    int numLeaves=leafRefs.size();
    int fillerNum=0;
    
    if (numLeaves>=4) fillerNum=1+(numLeaves-3)*2;
    if (commonBonus.size()==0 && numLeaves>1 && fillerNum==0) fillerNum=1;
    
    //if there is at least one non-common bonus block, at least 1 filler is needed so
    //that the non-common bonus block can not be mistaken for a common bonus block
    if (numLeafBonusBlocks!=0 && fillerNum==0) fillerNum=1;
        
    int numBlocks=1 + 2*commonBonus.size() + 2*numLeafBonusBlocks + leafRefs.size() + fillerNum;
    
    myBlocks_=new Vector(numBlocks);
    
    IsoLocation location=new IsoLocation(0,0,0);
    myBlocks_.add(new RootDominoBlock(location,true,tree));
    ++location.y;
    
    iter=commonBonus.iterator();
    while(iter.hasNext())
    {
      BonusRef bRef=(BonusRef)iter.next();
      //add inactive bonus block
      myBlocks_.add(new BonusDominoBlock(location,false,bRef.owner,bRef.subtreeIndex,bRef.newBonusForSubtreeIndex));
      //add active bonus block on top
      ++location.z;
      myBlocks_.add(new BonusDominoBlock(location,true,bRef.owner,bRef.subtreeIndex,bRef.newBonusForSubtreeIndex));
      --location.z;
      ++location.y;
    };
    
    //leftx is coordinate left of left-most filler
    int leftx=location.x - (fillerNum/2) - 1;
    //rightx is coordinate right of right-most filler
    int rightx=leftx + (fillerNum==0? 1: fillerNum) + 1;
    
    if (fillerNum>0)
    {
      location.x=leftx;
      for (int i=0; i<fillerNum; ++i)
      {
        ++location.x;
        myBlocks_.add(new EmptyDominoBlock(location));
      };
      ++location.y;
    };
    
    if (numLeaves>=2)
    {
      --location.y;
      iter=leafRefs.iterator();
      location.x=leftx;
      int sign=-1;
      for (int i=0; i<2; ++i)
      {
        LeafRef lRef=(LeafRef)iter.next();
        iter.remove();
        Iterator bonusIter=lRef.bonus.iterator();
        while (bonusIter.hasNext())
        {
          BonusRef bRef=(BonusRef)bonusIter.next();
          //add inactive bonus block
          myBlocks_.add(new BonusDominoBlock(location,false,bRef.owner,bRef.subtreeIndex,bRef.newBonusForSubtreeIndex));
          //add active bonus block on top
          ++location.z;
          myBlocks_.add(new BonusDominoBlock(location,true,bRef.owner,bRef.subtreeIndex,bRef.newBonusForSubtreeIndex));
          --location.z;
          location.x+=sign;
        };
        myBlocks_.add(new SubtreeConnectionDominoBlock(location,true,lRef.owner,lRef.index));
        location.x=rightx;
        sign=1;
      };
      numLeaves-=2;
      ++location.y;
    };
    
    if (numLeaves>0)
    {
      location.x=leftx+1;
      int ystart=location.y;
      iter=leafRefs.iterator();
      while(iter.hasNext())
      {
        LeafRef lRef=(LeafRef)iter.next();
        Iterator bonusIter=lRef.bonus.iterator();
        while (bonusIter.hasNext())
        {
          BonusRef bRef=(BonusRef)bonusIter.next();
          //add inactive bonus block
          myBlocks_.add(new BonusDominoBlock(location,false,bRef.owner,bRef.subtreeIndex,bRef.newBonusForSubtreeIndex));
          //add active bonus block on top
          ++location.z;
          myBlocks_.add(new BonusDominoBlock(location,true,bRef.owner,bRef.subtreeIndex,bRef.newBonusForSubtreeIndex));
          --location.z;
          ++location.y;
        };
        myBlocks_.add(new SubtreeConnectionDominoBlock(location,true,lRef.owner,lRef.index));
        location.x+=2;
        location.y=ystart;
      };
    };
    
    /*
      Currently the RootDominoBlock is the reference block. Now we compute the
      center of the DominoTile's mass (ignoring blocks with z!=0) 
      and make the block closest to this center
      the new reference block. Note that the center may not contain an actual
      *DominoBlock.
      
      NOTE: It is possible that leafRefs.isEmpty(), i.e. there are no open leaves.
    */
    
    location.x=0;
    location.y=0;
    location.z=0;
    
    iter=blocksIterator();
    while (iter.hasNext()) 
    {
      IsoLocation adder=((IsoBlock)iter.next()).getRelativeLocation();
      if (adder.z==0) location.add(adder);
    };  
    location.x=(int)-Math.round((double)location.x/(double)numberOfBlocks());
    location.y=(int)-Math.round((double)location.y/(double)numberOfBlocks());
    location.z=(int)-Math.round((double)location.z/(double)numberOfBlocks());
    
    iter=blocksIterator();
    while (iter.hasNext()) 
    {
      DominoBlock block=(DominoBlock)iter.next();
      IsoLocation relLocation=block.getRelativeLocation();
      relLocation.add(location);
      block.setRelativeLocation(relLocation);
    };  
  };
  
  
  /**
  * <p>After making changes to the underlying {@link FormulaDominoTree}, this
  * method has to be called to update the <code>DominoTile</code>.</p>
  * <p>This method will not only work on the <code>DominoTile</code> for which
  * it is called but also on all <code>DominoTile</code>s connected to it.</p>
  * <p>This method has to be called from the event-dispatching thread, so you should
  * use {@link javax.swing.SwingUtilities#invokeAndWait(Runnable) invokeAndWait()} or
  * {@link javax.swing.SwingUtilities#invokeLater(Runnable) invokeLater()}. In any case you
  * have to make sure that no thread accesses the {@link FormulaDominoTree} 
  * <code>this DominoTile</code> is linked to until <code>updateFormulas()</code> 
  * has finished.</p>
  */
  public void updateFormulas()
  {
    DominoTile tile=this;
    while (tile.rootConnection_!=null) tile=tile.rootConnection_;
    updateFormulasDown(tile);
  }
  
  protected static void updateFormulasDown(DominoTile tile)
  {
    Iterator iter=tile.blocksIterator();
    while(iter.hasNext())
    {
      DominoBlock block=(DominoBlock)iter.next();
      block.updateFormula();
    };
    
    for (int i=0; i<tile.leafConnections_.length; ++i)
      if (tile.leafConnections_[i]!=null) updateFormulasDown(tile.leafConnections_[i]);
  };




//###################################################################################
//############################ IsoObject methods ####################################
//###################################################################################

  
  //see IsoObject
  public IsoLocation getReferenceBlockLocation() //TESTED
  {
    return new IsoLocation(refBlockLocation_);
  };
  
  //see IsoObject
  public void setReferenceBlockLocation(IsoLocation location) //TESTED
  {
    refBlockLocation_=new IsoLocation(location);
  };
  
  //see IsoObject
  public IsoShift getShift() //TESTED
  {
    return new IsoShift(shift_);
  };
  
  //see IsoObject
  public void setShift(IsoShift shift,Collection posChangedBlocks) //TESTED
  {
    if (Debugging.enabled)
    {
      if (shift.x<0 || shift.y<0 || shift.z<0 || shift.x>=1.0 || shift.y>=1.0 
      || shift.z>=1.0) throw new RuntimeException("illegal shift");
    };
    shift_=new IsoShift(shift);
    //Because all DominoBlocks have a (0,0,0) shift, no DominoBlock can change
    //position. Therefore posChangedBlocks does not need to be considered.
  };
  
  //see IsoObject
  public int numberOfBlocks() { return myBlocks_.size(); }; //TESTED
  
  //see IsoObject
  public Iterator blocksIterator() { return myBlocks_.iterator(); };  //TESTED
  
  //see IsoObject
  public void setWorld(IsoWorld world) //TESTED
  {
    myWorld_=world;
  };

  //see IsoObject
  public IsoWorld getWorld() {return myWorld_;};



//###################################################################################
//###################################################################################
//###################################################################################
//###################################################################################
//###################################################################################

  
  /** <p>The color used to paint the top side of empty (unlabelled) 
  * <code>DominoBlock</code>s.</p>*/
  public static final Color TOP_COLOR=Color.lightGray;
  /** <p>The color used for the lines drawn around the block's faces.</p>*/
  public static final Color OUTLINE_COLOR=Color.black;
  
  /** <p>The color used for the lighter side of a bonus block or an inactive root
  * block.</p> */
  public static final Color BONUS_BLUE=new Color(60,100,230);
  /** <p>The color used for the darker side of a bonus block or an inactive root
  * block.</p> */
  public static final Color BONUS_BLUE_DARK=new Color(50,90,190);
    
  /** <p>The colors used to represent unbound propositional variables.</p> */
  public static final Color[] GRAY_COLORS={new Color(128,128,128),new Color(96,96,96),new Color(198,198,198),new Color(152,152,152),new Color(32,32,32),new Color(250,250,250),new Color(226,226,226),new Color(172,172,172),new Color(212,212,212),new Color(238,238,238)};
  /** <p>The colors used to represent bound propositional variables.</p> */
  public static final Color[] NON_GRAY_COLORS=
  {new Color(0,192,255),  new Color(255,64,240), new Color(255,255,0),
   new Color(128,255,224),new Color(144,0,224),  new Color(192,0,96),
   new Color(255,160,80), new Color(0,160,96),   new Color(160,192,0),
   new Color(192,0,128),  new Color(160,255,0),  new Color(192,128,255),
   new Color(0,96,160),   new Color(0,224,224),  new Color(48,0,192),
   new Color(160,48,0),   new Color(192,144,0),  new Color(0,128,128),  
   new Color(255,192,255),new Color(224,255,144),new Color(224,255,192),
   new Color(255,224,224),new Color(208,255,255),new Color(255,255,192),
   new Color(224,192,112),new Color(255,234,255)};

  /**
  * <p>The class of the blocks that form a {@link DominoTile}.</p>
  */
  public abstract class DominoBlock extends JPanel implements IsoBlock, JIsoBlock
  {
    /** <p>location relative to the reference block.</p> */
    protected IsoLocation relativeLocation_;
    /** <p>see {@link #getPaintOrderPosition()}.</p>*/
    protected int paintOrderPosition_;
    /** <p>see {@link #isActive()}.</p> */
    protected boolean active_;
    /** 
    * <p>Caches the formula whose graphical representation labels this block.
    * This is done for efficiency and to allow changing the underlying 
    * {@link FormulaSyntaxNet} from a different thread.</p> */
    protected String formula_;
    
    
  
    /** 
    * <p>number of pixels, the top side of the block is shifted to the right
    * relative to the bottom side.</p>
    */
    private int xoffset_=0;
    /** 
    * <p>number of pixels, the top side of the block is shifted down
    * relative to the bottom side.</p>
    */
    private int yoffset_=0;
    
    
    /**
    * <p>If this is <code>false</code>, then <code>this DominoBlock</code> is just a 
    * filler for displaying purposes that can not be attached to another 
    * <code>DominoBlock</code>.</p>
    */
    public boolean isActive()
    {
      return active_;
    };
    
    /**
    * <p>Deactivates this block.</p>
    *
    * @see #isActive()
    */
    public void deactivate()
    {
      active_=false;
      repaint();
    };
    
    /**
    * <p>Returns the formula whose graphical representation currently labels
    * <code>this DominoBlock</code>.</p>
    */
    public String currentFormula() {return formula_;}; //TESTED
        
    /**
    * <p>Checks if the formula that labels this block has changed in the
    * underlying <code>FormulaDominoTree</code> and updates and repaints this
    * block if necessary.</p>
    * <p>Multi-threading issues are discussed at {@link DominoTile#updateFormulas()}.</p>
    */
    public abstract void updateFormula();

    /**
    * <p>default constructor.</p>
    */
    public DominoBlock() //TESTED
    {
      super(null);
      setOpaque(false); //because of the non-rectangular shape
    };


    /** <p>Returns the color used to paint the side of the block parallel to the x axis.</p>*/
    public Color xColor() {return Color.gray;};
    /** <p>Returns the color used to paint the side of the block parallel to the y axis.</p>*/
    public Color yColor() {return Color.gray;};

    
    //#############################################################################
    //######################### JComponent methods ################################
    //#############################################################################
  
    
    /** <p>see {@link JComponent#paintComponent(Graphics)}.</p>*/
    protected void paintComponent(Graphics g) //TESTED
    {
      Insets insets=getInsets();
      int currentWidth = getWidth() - insets.left - insets.right;
      int currentHeight = getHeight() - insets.top - insets.bottom;
      int x=insets.left;
      int y=insets.top;
    
      //yes, I know that having different width and height for a square doesn't make 
      //sense but I don't want to force the owner of this object to display it with a
      //1:1 aspect ratio
      int squareWidth=currentWidth-Math.abs(xoffset_);
      int squareHeight=currentHeight-Math.abs(yoffset_);
    
      int x0,y0,x1,y1,x2,y2,x3,y3,x4,y4,x5,y5,x6,y6;
      if (xoffset_<=0)
      {
        if (yoffset_<=0)
        {
          x0=x;
          y0=y;
          x1=x+(squareWidth-1);
          y1=y;
          x2=x1;
          y2=y1+(squareHeight-1);        
          x3=x0;
          y3=y2;
   
          x4=x1-xoffset_;
          y4=y1-yoffset_;
          x5=x4;
          y5=y4+(squareHeight-1);
          x6=x5-(squareWidth-1);
          y6=y5;
        }
        else //if (yoffset_>0)
        {
          throw new RuntimeException("unsupported");
        };
      }
      else //if  (xoffset_>0)
      {
        if (yoffset_<=0)
        {
          throw new RuntimeException("unsupported");
        }
        else //if (yoffset_>0)
        {
          throw new RuntimeException("unsupported");
        }
      };
    
    
      int[] xPoints1={x3,x2,x5,x6};
      int[] yPoints1={y3,y2,y5,y6};
      int[] xPoints2={x1,x4,x5,x2};
      int[] yPoints2={y1,y4,y5,y2};
    
      g.setColor(xColor());
      g.fillPolygon(xPoints1,yPoints1,4);
      g.setColor(yColor());
      g.fillPolygon(xPoints2,yPoints2,4);
    
      g.setColor(OUTLINE_COLOR);
      g.drawPolygon(xPoints1,yPoints1,4);
      g.drawPolygon(xPoints2,yPoints2,4);
    
      paintLabel(g,x0,y0,squareWidth,squareHeight);
    };
    
    
    /**
    * <p>Paints the top side (the label) of the block.</p>
    *
    * @param g the graphics context for drawing.
    * @param x the x coordinate of the upper-left corner of the label rectangle.
    * @param y the y coordinate of the upper-left corner of the label rectangle.
    * @param width the width of the label rectangle.
    * @param height the height of the label rectangle.
    */
    protected void paintLabel(Graphics g,int x,int y, int width, int height) //TESTED
    {
      paintRectangle(currentFormula(),0,g,x,y,width,height,true);
    };
    
    
    /**
    * <p>Paints a framed rectangle filled with one or more colors.</p>
    *
    * @param g the graphics context for drawing.
    * @param x the x coordinate of the upper-left corner of the rectangle.
    * @param y the y coordinate of the upper-left corner of the rectangle.
    * @param width the width of the rectangle.
    * @param height the height of the rectangle.
    * @param color "_|_" or propositional variable as used in strings returned by
    *        {@link FormulaDominoSquare#toPrefixRepresentation()}. This string will be
    *        used to determine color(s) for the rectangle.
    *
    * @see #paintRectangle(String,int,Graphics,int,int,int,int,boolean)  
    */
    protected void paintRectangle(Graphics g,int x,int y,int width,int height,String color)
    { //TESTED
      
      Color[] col;
      
      if (color.equals("_|_"))
      {
        col=new Color[1];
        col[0]=Color.red;
      }
      else
      {
        boolean unbound=true;
        if (color.charAt(0)!='X' || color.length()<2) 
          unbound=false;
        else
        {
          for (int i=1; i<color.length(); ++i)
            if (!Character.isDigit(color.charAt(i))) {unbound=false; break;};
        };
        
        if (unbound)
        {
          int num=color.length()-1; //-1 because of 1st character 'X' 
          col=new Color[num];
          for (int i=0; i<num; ++i) col[i]=GRAY_COLORS[Character.digit(color.charAt(num-i),10)];
        }
        else
        {
          int num=color.length();
          col=new Color[num];
          for (int i=0; i<num; ++i) 
            col[i]=NON_GRAY_COLORS[color.charAt(i)-'A'];
        };
      };  
      
      int rectWidth=(width/col.length)-4; //-4 to account for black and white frame
      int rectHeight=(height/col.length)-4; //-4 to account for black and white frame
      if (rectWidth<2) rectWidth=2;
      if (rectHeight<2) rectHeight=2;
      
      for (int i=0; i<col.length; ++i)
      {
        if (width<=0 || height<=0) break;
        g.setColor(Color.black);
        g.drawRect(x,y,width-1,height-1);
        ++x;++y;width-=2;height-=2;
        if (width<=0 || height<=0) break;
        g.setColor(Color.white);
        g.drawRect(x,y,width-1,height-1);
        ++x;++y;width-=2;height-=2;
        if (width<=0 || height<=0) break;
        
        g.setColor(col[i]);
        g.fillRect(x,y,width,height);
        x+=rectWidth/2;y+=rectHeight/2;width-=rectWidth;height-=rectHeight;
      };  
    };
    
    
    /**
    * <p>Paints a rectangle with a pattern based on a formula.</p>
    *
    * @param f a prefix formula as returned by 
    *        {@link FormulaDominoSquare#toPrefixRepresentation()}.
    * @param i index in <code>f</code> of the subformula to use for determining the
    *        pattern for the rectangle.
    * @param g the graphics context for drawing.
    * @param x the x coordinate of the upper-left corner of the rectangle.
    * @param y the y coordinate of the upper-left corner of the rectangle.
    * @param width the width of the rectangle.
    * @param height the height of the rectangle.
    * @param verticalSplit if the subformula to use for the pattern is an
    *        implication, then 2 rectangles need to be drawn. If 
    *        <code>verticalSplit</code> is <code>true</code>, then the 2 rectangles
    *        will be the top and bottom half of the rectangle 
    *        <code>(x,y,width,height)</code>. If <code>verticalSplit</code> is
    *        <code>false</code> the split will be into left and right half.
    *
    * @return the index of the character after the subformula that has been
    *         processed.
    *
    * @see #paintRectangle(Graphics,int,int,int,int,String)
    */
    protected int paintRectangle(String f,int i,Graphics g,int x,int y,int width,int height,boolean verticalSplit)  
    { //TESTED
      if (i<f.length())
      {
        if (f.charAt(i)=='-')  //->
        {
          i+=2;
          if (verticalSplit)
          {
            i=paintRectangle(f,i,g,x,y,width,height/2,verticalSplit^true);
            i=paintRectangle(f,i,g,x,y+height/2,width,height-(height/2),verticalSplit^true);
          }
          else //if horizontal split
          {
            i=paintRectangle(f,i,g,x,y,width/2,height,verticalSplit^true);
            i=paintRectangle(f,i,g,x+width/2,y,width-(width/2),height,verticalSplit^true);
          }
        }
        else //must be "("
        {
          ++i;
          int j=i;
          while (f.charAt(i)!=')') ++i;
          paintRectangle(g,x,y,width,height,f.substring(j,i));
          ++i;
        };
      };
      
      return i;
    };




    //#############################################################################
    //######################### IsoBlock methods ##################################
    //#############################################################################

    
    //see IsoBlock
    public IsoObject parentObject() {return DominoTile.this;};//TESTED
 
    //see IsoBlock
    public IsoShift getAbsoluteShift() {return DominoTile.this.getShift();};//TESTED
    
    //see IsoBlock
    public IsoShift getRelativeShift() {return new IsoShift(0,0,0);};
  
    //see IsoBlock
    public int getPaintOrderPosition() //TESTED
    {
      return paintOrderPosition_;
    };
    
    //see IsoBlock
    public void setPaintOrderPosition(int pos) //TESTED
    {
      paintOrderPosition_=pos;
    };
  
    //see IsoBlock
    public IsoLocation getAbsoluteLocation() //TESTED
    {
      IsoLocation loc=getRelativeLocation();
      loc.add(getReferenceBlockLocation());
      //shift does not have to be taken into account because DominoBlocks always
      //have a (0,0,0) relative shift and the parent object's shift is always
      //between 0.0 and 1.0
      return loc;
    };
  
    //see IsoBlock
    public IsoLocation getRelativeLocation() //TESTED
    {
      switch(getOrientation())
      {
        //NOTE: greater y coordinates are further SOUTH!
        case 1: return new IsoLocation(-relativeLocation_.y,relativeLocation_.x,relativeLocation_.z);
        case 2: return new IsoLocation(-relativeLocation_.x,-relativeLocation_.y,relativeLocation_.z);
        case 3: return new IsoLocation(relativeLocation_.y,-relativeLocation_.x,relativeLocation_.z);
      };
      return new IsoLocation(relativeLocation_);
    };  
    
    /** 
    * <p>Changes the value returned by {@link #getRelativeLocation}.</p>
    * <p><code>location</code> is copied, so later modifications will not affect
    * <code>this</code>.</p>
    */
    protected void setRelativeLocation(IsoLocation location) //TESTED
    {
      relativeLocation_=new IsoLocation(location);
    };  
  
    //see IsoBlock
    public boolean isMassive()
    {
      return massive_;
    };  
  
    //see IsoBlock
    public JIsoBlock graphicalRepresentation() {return this;}; //TESTED
    
   
   
   
    //#############################################################################
    //######################### JIsoBlock methods ##################################
    //#############################################################################
   
   
    //see JIsoBlock
    public JComponent toJComponent() {return this;}; //TESTED
    
    //see JIsoBlock
    public void setIsoOffsets(int xoffset, int yoffset) //TESTED
    {
      xoffset_=xoffset;
      yoffset_=yoffset;
    };
  };
  


//###################################################################################
//###################################################################################
//###################################################################################
//###################################################################################
//###################################################################################

  /**
  * <p>Class for {@link DominoTile.DominoBlock}s labelled with a bonus formula.</p>
  */
  public class BonusDominoBlock extends DominoBlock
  {
    /** <p>The tree that provides the bonus.</p> */
    private FormulaDominoTree myTree_;
    /** <p>The subtree connection index for which the bonus is valid.</p> */
    private int subtreeIndex_;
    /** <p>The index of the bonus in the list of bonuses.</p>*/
    private int  newBonusForSubtreeIndex_;
    /** 
    * <p>If this block has been attached to a tree, this is the
    * {@link FormulaDominoTree} that represents the attached bonus square, i.e. the
    * tree returned by {@link FormulaDominoTree#useBonusSquare(int,int)}.</p>
    */
    private FormulaDominoTree bonusTree_;
    
    /** 
    * <p>If this block has been {@link #attachBonusTree(DominoTree,int) attached to a tree}, 
    * this function returns the
    * {@link FormulaDominoTree} that represents the attached bonus square (the
    * tree returned by {@link FormulaDominoTree#useBonusSquare(int,int)}), otherwise
    * returns <code>null</code>.</p>
    */
    public FormulaDominoTree getBonusTree() {return bonusTree_;};
    
    /**
    * <p>See {@link FormulaDominoTree#useBonusSquare(int,DominoTree,int,int)}.</p>
    */
    public void attachBonusTree(DominoTree toTree, int subtreeConnectionIndex)
    {
      bonusTree_=(FormulaDominoTree)toTree.useBonusSquare(subtreeConnectionIndex,formulaDominoTree(),subtreeConnectionIndex(),newBonusForSubtreeConnectionIndex());
    };
    
    /**
    * <p>See {@link FormulaDominoTree#detachFromParent()}.</p>
    */
    public boolean detachBonusTree()
    {
      if (bonusTree_==null) return false; 
      boolean ret=bonusTree_.detachFromParent();
      bonusTree_=null;
      return ret;
    };
    
    /** <p>Returns the {@link FormulaDominoTree} that owns the bonus this tile 
    * represents.</p> */
    public FormulaDominoTree formulaDominoTree() {return myTree_;};
    
    /** <p>Returns the index of the subtree connection in
    * {@link #formulaDominoTree()} for which the bonus is valid.</p> 
    */
    public int subtreeConnectionIndex() {return subtreeIndex_;};
    
    /** <p>Returns the number of the bonus in the list of bonuses that
    * {@link #formulaDominoTree()} <em>adds to</em> the the bonuses valid for
    * {@link #subtreeConnectionIndex()}.</p> 
    */
    public int newBonusForSubtreeConnectionIndex() {return newBonusForSubtreeIndex_;};
        
    /** 
    * <p>Creates a new <code>BonusDominoBlock</code>.</p>
    *
    * @param relativeLocation value to be returned by {@link #getRelativeLocation()}.
    * @param active <code>false</code> if this block is just a dummy for displaying.
    * @param tree the subtree of {@link DominoTile}'s tree that provides the bonus.
    * @param subtreeIndex the subtree connection index for which the bonus is valid.
    * @param newBonusForSubtreeIndex the index of the bonus in the list of bonuses.
    */
    public BonusDominoBlock(IsoLocation relativeLocation, boolean active, FormulaDominoTree tree,int subtreeIndex,int newBonusForSubtreeIndex)
    {
      relativeLocation_=new IsoLocation(relativeLocation);
      active_=active;
      myTree_=tree;
      subtreeIndex_=subtreeIndex;
      newBonusForSubtreeIndex_=newBonusForSubtreeIndex;
      FormulaDominoSquare sq=(FormulaDominoSquare)tree.newBonusSquareForSubtreeConnection(subtreeIndex,newBonusForSubtreeIndex);
      formula_=sq.toPrefixRepresentation(globalUnboundIdAssigner);
    };
    
    
    /** <p>Returns the color used to paint the side of the block parallel to the x axis.</p>*/
    public Color xColor() {return BONUS_BLUE;};
    /** <p>Returns the color used to paint the side of the block parallel to the y axis.</p>*/
    public Color yColor() {return BONUS_BLUE_DARK;};
    
    //see DominoBlock
    public void updateFormula()
    {
      FormulaDominoSquare sq=(FormulaDominoSquare)myTree_.newBonusSquareForSubtreeConnection(subtreeIndex_,newBonusForSubtreeIndex_);
      String newFormula=sq.toPrefixRepresentation(globalUnboundIdAssigner);
      if (!newFormula.equals(formula_))
      {
        formula_=newFormula;
        repaint();
      };
    };
  };
  
  /**
  * <p>Class for {@link DominoTile.DominoBlock}s labelled with a root formula.</p>
  */
  public class RootDominoBlock extends DominoBlock
  {
    /** <p>The tree that has the formula as root.</p> */
    private FormulaDominoTree myTree_;
    
    /** 
    * <p>Returns the {@link FormulaDominoTree} whose root formula this tile 
    * represents.</p> 
    */
    public FormulaDominoTree formulaDominoTree() {return myTree_;};
    
    /** 
    * <p>Creates a new <code>RootDominoBlock</code>.</p>
    *
    * @param relativeLocation value to be returned by {@link #getRelativeLocation()}.
    * @param active <code>false</code> if this block is just a dummy for displaying.
    * @param tree the subtree of {@link DominoTile}'s tree that provides the bonus.
    */
    public RootDominoBlock(IsoLocation relativeLocation, boolean active, FormulaDominoTree tree)
    { //TESTED
      relativeLocation_=new IsoLocation(relativeLocation);
      active_=active;
      myTree_=tree;
      FormulaDominoSquare sq=(FormulaDominoSquare)tree.rootSquare();
      formula_=sq.toPrefixRepresentation(globalUnboundIdAssigner);
    };
    
    
    /** <p>Returns the color used to paint the side of the block parallel to the x axis.</p>*/
    public Color xColor() { if (!isActive()) return super.xColor(); else return (numberOfBlocks()>1) ? new Color(220,50,20) : BONUS_BLUE; };
    /** <p>Returns the color used to paint the side of the block parallel to the y axis.</p>*/
    public Color yColor() { if (!isActive()) return super.yColor(); else return (numberOfBlocks()>1) ? new Color(180,40,15) : BONUS_BLUE_DARK; };
    
    //see DominoBlock
    public void updateFormula()
    {
      FormulaDominoSquare sq=(FormulaDominoSquare)myTree_.rootSquare();
      String newFormula=sq.toPrefixRepresentation(globalUnboundIdAssigner);
      if (!newFormula.equals(formula_))
      {
        formula_=newFormula;
        repaint();
      };
    };
  };
  
  /**
  * <p>Class for {@link DominoTile.DominoBlock}s labelled with a subtree connection formula.</p>
  */
  public class SubtreeConnectionDominoBlock extends DominoBlock
  {
    /** <p>The tree that owns the subtree connection.</p> */
    private FormulaDominoTree myTree_;
    /** <p>The index of the subtree connection.</p> */
    private int subtreeConnectionIndex_;
    
    
    /** <p>Returns the {@link FormulaDominoTree} that owns the subtree connection
    * this tile represents.</p> */
    public FormulaDominoTree formulaDominoTree() {return myTree_;};
    
    /** <p>Returns the index of this subtree connection in
    * {@link #formulaDominoTree()}.</p> 
    */
    public int subtreeConnectionIndex() {return subtreeConnectionIndex_;};
    
    /** 
    * <p>Creates a new <code>SubtreeConnectionDominoBlock</code>.</p>
    *
    * @param relativeLocation value to be returned by {@link #getRelativeLocation()}.
    * @param active <code>false</code> if this block is just a dummy for displaying.
    * @param tree the subtree of {@link DominoTile}'s tree that owns the subtree connection.
    * @param subtreeConnectionIndex the index of the subtree connection.
    */
    public SubtreeConnectionDominoBlock(IsoLocation relativeLocation, boolean active, FormulaDominoTree tree, int subtreeConnectionIndex)
    { //TESTED
      relativeLocation_=new IsoLocation(relativeLocation);
      active_=active;
      myTree_=tree;
      subtreeConnectionIndex_=subtreeConnectionIndex;
      FormulaDominoSquare sq=(FormulaDominoSquare)tree.subtreeConnectionSquare(subtreeConnectionIndex_);
      formula_=sq.toPrefixRepresentation(globalUnboundIdAssigner);
    };
    
    /** <p>Returns the color used to paint the side of the block parallel to the x axis.</p>*/
    public Color xColor() {return new Color(50,200,50);};
    /** <p>Returns the color used to paint the side of the block parallel to the y axis.</p>*/
    public Color yColor() {return new Color(40,150,40);};
    
    //see DominoBlock
    public void updateFormula()
    {
      FormulaDominoSquare sq=(FormulaDominoSquare)myTree_.subtreeConnectionSquare(subtreeConnectionIndex_);
      String newFormula=sq.toPrefixRepresentation(globalUnboundIdAssigner);
      if (!newFormula.equals(formula_))
      {
        formula_=newFormula;
        repaint();
      };
    };
    
  };
  
  /** <p>An unlabelled {@link DominoTile.DominoBlock}.</p>*/
  public class EmptyDominoBlock extends DominoBlock
  {
    /** 
    * <p>Creates a new <code>EmptyDominoBlock</code>.</p>
    *
    * @param relativeLocation value to be returned by {@link #getRelativeLocation()}.
    */
    EmptyDominoBlock(IsoLocation relativeLocation) //TESTED
    {
      relativeLocation_=new IsoLocation(relativeLocation);
      active_=false;
      formula_=null;
    };
    
    //see DominoBlock, does nothing
    public void updateFormula() {};
    
    /** <p>See {@link DominoTile.DominoBlock#paintLabel(Graphics,int,int,int,int)}. */
    protected void paintLabel(Graphics g,int x,int y, int width, int height)
    {
      g.setColor(TOP_COLOR);
      g.fillRect(x,y,width,height);
      g.setColor(OUTLINE_COLOR);
      --width;
      --height;
      if (width>=0 && height>=0) g.drawRect(x,y,width,height);
    };
  };
  
};



