/* Copyright (C) 2002 Matthias S. Benkmann <m.s.b@gmx.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License (ONLY THIS VERSION).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package domino.ui;

/**
* <p><code>IsoBlock</code>s are the fundamental building blocks of 
* an {@link IsoWorld}.
* Every <code>IsoBlock</code> is part of exactly one {@link IsoObject}.</p>
*/
public interface IsoBlock
{
  /**
  * <p>Returns the {@link IsoObject} this block belongs to.</p>
  */
  public IsoObject parentObject();
 
  /**
  * <p>Returns the location of <code>this</code> block relative to the parent
  * {@link IsoObject}'s reference block (which has a relative location of (0,0,0)).</p>
  * <p>The returned {@link IsoLocation} is a copy of the data, so modifications made
  * to it will not affect the {@link IsoBlock}.</p>
  * <p>Unlike {@link #getAbsoluteLocation()} this method does not take shift 
  * into account.</p>
  *
  * @see #getAbsoluteLocation()
  */
  public IsoLocation getRelativeLocation();
 
  /**
  * <p>Returns the location of <code>this</code> block within the 
  * {@link IsoWorld}.</p>
  * <p>The returned {@link IsoLocation} is a copy of the data, so modifications made
  * to it will not affect the {@link IsoBlock}.</p>
  * <p>This method takes shift into account. If <code>this</code> block has a 
  * shift relative to the parent object and the parent object has been shifted so
  * that <code>this</code> block's shift would have a coordinate &gt;=1.0, then
  * the value returned by <code>getAbsoluteLocation()</code> will be adjusted to
  * get the shift into the legal range again.</p>
  *
  * @see #getRelativeLocation()
  */
  public IsoLocation getAbsoluteLocation();
  
  /**
  * <p>Returns the shift of <code>this</code> block relative to the parent object.</p>
  * <p>The returned {@link IsoShift} is a copy of the data, so modifications made
  * to it will not affect the {@link IsoBlock}.</p>
  *
  * @see #getAbsoluteShift()
  */
  public IsoShift getRelativeShift();
  
  /**
  * <p>Returns the shift of <code>this</code> block relative to the location returned
  * by {@link #getAbsoluteLocation()}.</p>
  * <p>The returned {@link IsoShift} is a copy of the data, so modifications made
  * to it will not affect the {@link IsoBlock}.</p>
  *
  * @see #getAbsoluteLocation()
  */
  public IsoShift getAbsoluteShift();
  
  /**
  * <p>Returns the value set with the most recent call to
  * {@link #setPaintOrderPosition(int)}. This method is meant to be used by a
  * container managing <code>IsoBlocks</code> such as {@link IsoWorld}.</p>
  */
  public int getPaintOrderPosition();
  
  /**
  * <p>Sets the value to be returned by {@link #getPaintOrderPosition()}.</p>
  * <p>This method is meant to be used by a
  * container managing <code>IsoBlocks</code> such as {@link IsoWorld}.</p>
  */
  public void setPaintOrderPosition(int pos); //TESTED
  
  /**
  * <p>Returns <code>true</code> if <code>this</code> block can not share the same
  * {@link IsoLocation} with another <code>IsoBlock</code>.</p>
  */
  public boolean isMassive();
  
  /**
  * <p>Returns the component to be used for rendering <code>this</code> block on
  * the screen.</p>
  */
  public JIsoBlock graphicalRepresentation();
};

