/* Copyright (C) 2002 Matthias S. Benkmann <m.s.b@gmx.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License (ONLY THIS VERSION).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package domino.ui;

import java.util.Collection;
import java.util.Iterator;


/**
* <p>An <code>IsoObject</code> consists of 0 or more {@link IsoBlock}s and can
* be added to an {@link IsoWorld}.</p>
*/
public interface IsoObject
{
  /**
  * <p>Returns the location of the reference block within the {@link IsoWorld}
  * <code>this</code> object is part of.</p>
  * <p>If <code>this</code> object is not part of an {@link IsoWorld}, the result
  * is undefined.</p>
  * <p>Note, that the <code>IsoObject</code> does not necessarily have an actual
  * {@link IsoBlock} at the location returned by this method. The location is just
  * the point of reference for {@link IsoBlock#getAbsoluteLocation()}.</p>
  *
  * @see #getShift()
  */
  public IsoLocation getReferenceBlockLocation();
  
  /**
  * <p>Sets the position of the reference block in the {@link IsoWorld}. This
  * method is intended to be used only by the {@link IsoWorld} managing this object.
  * Use {@link IsoWorld#moveObject(IsoObject,IsoLocation,IsoShift)} to move the
  * <code>IsoObject</code>.</p>
  * <p>This method creates a copy of the <code>location</code> argument.</p>
  *
  * @see #setShift(IsoShift,Collection)
  */ 
  public void setReferenceBlockLocation(IsoLocation location);
  
  /**
  * <p>Sets the shift for the <code>IsoObject</code> as a whole. This shift is
  * added to each {@link IsoBlock}'s 
  * {@link IsoBlock#getRelativeShift() relative shift} value. Because no
  * shift coordinate may exceed 1.0, this may cause the position of an
  * {@link IsoBlock} to change to bring the shift back into the permitted range.</p>
  * <p>This method is intended to be called by {@link IsoWorld} only. Use
  * {@link IsoWorld#moveObject(IsoObject,IsoLocation,IsoShift)} to move
  * an <code>IsoObject</code> around within the {@link IsoWorld}.</p>
  *
  * @param shift the new shift for the object. The argument is copied, so later
  *        changes to this object will not affect <code>this IsoObject</code>.
  * @param posChangedBlocks if non-<code>null</code>, all {@link IsoBlock}s that
  *        change position due to the new shift will be added to this
  *        {@link Collection}.
  *
  * @see #getShift()
  * @see #setReferenceBlockLocation(IsoLocation)
  *
  */
  public void setShift(IsoShift shift, Collection posChangedBlocks);
  
  /**
  * <p>Returns the current shift value of <code>this IsoObject</code>.</p>
  *
  * @see #setShift(IsoShift, Collection)
  * @see #getReferenceBlockLocation()
  */
  public IsoShift getShift();
  
  /**
  * <p>Returns the number of {@link IsoBlock}s, <code>this</code> object consists
  * of.</p>
  * <p>Note that <code>IsoObject</code>s with 0 {@link IsoBlock}s are possible.</p>
  */
  public int numberOfBlocks();
  
  /**
  * <p>Returns an iterator to iterate over all {@link IsoBlock}s <code>this</code>
  * object consists of.</p>
  * <p>Note that <code>IsoObject</code>s without any {@link IsoBlock}s are 
  * possible.</p>
  */
  public Iterator blocksIterator();
  
  /**
  * <p>Called by {@link IsoWorld} <code>world</code> when <code>this</code> 
  * <code>IsoObject</code> is added to it.</p>
  * <p>When the <code>IsoObject</code> is removed again, this will be called with
  * <code>world==null</code>.
  *
  * @see #getWorld()
  */
  public void setWorld(IsoWorld world);
  
  /**
  * <p>Returns the {@link IsoWorld} <code>this</code> object is part of or
  * <code>null</code> if this object is not part of a world.</p>
  *
  * @see #setWorld(IsoWorld)
  */
  public IsoWorld getWorld();
};



