/* Copyright (C) 2002 Matthias S. Benkmann <m.s.b@gmx.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License (ONLY THIS VERSION).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package domino.ui;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.HashMap;
import javax.swing.JComponent;

/**
* <p>Manages a world with an isometric view.</p>
*/
public class IsoWorld
{
  /** <p>manages the actual graphics.</p>*/
  private JIsoWorld myJIsoWorld_;
  
  /**
  * <p>Maps a {@link IsoLocation}s to {@link Collection}s of {@link IsoBlock}s
  * found at that location. 
  * If a block has a non-zero {@link IsoBlock#getAbsoluteShift() shift}, 
  * it will be contained in multiple Collections.</p>
  * <p>This map does not contain empty {@link Collection}s. If a {@link Collection}
  * becomes empty, its mapping is removed.</p>
  */
  private Map blocks_=new HashMap();
  
  
  /**
  * <p>Creates a new IsoWorld.</p>
  *
  * @param minusxdim <code>-minusxdim</code> is the index of the left-most column of 
  *        the board (a negative number, because <code>minusxdim&gt;=0</code>).
  * @param plusxdim the index of the right-most column of
  *        the board.
  * @param minusydim <code>-minusydim</code> is the index of the down-most row of 
  *        the board (a negative number, because <code>minusydim&gt;=0</code>).
  * @param plusydim the index of the up-most row of the board.
  * @param squareSide initial side length of the bottom and top side square of a block.
  * @param xoffsetRel <code>xoffsetRel_*squareSide</code> is the number of pixels, 
  *        the top side of the block is shifted to the right in comparison to the 
  *        bottom side.
  * @param yoffsetRel <code>yoffsetRel_*squareSide</code> is the number of pixels, 
  *        the top side of the block is shifted down in comparison to the 
  *        bottom side.
  */
  public IsoWorld(int minusxdim, int plusxdim, int minusydim, int plusydim, int squareSide, double xoffsetRel, double yoffsetRel) 
  { //TESTED
    myJIsoWorld_=new JIsoWorld(minusxdim, plusxdim, minusydim, plusydim, squareSide, xoffsetRel, yoffsetRel);
  };
  
  
  /**
  * <p>Adds or removes rows or columns from the board.</p>
  * <p>This method only increases or decreases the visible part of the board.
  * Removing rows/colums will not remove any blocks located there and will not
  * prevent new blocks from being added there.</p>
  *
  * @param leftrighttopbottom Where to add/remove: 
  *        <code>0=left, 1=right, 2=top, 3=bottom</code>
  * @param numLines how much space to add. Negative values remove space.
  */
  public void addLines(int leftrighttopbottom, int numLines) 
  {
    myJIsoWorld_.addLines(leftrighttopbottom, numLines);
  };
  
  /** 
  * <p>Returns the index of the right-most column of the board.</p>
  */
  public int plusXDim() {return myJIsoWorld_.plusXDim();};
  /** 
  * <p>Returns abs() of the index of the left-most column of the board.</p> 
  */
  public int minusXDim() {return myJIsoWorld_.minusXDim();};
  /** 
  * <p>Returns the index of the up-most row of the board.</p>
  */
  public int plusYDim() {return myJIsoWorld_.plusYDim();};
  /** 
  * <p>Returns abs() of the index of the down-most row of the board.</p> 
  */
  public int minusYDim() {return myJIsoWorld_.minusYDim();};
  
  /** 
  * <p>Returns <code>true</code> if <code>object</code> would cause a collision if 
  * its reference block were put at <code>location</code> with the given
  * <code>shift</code>. <code>object</code> may or may not be part of 
  * <code>this IsoWorld</code>. Collisions of <code>object</code> with itself do 
  * not count.</p>
  */
  public boolean collision(IsoObject object, IsoLocation location, IsoShift shift)
  { //TESTED
    Iterator i=object.blocksIterator();
    while(i.hasNext())
    {
      IsoBlock block=(IsoBlock)i.next();
      if (!block.isMassive()) continue;
      IsoLocation loc=block.getRelativeLocation();
      loc.add(location);
      
      int xm=0; if (shift.x>0) xm=1;
      int ym=0; if (shift.y>0) ym=1;
      int zm=0; if (shift.z>0) zm=1;
      
      for (int x=loc.x; x<=loc.x+xm; ++x)
      {
        for (int y=loc.y; y<=loc.y+ym; ++y)
        {
          for (int z=loc.z; z<=loc.z+zm; ++z)
          {
            Collection c=blocksAtLocation(new IsoLocation(x,y,z));
            if (c==null) continue;
            Iterator j=c.iterator();
            while (j.hasNext())
            {
              IsoBlock otherBlock=(IsoBlock)j.next();
              if (otherBlock.parentObject()==object) continue;
              if (otherBlock.isMassive()) return true;
            };
          };
        };
      };
    };

    return false;
  };
 
  
  /**
  * <p>Adds <code>block</code> to the {@link Collection} of {@link IsoBlock}s 
  * at <code>location</code> (see {@link #blocksAtLocation(IsoLocation)}), creating
  * this {@link Collection} if necessary.</p>
  *
  * @see #removeFromBlocks(IsoLocation,IsoBlock)
  */
  protected void addToBlocks(IsoLocation location,IsoBlock block) //TESTED
  {
    Collection c=blocksAtLocation(location);
    if (c==null)
    {
      c=new LinkedList();
      blocks_.put(new IsoLocation(location),c);
    };
    c.add(block);
  };
  
  /**
  * <p>Removes <code>block</code> from the {@link Collection} of {@link IsoBlock}s 
  * at <code>location</code> (see {@link #blocksAtLocation(IsoLocation)}), deleting
  * this {@link Collection} if it is empty afterwards.</p>
  *
  * @see #addToBlocks(IsoLocation,IsoBlock)
  */
  protected void removeFromBlocks(IsoLocation location, IsoBlock block) //TESTED
  {
    Collection c=blocksAtLocation(location);
    if (c!=null)
    {
      c.remove(block);
      if (c.isEmpty()) blocks_.remove(location);
    };  
  };
  
  
  /** 
  * <p>Returns -1 if when displaying an {@link IsoBlock} the top side has a 
  * lower y coordinate than the bottom side, +1 if the top side has the same or 
  * a higher y coordinate than the bottom side.</p>
  *
  * @see #xOffsetSign2()
  */
  public int yOffsetSign2() //TESTED
  {
    int i=graphicalRepresentation().yOffset();
    if (i<0) return -1;
    return +1;
  };
  
  /** 
  * <p>Returns -1 if when displaying an {@link IsoBlock} the top side has a 
  * lower x coordinate than the bottom side, +1 if the top side has the same or 
  * a higher x coordinate than the bottom side.</p>
  *
  * @see #yOffsetSign2()
  */
  public int xOffsetSign2()
  {
    int i=graphicalRepresentation().xOffset();
    if (i<0) return -1;
    return +1;
  };
  
  /** 
  * <p>Adjust the paint order so that <code>block</code> is painted correctly.</p>
  * <p>The layer on which a block is painted only depends on its {@link IsoLocation}.
  * However, more than 1 {@link IsoBlock} can occupy the same location, so
  * additional paint order tuning has to be performed based on the blocks'
  * {@link IsoShift}s. That is the purpose of this function.</p>
  * <p><code>block</code> must have been added to the world before using this
  * function.</p>
  */
  protected void insertInPaintOrder(IsoBlock block)
  {
    IsoLocation loc=block.getAbsoluteLocation();
    IsoShift shift1=block.getAbsoluteShift();
    Collection blocks=blocksAtLocation(loc);
    int paintOrderPosition=Integer.MAX_VALUE;
    int numOtherBlocks=0;
    Iterator i=blocks.iterator();
    while (i.hasNext())
    {
      IsoBlock otherBlock=(IsoBlock)i.next();
      if (block==otherBlock) continue;
      
      //only consider blocks whose main location is the same as that of block
      if (!otherBlock.getAbsoluteLocation().equals(loc)) continue;
      
      ++numOtherBlocks;
      
      IsoShift shift2=otherBlock.getAbsoluteShift();
      
      int cmp2;
      int cmp1=IsoShift.shiftCompare(shift1.y,shift2.y);
      if (cmp1==0)
      { 
        cmp1=IsoShift.shiftCompare(shift1.x,shift2.x);
        if (cmp1==0)
        {
          cmp1=IsoShift.shiftCompare(shift1.z,shift2.z);
          cmp2=-1;
        }
        else cmp2=xOffsetSign2();
      }
      else cmp2=yOffsetSign2();
      
      if (cmp1!=cmp2) 
      {
        int pop=otherBlock.getPaintOrderPosition();
        if (pop<paintOrderPosition) paintOrderPosition=pop;
      };  
    };
    
    if (paintOrderPosition==Integer.MAX_VALUE)
      paintOrderPosition=numOtherBlocks;
    else
    {  
      i=blocks.iterator();
      while (i.hasNext())
      {
        IsoBlock otherBlock=(IsoBlock)i.next();
        if (block==otherBlock) continue;
      
        //only consider blocks whose main location is the same as that of block
        if (!otherBlock.getAbsoluteLocation().equals(loc)) continue;
      
        int pop=otherBlock.getPaintOrderPosition();
        if (pop>=paintOrderPosition) 
        {
          otherBlock.setPaintOrderPosition(pop+1);
          myJIsoWorld_.setPaintOrderPosition(otherBlock.graphicalRepresentation().toJComponent(),pop+1);
        };
      };
    };  

    block.setPaintOrderPosition(paintOrderPosition);
    myJIsoWorld_.setPaintOrderPosition(block.graphicalRepresentation().toJComponent(),paintOrderPosition);
  };
  

  /** 
  * <p>Removes block's paint order assignment, adjust other blocks' paint order
  * positions to close the gap.</p>
  */
  protected void removeFromPaintOrder(IsoBlock block)
  {
    IsoLocation loc=block.getAbsoluteLocation();
    Collection blocks=blocksAtLocation(loc);
    if (blocks!=null)
    {
      int paintOrderPosition=block.getPaintOrderPosition();
      block.setPaintOrderPosition(-1);
      myJIsoWorld_.setPaintOrderPosition(block.graphicalRepresentation().toJComponent(),-1);
      Iterator i=blocks.iterator();
      while (i.hasNext())
      {
        IsoBlock otherBlock=(IsoBlock)i.next();
        if (block==otherBlock) continue;
      
        //only consider blocks whose main location is the same as that of block
        if (!otherBlock.getAbsoluteLocation().equals(loc)) continue;
      
        int pop=otherBlock.getPaintOrderPosition();
        if (pop>paintOrderPosition) otherBlock.setPaintOrderPosition(pop-1);
      };
    
      i=blocks.iterator();
      while (i.hasNext())
      {
        IsoBlock otherBlock=(IsoBlock)i.next();
        if (block==otherBlock) continue;
      
        //only consider blocks whose main location is the same as that of block
        if (!otherBlock.getAbsoluteLocation().equals(loc)) continue;
      
        int pop=otherBlock.getPaintOrderPosition();
        if (pop>=paintOrderPosition) // >= because it may have been decremented above
          myJIsoWorld_.setPaintOrderPosition(otherBlock.graphicalRepresentation().toJComponent(),pop);
      };  
    };  
  };
  
  
  
  /**
  * <p>Puts <code>object</code>'s {@link IsoBlock}s into the top layers, 
  * regardless of their {@link IsoLocation}s. The relative layer order of the
  * blocks is preserved.</p>
  *
  * @see #putInDefaultLayer(IsoObject)
  */
  public void putInTopLayer(IsoObject object)
  {
    class CompObj implements java.util.Comparator
    {
      public IsoBlock block;
      public int layer;
      public int compare(Object o1, Object o2)
      {
        int l1=((CompObj)o1).layer;
        int l2=((CompObj)o2).layer;
        if (l1==l2) return 0;
        if (l1<l2) return -1;
        return +1;
      };
      public CompObj(IsoBlock b,int l){block=b; layer=l;};
    };
    
    CompObj[] c=new CompObj[object.numberOfBlocks()];
    int i=0;
    if (c.length>0)
    {
      Iterator blocks=object.blocksIterator();
      while (blocks.hasNext()) 
      {
        IsoBlock block=(IsoBlock)blocks.next();
        c[i++]=new CompObj(block,myJIsoWorld_.getLayer(block.graphicalRepresentation().toJComponent()));
      };
    
      java.util.Arrays.sort((Object[])c,c[0]);
    };  
      
    int l=Integer.MAX_VALUE;
    while (i>0) 
    {
      myJIsoWorld_.setLayer(c[--i].block.graphicalRepresentation().toJComponent(),l--);
    };  
  };
  

  /**
  * <p>Puts <code>object</code> into the layer that is appropriate for its
  * {@link IsoLocation}.</p>
  *
  * @see #putInTopLayer(IsoObject)
  */
  public void putInDefaultLayer(IsoObject object)
  {
    Iterator blocks=object.blocksIterator();
    while (blocks.hasNext())
    {
      IsoBlock block=(IsoBlock)blocks.next();
      myJIsoWorld_.setLayer(block.graphicalRepresentation().toJComponent(),block.getAbsoluteLocation());
    };
  };
  

  /**
  * <p>Calls {@link JIsoWorld#setBlock(JIsoBlock,IsoLocation,IsoShift)} for every
  * block in this {@link IsoWorld}.</p>
  */
  protected void setBlocks()
  {
    Set coordsAndBlocks=blocks_.entrySet();
    Iterator i=coordsAndBlocks.iterator();
    while (i.hasNext())
    {
      Map.Entry ent=(Map.Entry)i.next();
      Collection blocksAtPosition=(Collection)ent.getValue();
      Iterator j=blocksAtPosition.iterator();
      while (j.hasNext())
      {
        IsoBlock block=(IsoBlock)j.next();
        myJIsoWorld_.setBlock(block.graphicalRepresentation(),block.getAbsoluteLocation(),block.getAbsoluteShift());
      };
    };
  };
  
  
  /**
  * <p>Changes the size of the squares that form the top and bottom sides of the
  * blocks.</p>
  * <p>The center of the viewport will be a fixpoint, i.e. it will still show
  * the same location on the board after the resize.</p>
  *
  * @param sideLengthInPixels new side length of a square.
  *
  * @see #setSquareSize(int,double,double)
  * @see #setSquareSize(int,double,double,int,int)
  */
  public void setSquareSize(int sideLengthInPixels)
  {
    graphicalRepresentation().setSquareSize(sideLengthInPixels);
    setBlocks();
  };
  
  
  /**
  * <p>Changes the size of the squares that form the top and bottom sides of the
  * blocks.</p>
  *
  * @param sideLengthInPixels new side length of a square.
  * @param fixPointX x coordinate of fixpoint (normalized board coordinate). This
  *        board location will still be visible at the same position in the viewport
  *        after the resize.
  * @param fixPointY y coordinate of fixpoint (normalized board coordinate).
  *
  * @see #setSquareSize(int)
  * @see #setSquareSize(int,double,double,int,int)
  */
  public void setSquareSize(int sideLengthInPixels, double fixPointX, double fixPointY)
  {
    graphicalRepresentation().setSquareSize(sideLengthInPixels, fixPointX, fixPointY);
    setBlocks();
  };
  
  
  /**
  * <p>Changes the size of the squares that form the top and bottom sides of the
  * blocks.</p>
  *
  * @param sideLengthInPixels new side length of a square.
  * @param fixPointX x coordinate of fixpoint (normalized board coordinate).
  * @param fixPointY y coordinate of fixpoint (normalized board coordinate).
  * @param targetX x coordinate where fixpoint should be located after resize 
  *        (viewport coordinate).
  * @param targetY y coordinate where fixpoint should be located after resize 
  *        (viewport coordinate).
  *
  * @see #setSquareSize(int)
  * @see #setSquareSize(int,double,double)
  */
  public void setSquareSize(int sideLengthInPixels, double fixPointX, double fixPointY, int targetX, int targetY)
  { //TESTED
    graphicalRepresentation().setSquareSize(sideLengthInPixels, fixPointX, fixPointY, targetX, targetY);
    setBlocks();
  };
  
  
  /**
  * <p>Returns the current side length (in pixels) of the top and bottom square of
  * a block.</p>
  *
  * @see #setSquareSize(int,double,double,int,int)
  */
  public int getSquareSize() //TESTED
  {
    return graphicalRepresentation().getSquareSize();
  };
  
  /**
  * <p>Returns the viewport to the <code>IsoWorld</code>.</p>
  */
  public JIsoWorld graphicalRepresentation() //TESTED
  {
    return myJIsoWorld_;
  };
  
  
  /**
  * <p>Adds <code>object</code> to the world at the given <code>location</code>
  * with the given <code>shift</code>.</p>
  * <p>After being added to the world, the object must not change its shape, i.e.
  * its {@link IsoBlock}s must not change their properties.</p>
  * <p>It is possible to add objects at locations that exceed the visible board 
  * coordinates. In that case, the object will not be drawn (completely).</p>
  *
  * @throws IsoObjectCollision if adding <code>object</code> would cause a collision.
  *         In that case, <code>object</code> will not be added to the world.
  *
  * @see #addObjectNoCollision(IsoObject,IsoLocation,IsoShift)
  */
  public void addObject(IsoObject object, IsoLocation location, IsoShift shift) //TESTED
  {
    if (collision(object,location,shift)) throw new IsoObjectCollision();
    addObjectNoCollision(object, location, shift);
  }; 
  

  /**
  * <p>Like {@link #addObject(IsoObject object, IsoLocation location, IsoShift shift)}
  * but without collision check.</p>
  */
  public void addObjectNoCollision(IsoObject object, IsoLocation location, IsoShift shift)
  { //TESTED
    object.setWorld(this);
    object.setReferenceBlockLocation(location);
    object.setShift(shift,null);
    
    Iterator i=object.blocksIterator();
    while(i.hasNext())
    {
      IsoBlock block=(IsoBlock)i.next();
      IsoLocation loc=block.getAbsoluteLocation();
      IsoShift sh=block.getAbsoluteShift();
      
      int xm=0; if (sh.x>0) xm=1;
      int ym=0; if (sh.y>0) ym=1;
      int zm=0; if (sh.z>0) zm=1;
      
      for (int x=loc.x; x<=loc.x+xm; ++x)
      {
        for (int y=loc.y; y<=loc.y+ym; ++y)
        {
          for (int z=loc.z; z<=loc.z+zm; ++z)
          {
            addToBlocks(new IsoLocation(x,y,z),block);
          };
        };
      };
      
      myJIsoWorld_.addBlock(block.graphicalRepresentation(),loc,shift);
      insertInPaintOrder(block);
    };
  };
  
  
  /**
  * <p>Moves <code>object</code> (which must already be part of the world) to
  * <code>newLocation</code> with <code>newShift</code>.</p>
  *
  * @throws IsoObjectCollision if moving <code>object</code> would cause a collision.
  *         In that case, <code>object</code> will not be moved.
  *
  * @see #removeObject(IsoObject)
  * @see #addObject(IsoObject,IsoLocation,IsoShift)
  */
  public void moveObject(IsoObject object, IsoLocation newLocation, IsoShift newShift)
  {
    if (collision(object,newLocation,newShift)) throw new IsoObjectCollision();
    
    //could be made more efficient! If this is improved to not remove the object,
    //then don't forget to adjust the layer of every block
    
    int layer=-1;
    if (object.numberOfBlocks()>0)
    {
      IsoBlock block=(IsoBlock)object.blocksIterator().next();
      layer=myJIsoWorld_.getLayer(block.graphicalRepresentation().toJComponent());
    };  
    removeObject(object);
    addObjectNoCollision(object, newLocation, newShift);
    if (layer>=Integer.MAX_VALUE-object.numberOfBlocks()) putInTopLayer(object);
  };
  

  /**
  * <p>Removes <code>object</code> from the world.</p>
  *
  * @see #moveObject(IsoObject,IsoLocation,IsoShift)
  * @see #addObject(IsoObject,IsoLocation,IsoShift)
  */
  public void removeObject(IsoObject object) //TESTED
  {
    Iterator i=object.blocksIterator();
    while(i.hasNext())
    {
      IsoBlock block=(IsoBlock)i.next();
      IsoLocation loc=block.getAbsoluteLocation();
      IsoShift sh=block.getAbsoluteShift();
      
      int xm=0; if (sh.x>0) xm=1;
      int ym=0; if (sh.y>0) ym=1;
      int zm=0; if (sh.z>0) zm=1;
      
      for (int x=loc.x; x<=loc.x+xm; ++x)
      {
        for (int y=loc.y; y<=loc.y+ym; ++y)
        {
          for (int z=loc.z; z<=loc.z+zm; ++z)
          {
            removeFromBlocks(new IsoLocation(x,y,z),block);
          };
        };
      };
      removeFromPaintOrder(block); 
      myJIsoWorld_.removeBlock(block.graphicalRepresentation());
    };
    object.setWorld(null);
  };
  
  
  /**
  * <p>Removes all <code>IsoObject</code>s from the world.</p>
  */
  public void removeAllIsoObjects()
  {
    Iterator i=blocks_.entrySet().iterator();
    while(i.hasNext())
    {
      Map.Entry ent=(Map.Entry)i.next();
      Collection c=(Collection)ent.getValue();
      removeObject(((IsoBlock)c.iterator().next()).parentObject());
      i=blocks_.entrySet().iterator();
    };
  };
  
  /**
  * <p>Adds <code>object</code> to the viewport as an object that does not
  * interact with the blocks from the world.</p>
  *
  * @see #removeNonWorldObject(JComponent)
  */
  public void addNonWorldObject(JComponent object)
  {
    graphicalRepresentation().getContentPane().add(object);
  };
  
  
  /**
  * <p>Removes <code>object</code> which must have been added using
  * {@link #addNonWorldObject(JComponent)}.</p>
  */
  public void removeNonWorldObject(JComponent object)
  {
    graphicalRepresentation().getContentPane().remove(object);
  };
  
  /**
  * <p>Returns the {@link Collection} of all {@link IsoBlock}s at 
  * <code>location</code> or <code>null</code> if there are none (meaning that
  * the returned {@link Collection} will never be empty). The returned
  * <code>Collection</code> includes blocks that overlap with <code>location</code>
  * due to shift.</p>
  * <p>Note, that for efficiency reasons, the returned <code>Collection</code> 
  * may not be a copy
  * but a reference to the actual internal structure used. DO NOT MODIFY IT!</p>
  */
  public Collection blocksAtLocation(IsoLocation location) //TESTED
  {
    return (Collection)blocks_.get(location);
  };
};

