/* Copyright (C) 2002 Matthias S. Benkmann <m.s.b@gmx.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License (ONLY THIS VERSION).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package domino.ui;

import java.util.*;
import javax.swing.*;
import java.awt.*;
import java.awt.geom.Point2D;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;


/**
* <p><code>JIsoWorld</code> is the class responsible for the graphical
* representation of an {@link IsoWorld}.</p>
*/
public class JIsoWorld extends JScrollPane
{
  /** 
  * <p><code>XDIM_MAX, YDIM_MAX and ZDIM_MAX</code> specify the maximum number of blocks
  * along the respective axis. The product of the 3 must not exceed 
  * {@link Integer#MAX_VALUE} because a unique layer is assigned to each block.</p>
  * <p>x coordinates of blocks in the <code>JIsoWorld</code> must be
  * in the range <code>]-XDIM_MAX/2,XDIM_MAX/2[</code>. The same applies to
  * y coordinates and <code>YDIM_MAX</code> and z coordinates and 
  * <code>ZDIM_MAX</code>.</p> 
  */
  private static final int XDIM_MAX=4096;
  /** <p>see {@link #XDIM_MAX}.</p>*/
  private static final int YDIM_MAX=4096;
  /** <p>see {@link #XDIM_MAX}.</p>*/
  private static final int ZDIM_MAX=64;
  
  /**
  * <p>When determining the size of the drawing canvas, space needs to be added to
  * make sure the top sides (which are offset due to the isometric view) of blocks 
  * at the sides are still on the canvas. Blocks with a z coordinate that exceeds this
  * constants value will not be completely visible anymore if placed at the very side
  * of the board.</p>
  */
  private static final int MAXZ_TO_TAKE_INTO_ACCOUNT_FOR_SIZING_BOARD=4;
  
  /**
  * <p>The minimum side length (in pixels) of a block's top and bottom side square.</p>
  */
  private static final int MIN_SQUARE_SIZE=10;
  
  /** 
  * <p><code>xoffsetRel_*squareSide_</code> is the number of pixels, the top side of
  * the block is shifted to the right relative to the bottom side.</p>
  */
  private double xoffsetRel_;
  /** 
  * <p><code>yoffsetRel_*squareSide_</code> is the number of pixels, the top side of
  * the block is shifted down relative to the bottom side.</p>
  */
  private double yoffsetRel_;
  
  /** <p><code>xoffsetRel_*squareSide_</code></p> */
  private int xoffset_;
  /** <p><code>yoffsetRel_*squareSide_</code></p> */
  private int yoffset_;
  
  /** <p>Side length of the bottom and top side square of a block.</p> */
  private int squareSide_;
  
  /** 
  * <p><code>plusxdim_</code> is the index of the right-most column of the board.</p>
  */
  private int plusxdim_;
  /** 
  * <p><code>-minusxdim_</code> is the index of the left-most column of the board
  *   (a negative number, because <code>minusxdim_&gt;=0</code>).</p> 
  */
  private int minusxdim_;
  /** 
  * <p><code>plusydim_</code> is the index of the up-most row of the board.</p>
  */
  private int plusydim_;
  /** 
  * <p><code>-minusydim_</code> is the index of the down-most row of the board
  *   (a negative number, because <code>minusydim_&gt;=0</code>).</p> 
  */
  private int minusydim_;
  
  /** Number of pixels added at the left side of the board to satisfy the fixpoint
  * restrictions from the last call to {@link #setSquareSize(int,double,double,int,int)}. */
  private int leftpadding_;
  /** Number of pixels added at the top of the board to satisfy the fixpoint
  * restrictions from the last call to {@link #setSquareSize(int,double,double,int,int)}. */
  private int toppadding_;
  /** Number of pixels added at the right side of the board to satisfy the fixpoint
  * restrictions from the last call to {@link #setSquareSize(int,double,double,int,int)}. */
  private int rightpadding_;
  /** Number of pixels added at the bottom of the board to satisfy the fixpoint
  * restrictions from the last call to {@link #setSquareSize(int,double,double,int,int)}. */
  private int bottompadding_;
  
  /** <p>Width of the board without padding.</p> */
  private int boardWidth_;
  /** <p>Height of the board without padding.</p> */
  private int boardHeight_;
  
  /** <p>The actual scroll area that contains the graphics objects for the blocks.</p> */
  private JIsoWorldPane myBoardPane_;
  
  /**
  * <p>Returns the number of pixels the top side of a block is moved to the right
  * when drawing, relative to the bottom side. This number changes if the squares
  * are resized.</p>
  * @see #yOffset()
  */
  public int xOffset() {return xoffset_;};
  /**
  * <p>Returns the number of pixels the top side of a block is moved down
  * when drawing, relative to the bottom side. This number changes if the squares
  * are resized.</p>
  * @see #xOffset()
  */
  public int yOffset() {return yoffset_;}; //TESTED
  
  
  /** 
  * <p>Returns the index of the right-most column of the board.</p>
  */
  public int plusXDim() { return plusxdim_;};
  /** 
  * <p>Returns abs() of the index of the left-most column of the board.</p> 
  */
  public int minusXDim() {return minusxdim_;};
  /** 
  * <p>Returns the index of the up-most row of the board.</p>
  */
  public int plusYDim() {return plusydim_;};
  /** 
  * <p>Returns abs() of the index of the down-most row of the board.</p> 
  */
  public int minusYDim() {return minusydim_;};
  
  /**
  * <p>Returns the canvas that contains the graphics objects for the blocks.</p>
  */
  public JComponent getContentPane()
  {
    return myBoardPane_;
  };
  
  /**
  * <p>Translates coordinates relative to the viewport to 
  * coordinates relative to the board. Padding is regarded as separate from the board,
  * so points in the padding area have negative board coordinates.</p>
  *
  * @see #translateToViewport(Point)
  * @see #translateNormalizedToViewport(Point2D.Double)
  * @see #translateToBoardNormalized(Point)
  * @see #translateToIsoWorld(Point, int)
  */
  public Point translateToBoard(Point p)
  {
    Point viewPos=getViewport().getViewPosition();
    return new Point(p.x+viewPos.x-leftpadding_, p.y+viewPos.y-toppadding_);
  };
  
  /**
  * <p>Translates coordinates relative to the board to 
  * coordinates relative to the viewport. 
  * Padding is regarded as separate from the board,
  * so points in the padding area have negative board coordinates.</p>
  *
  * @see #translateToBoard(Point)
  * @see #translateToBoardNormalized(Point)
  * @see #translateNormalizedToViewport(Point2D.Double)
   * @see #translateToIsoWorld(Point, int)
  */
  public Point translateToViewport(Point p)
  {
    Point viewPos=getViewport().getViewPosition();
    return new Point(p.x-viewPos.x+leftpadding_, p.y-viewPos.y+toppadding_);
  };
  
  
  /**
  * <p>Translates normalized board coordinates to 
  * (pixel) coordinates relative to the viewport. 
  * Padding is regarded as separate from the board,
  * so points in the padding area have negative board coordinates.</p>
  *
  * @see #translateToViewport(Point)
  * @see #translateToBoard(Point)
  * @see #translateToBoardNormalized(Point)
  * @see #translateToIsoWorld(Point, int)
  */
  public Point translateNormalizedToViewport(Point2D.Double p)
  {
    Point viewPos=getViewport().getViewPosition();
    return new Point((int)Math.round(p.x*boardWidth_)-viewPos.x+leftpadding_, (int)Math.round(p.y*boardHeight_)-viewPos.y+toppadding_);
  };
  
  
  /**
  * <p>Like {@link #translateToBoard(Point)} but returns coordinates relative to
  * the board's width and height.</p>
  *
  * @see #translateToViewport(Point)
  * @see #translateNormalizedToViewport(Point2D.Double)
  * @see #translateToBoard(Point)
  * @see #translateToIsoWorld(Point, int)
  */
  public Point2D.Double translateToBoardNormalized(Point p)
  {
    p=translateToBoard(p);
    return new Point2D.Double((double)p.x/(double)boardWidth_, (double)p.y/(double)boardHeight_);
  };
  

  /**
  * <p>Translates coordinates relative to the viewport to the {@link IsoLocation}
  * with z coordinate <code>z</code> that an {@link IsoBlock} would need to have for 
  * its top
  * side to be at that location.</p>
  */
  public IsoLocation translateToIsoWorld(Point p, int z)
  {
    return translateToIsoWorld(p,z,null);
  };
  
  /**
  * <p>Translates coordinates relative to the viewport to the {@link IsoLocation}
  * with z coordinate <code>z</code> that an {@link IsoBlock} would need to have for 
  * its top side to be at that location. If <code>sh!=null</code>, it will be
  * set to reflect the {@link IsoShift} the block would need to have in order to
  * have the upper left corner of the top side exactly at point <code>p</code>.</p>
  *
  * @see #translateToViewport(Point)
  * @see #translateNormalizedToViewport(Point2D.Double)
  * @see #translateToBoard(Point)
  * @see #translateToBoardNormalized(Point)
  */
  public IsoLocation translateToIsoWorld(Point p, int z, IsoShift sh)
  {
    p=translateToBoard(p);
    int width=squareSide_+Math.abs(xoffset_);
    int height=squareSide_+Math.abs(yoffset_);
    //no xzadd and yzadd in the following formulas!
    double x=(p.x-(1+z-MAXZ_TO_TAKE_INTO_ACCOUNT_FOR_SIZING_BOARD)*xoffset_-minusxdim_*squareSide_)/(double)squareSide_;
    double y=(p.y-(1+z-MAXZ_TO_TAKE_INTO_ACCOUNT_FOR_SIZING_BOARD)*yoffset_-minusydim_*squareSide_)/(double)squareSide_;
    int xi=(int)Math.floor(x);
    int yi=(int)Math.floor(y);
    if (sh!=null) {sh.x=x-xi; sh.y=y-yi; sh.z=0;};
    return new IsoLocation(xi,yi,z);
  };
 
  /**
  * <p>Moves the viewport <code>xdif</code> pixels to the right and
  * <code>ydif</code> pixels down.</p>
  */ 
  public void moveViewportRelative(int xdif, int ydif)
  {
    JViewport myViewport=getViewport();
    Point p=myViewport.getViewPosition();
    p.x+=xdif;
    p.y+=ydif;
    moveViewportAbsolute(p.x,p.y);
  }
  
  /**
  * <p>Moves the viewport so that the coordinates <code>(x,y)</code> of the world
  * appear in the upper left corner.</p>
  */ 
  public void moveViewportAbsolute(int x, int y)
  {  
    JViewport myViewport=getViewport();
    Dimension d=myViewport.getExtentSize();
    Dimension s=myViewport.getViewSize();
    int right=x+d.width-s.width;
    int bottom=y+d.height-s.height;
    if (right>0)  x-=right;
    if (bottom>0) y-=bottom;
    if (x<0) x=0;
    if (y<0) y=0;
    myViewport.setViewPosition(new Point(x,y));
  };
  
  /**
  * <p>Returns the current side length (in pixels) of the top and bottom square of
  * a block.</p>
  *
  * @see #setSquareSize(int,double,double,int,int)
  */
  public int getSquareSize() //TESTED
  {
    return squareSide_;
  };
  
  /**
  * <p>Creates a new JIsoWorld.</p>
  *
  * @param minusxdim <code>-minusxdim</code> is the index of the left-most column of 
  *        the board (a negative number, because <code>minusxdim&gt;=0</code>).
  * @param plusxdim the index of the right-most column of
  *        the board.
  * @param minusydim <code>-minusydim</code> is the index of the down-most row of 
  *        the board (a negative number, because <code>minusydim&gt;=0</code>).
  * @param plusydim the index of the up-most row of the board.
  * @param squareSide initial side length of the bottom and top side square of a block.
  * @param xoffsetRel <code>xoffsetRel_*squareSide</code> is the number of pixels, 
  *        the top side of the block is shifted to the right in comparison to the 
  *        bottom side.
  * @param yoffsetRel <code>yoffsetRel_*squareSide</code> is the number of pixels, 
  *        the top side of the block is shifted down in comparison to the 
  *        bottom side.
  */
  public JIsoWorld(int minusxdim, int plusxdim, int minusydim, int plusydim, int squareSide, double xoffsetRel, double yoffsetRel)
  { //TESTED
    super(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);

    //the following assignments must be performed before creating myBoardPane_
    xoffsetRel_=xoffsetRel;
    yoffsetRel_=yoffsetRel;
    squareSide_=squareSide;
    xoffset_=(int)Math.round(xoffsetRel*squareSide);
    yoffset_=(int)Math.round(yoffsetRel*squareSide);
    minusxdim_=minusxdim;
    plusxdim_=plusxdim;
    minusydim_=minusydim;
    plusydim_=plusydim;
    leftpadding_=0;
    toppadding_=0;
    rightpadding_=0;
    bottompadding_=0;

    myBoardPane_=new JIsoWorldPane();
    setViewportView(myBoardPane_);
    setBorder(null);
  };  

  
  /**
  * <p>Adds or removes rows or columns from the board.</p>
  * <p>This method only increases or decreases the visible part of the board.
  * Removing rows/colums will not remove any blocks located there and will not
  * prevent new blocks from being added there.</p>
  *
  * @param leftrighttopbottom Where to add/remove: 
  *        <code>0=left, 1=right, 2=top, 3=bottom</code>
  * @param numLines how much space to add. Negative values remove space.
  */
  public void addLines(int leftrighttopbottom, int numLines) 
  {
    Dimension d=getViewport().getExtentSize();
    Point viewportMiddle=new Point(d.width>>1,d.height>>1);
    Point boardP=translateToBoard(viewportMiddle);
    
    int old;
    int xplus=0;
    int yplus=0;
    switch(leftrighttopbottom)
    {
      case 0: old=minusxdim_;
              minusxdim_+=numLines;
              if (minusxdim_<0) minusxdim_=0;
              xplus=(minusxdim_-old)*squareSide_;
              boardP.x+=xplus;
              break;
              
      case 1: plusxdim_+=numLines;
              if (plusxdim_<0) plusxdim_=0;
              break;
              
      case 2: old=minusydim_;
              minusydim_+=numLines; 
              if (minusydim_<0) minusydim_=0;
              yplus=(minusydim_-old)*squareSide_;
              boardP.y+=yplus;
              break;
              
      case 3: plusydim_+=numLines; 
              if (plusydim_<0) plusydim_=0;
              break;
              
      default: throw new IllegalArgumentException();
    };
    
    myBoardPane_.setPrefSize();
    myBoardPane_.refocusViewport(xplus,yplus,boardP.x/(double)boardWidth_,boardP.y/(double)boardHeight_,viewportMiddle.x,viewportMiddle.y);
  };
  
  
  /**
  * <p>Adds <code>block</code> to the <code>JIsoWorld</code>.</p>
  */
  public void addBlock(JIsoBlock block,IsoLocation location, IsoShift shift) //TESTED
  {
    myBoardPane_.addBlock(block, location, shift);
  }; 
  
  /**
  * <p>Removes <code>block</code> from the <code>JIsoWorld</code>.</p>
  */
  public void removeBlock(JIsoBlock block) //TESTED
  {
    myBoardPane_.removeBlock(block);
  }; 
  
  /**
  * <p>Changes the size of the squares that form the top and bottom sides of the
  * blocks.</p>
  * <p>This method does not affect blocks already on the board. You have to call
  * {@link #setBlock(JIsoBlock,IsoLocation,IsoShift)} for each of these blocks, to 
  * update its position and size.</p>
  * <p>The center of the viewport will be a fixpoint, i.e. it will still show
  * the same location on the board after the resize.</p>
  *
  * @param sideLengthInPixels new side length of a square.
  *
  * @see #setSquareSize(int,double,double)
  * @see #setSquareSize(int,double,double,int,int)
  */
  public void setSquareSize(int sideLengthInPixels)
  {
    myBoardPane_.setSquareSize(sideLengthInPixels);
  };
  
  /**
  * <p>Changes the size of the squares that form the top and bottom sides of the
  * blocks.</p>
  * <p>This method does not affect blocks already on the board. You have to call
  * {@link #setBlock(JIsoBlock,IsoLocation,IsoShift)} for each of these blocks, to 
  * update its position and size.</p>
  *
  * @param sideLengthInPixels new side length of a square.
  * @param fixPointX x coordinate of fixpoint (normalized board coordinate). This
  *        board location will still be visible at the same position in the viewport
  *        after the resize.
  * @param fixPointY y coordinate of fixpoint (normalized board coordinate).
  *
  * @see #setSquareSize(int)
  * @see #setSquareSize(int,double,double,int,int)
  */
  public void setSquareSize(int sideLengthInPixels, double fixPointX, double fixPointY)
  {
    myBoardPane_.setSquareSize(sideLengthInPixels, fixPointX, fixPointY);
  };
  
  /**
  * <p>Changes the size of the squares that form the top and bottom sides of the
  * blocks.</p>
  * <p>This method does not affect blocks already on the board. You have to call
  * {@link #setBlock(JIsoBlock,IsoLocation,IsoShift)} for each of these blocks, to 
  * update its position and size.</p>
  *
  * @param sideLengthInPixels new side length of a square.
  * @param fixPointX x coordinate of fixpoint (normalized board coordinate).
  * @param fixPointY y coordinate of fixpoint (normalized board coordinate).
  * @param targetX x coordinate where fixpoint should be located after resize 
  *        (viewport coordinate).
  * @param targetY y coordinate where fixpoint should be located after resize 
  *        (viewport coordinate).
  *
  * @see #setSquareSize(int)
  * @see #setSquareSize(int,double,double)
  */
  public void setSquareSize(int sideLengthInPixels, double fixPointX, double fixPointY, int targetX, int targetY)
  { //TESTED
    myBoardPane_.setSquareSize(sideLengthInPixels, fixPointX, fixPointY, targetX, targetY);
  };
  
  
  /**
  * <p>Recomputes padding, repositions viewport and {@link Component}s.</p>
  *
  * @param xplus value to add to {@link Component}s' x coordinates.
  * @param yplus value to add to {@link Component}s' y coordinates.
  * @param fixPointX x coordinate of fixpoint (normalized board coordinate).
  * @param fixPointY y coordinate of fixpoint (normalized board coordinate).
  * @param targetX x coordinate where fixpoint should be located after resize 
  *        (viewport coordinate).
  * @param targetY y coordinate where fixpoint should be located after resize 
  *        (viewport coordinate).
  *
  */
  public void refocusViewport(int xplus, int yplus, double fixPointX, double fixPointY, int targetX, int targetY)
  {
    myBoardPane_.refocusViewport(xplus,yplus,fixPointX,fixPointY,targetX,targetY);
  };
  
  /**
  * <p>Adjusts position and size of <code>block</code> after changing the square
  * size, then causes a repaint.</p>
  *
  * @param block the block to adjust. This block must have been added to this
  *        <code>JIsoWorld</code>.
  * @param location the location of <code>block</code> within the world.
  * @param shift <code>block</code>'s shift.
  *
  * @see #setSquareSize(int,double,double,int,int)
  */
  public void setBlock(JIsoBlock block, IsoLocation location, IsoShift shift)
  {
    myBoardPane_.setBlock(block, location, shift);
  };
  
  
  /** <p>see {@link JLayeredPane#getLayer(Component)}.</p> */
  public int getLayer(Component c) //TESTED
  {
    return myBoardPane_.getLayer(c);
  };
  
  /** <p>see {@link JLayeredPane#setLayer(Component,int)}.</p> */
  public void setLayer(Component c, int layer)
  {
    myBoardPane_.setLayer(c, layer);
  };
  
  /** 
  * <p>Like {@link #setLayer(Component,int)} but instead of specifying the layer
  * explicitly, it is computed to be appropriate for a block at the given
  * <code>location</code>.</p>
  */
  public void setLayer(Component c, IsoLocation location)
  {
    myBoardPane_.setLayer(c, myBoardPane_.layerFor(location));
  };
  
  
  /**
  * <p>See {@link JLayeredPane#setPosition(Component,int)}.</p>
  */
  public void setPaintOrderPosition(Component c,int paintOrderPosition) //TESTED
  {
    myBoardPane_.setPosition(c,paintOrderPosition);
  };
  
  

//####################################################################################
//####################################################################################
//####################################################################################
//####################################################################################

    /**
    * <p>The content pane for a {@link JIsoWorld}.</p>
    */
    protected class JIsoWorldPane extends JLayeredPane implements Scrollable
    {
      public Dimension getPreferredScrollableViewportSize()
      {
        return getPreferredSize();
      };
  
      public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction)
      {
        return squareSide_;
      };
  
      public int getScrollableBlockIncrement(Rectangle visibleRect, int orientation, int direction)
      {
        return squareSide_;
      };
  
      public boolean getScrollableTracksViewportWidth() {return false;};
      public boolean getScrollableTracksViewportHeight() {return false;};
  

      /**
      * <p>Computes the size of the canvas and sets it as preferred size.</p>
      */
      private void setPrefSize() //TESTED
      {
        int cols=minusxdim_+plusxdim_+1;
        int rows=minusydim_+plusydim_+1;
        boardWidth_=squareSide_*cols+Math.abs(MAXZ_TO_TAKE_INTO_ACCOUNT_FOR_SIZING_BOARD*xoffset_);
        boardHeight_=squareSide_*rows+Math.abs(MAXZ_TO_TAKE_INTO_ACCOUNT_FOR_SIZING_BOARD*yoffset_);
        setPreferredSize(new Dimension(boardWidth_+leftpadding_+rightpadding_,boardHeight_+toppadding_+bottompadding_));
      };
  
      /** Creates a new <code>JIsoWorldPane</code>.*/
      public JIsoWorldPane()
      {
        setPrefSize();
        
        /*
        * We need to grow and shrink with our JIsoWorld to avoid ugly effects like
        * IsoObjects being cut off in the middle.
        */
        JIsoWorld.this.addComponentListener
        (
          new ComponentAdapter()
          {
            public void componentResized(ComponentEvent e)
            {
              Dimension d=getViewport().getExtentSize();
              Point2D.Double p=translateToBoardNormalized(new Point(d.width>>1,d.height>>1));
              refocusViewport(0, 0, p.x, p.y, d.width>>1, d.height>>1);
            };
          }
        );
      };
  

      /**
      * <p>Adjusts position and size of <code>block</code> after changing the square
      * size, then causes a repaint.</p>
      *
      * @param block the block to adjust. This block must have been added to this
      *        <code>JIsoWorld</code>.
      * @param location the location of <code>block</code> within the world.
      * @param shift <code>block</code>'s shift.
      *
      * @see #setSquareSize(int,double,double,int,int)
      */  
      public void setBlock(JIsoBlock block, IsoLocation location, IsoShift shift) //TESTED
      {
        int width=squareSide_+Math.abs(xoffset_);
        int height=squareSide_+Math.abs(yoffset_);
        int xzadd=(xoffset_<0) ? 1 : 0;
        int yzadd=(yoffset_<0) ? 1 : 0;
        int xpixel=(int)Math.round(leftpadding_ + ((location.x+minusxdim_)+shift.x)*squareSide_ + ((xzadd+location.z-MAXZ_TO_TAKE_INTO_ACCOUNT_FOR_SIZING_BOARD)+shift.z)*xoffset_);
        int ypixel=(int)Math.round(toppadding_ + ((location.y+minusydim_)+shift.y)*squareSide_ + ((yzadd+location.z-MAXZ_TO_TAKE_INTO_ACCOUNT_FOR_SIZING_BOARD)+shift.z)*yoffset_);
        JComponent comp=block.toJComponent();
        comp.setBounds(xpixel,ypixel,width,height);
        block.setIsoOffsets(xoffset_,yoffset_);
        //block.revalidate();
        comp.repaint();
      };
  
      /** <p>Returns the appropriate layer for a block at <code>location</code>.</p>*/
      public int layerFor(IsoLocation location) //TESTED
      {
        return (location.y+YDIM_MAX/2)*XDIM_MAX + (location.x+XDIM_MAX/2)*ZDIM_MAX + location.z+ZDIM_MAX/2;
      };
  
      /**
      * <p>Adds <code>block</code> to the <code>JIsoWorld</code>.</p>
      */
      public void addBlock(JIsoBlock block, IsoLocation location, IsoShift shift)//TESTED
      {
        this.add(block.toJComponent(),new Integer(layerFor(location)));
        setBlock(block, location, shift);
      };
      
      
      /**
      * <p>Removes <code>block</code> from the <code>JIsoWorld</code>.</p>
      */
      public void removeBlock(JIsoBlock block) //TESTED
      {
        Rectangle r=block.toJComponent().getBounds(null);
        this.remove(block.toJComponent());
        repaint(r);
      };
  
      
      /**
      * <p>Changes the size of the squares that form the top and bottom sides of the
      * blocks.</p>
      * <p>This method does not affect blocks already on the board. You have to call
      * {@link #setBlock(JIsoBlock,IsoLocation,IsoShift)} for each of these blocks, to 
      * update its position and size.</p>
      * <p>The center of the viewport will be a fixpoint, i.e. it will still show
      * the same location on the board after the resize.</p>
      *
      * @param sideLengthInPixels new side length of a square.
      *
      * @see #setSquareSize(int,double,double)
      * @see #setSquareSize(int,double,double,int,int)
      */
      public void setSquareSize(int sideLengthInPixels)
      {
        Dimension d=getViewport().getExtentSize();
        Point2D.Double p=translateToBoardNormalized(new Point(d.width>>1,d.height>>1));
        setSquareSize(sideLengthInPixels, p.x, p.y);
      };

      
      /**
      * <p>Changes the size of the squares that form the top and bottom sides of the
      * blocks.</p>
      * <p>This method does not affect blocks already on the board. You have to call
      * {@link #setBlock(JIsoBlock,IsoLocation,IsoShift)} for each of these blocks, to 
      * update its position and size.</p>
      *
      * @param sideLengthInPixels new side length of a square.
      * @param fixPointX x coordinate of fixpoint (normalized board coordinate). This
      *        board location will still be visible at the same position in the viewport
      *        after the resize.
      * @param fixPointY y coordinate of fixpoint (normalized board coordinate).
      *
      * @see #setSquareSize(int)
      * @see #setSquareSize(int,double,double,int,int)
      */
      public void setSquareSize(int sideLengthInPixels, double fixPointX, double fixPointY)
      {
        Point p=translateToViewport(new Point((int)Math.round(fixPointX*boardWidth_), (int)Math.round(fixPointY*boardHeight_)));
        setSquareSize(sideLengthInPixels, fixPointX, fixPointY, p.x, p.y);
      };
  

      /**
      * <p>Changes the size of the squares that form the top and bottom sides of the
      * blocks.</p>
      * <p>This method does not affect blocks already on the board. You have to call
      * {@link #setBlock(JIsoBlock,IsoLocation,IsoShift)} for each of these blocks, to 
      * update its position and size.</p>
      *
      * @param sideLengthInPixels new side length of a square.
      * @param fixPointX x coordinate of fixpoint (normalized board coordinate).
      * @param fixPointY y coordinate of fixpoint (normalized board coordinate).
      * @param targetX x coordinate where fixpoint should be located after resize 
      *        (viewport coordinate).
      * @param targetY y coordinate where fixpoint should be located after resize 
      *        (viewport coordinate).
      *
      * @see #setSquareSize(int)
      * @see #setSquareSize(int,double,double)
      */
      public void setSquareSize(int sideLengthInPixels, double fixPointX, double fixPointY, int targetX, int targetY)
      { //TESTED
        if (sideLengthInPixels<MIN_SQUARE_SIZE) sideLengthInPixels=MIN_SQUARE_SIZE;
        double scale=((double)sideLengthInPixels)/squareSide_;
        boardWidth_=(int)(boardWidth_*scale);
        boardHeight_=(int)(boardHeight_*scale);
        double newFixPointX=fixPointX*boardWidth_;
        double newFixPointY=fixPointY*boardHeight_;
        JViewport myViewport=getViewport();
        Point viewPos=new Point();
        viewPos.x=(int)Math.round(newFixPointX-targetX);
        viewPos.y=(int)Math.round(newFixPointY-targetY);
        leftpadding_=0;
        toppadding_=0;
        if (viewPos.x<0) {leftpadding_=-viewPos.x; viewPos.x=0;};
        if (viewPos.y<0) {toppadding_=-viewPos.y; viewPos.y=0;};
                
        Dimension d=myViewport.getExtentSize();
        rightpadding_=viewPos.x+d.width-boardWidth_;
        bottompadding_=viewPos.y+d.height-boardHeight_;
        if (rightpadding_<0) rightpadding_=0;
        if (bottompadding_<0) bottompadding_=0;
        
        squareSide_=sideLengthInPixels;
        xoffset_=(int)Math.round(xoffsetRel_*squareSide_);
        yoffset_=(int)Math.round(yoffsetRel_*squareSide_);
        setPrefSize();
    
        revalidate();
        myViewport.validate(); //force viewport to acknowledge new size of this
                               //so that it will not crop the new viewPos coordinates
        myViewport.setViewPosition(viewPos);
      };
    
    
    
      /**
      * <p>Recomputes padding, repositions viewport and {@link Component}s.</p>
      *
      * @param xplus value to add to {@link Component}s' x coordinates.
      * @param yplus value to add to {@link Component}s' y coordinates.
      * @param fixPointX x coordinate of fixpoint (normalized board coordinate).
      * @param fixPointY y coordinate of fixpoint (normalized board coordinate).
      * @param targetX x coordinate where fixpoint should be located after resize 
      *        (viewport coordinate).
      * @param targetY y coordinate where fixpoint should be located after resize 
      *        (viewport coordinate).
      *
      */
      public void refocusViewport(int xplus, int yplus, double fixPointX, double fixPointY, int targetX, int targetY)
      { 
        double newFixPointX=fixPointX*boardWidth_;
        double newFixPointY=fixPointY*boardHeight_;
        
        Point viewPos=new Point();
        viewPos.x=(int)Math.round(newFixPointX-targetX);
        viewPos.y=(int)Math.round(newFixPointY-targetY);
        int oldleftpadding=leftpadding_;
        int oldtoppadding=toppadding_;
        leftpadding_=0;
        toppadding_=0;
        if (viewPos.x<0) {leftpadding_=-viewPos.x; viewPos.x=0;};
        if (viewPos.y<0) {toppadding_=-viewPos.y; viewPos.y=0;};
        xplus+=(leftpadding_-oldleftpadding);
        yplus+=(toppadding_-oldtoppadding);
        
        if (xplus!=0 || yplus!=0)
        {
          Component[] comp=myBoardPane_.getComponents();
          for (int i=0; i<comp.length; ++i)
          {
            Point p=comp[i].getLocation();
            comp[i].setLocation(p.x+xplus,p.y+yplus);
          };
        };
        
        JViewport myViewport=getViewport();        
        Dimension d=myViewport.getExtentSize();
        rightpadding_=viewPos.x+d.width-boardWidth_;
        bottompadding_=viewPos.y+d.height-boardHeight_;
        if (rightpadding_<0) rightpadding_=0;
        if (bottompadding_<0) bottompadding_=0;
        
        setPrefSize();
    
        revalidate();
        myViewport.validate(); //force viewport to acknowledge new size of this
                               //so that it will not crop the new viewPos coordinates
        myViewport.setViewPosition(viewPos);
      };

    };
};

